!=========================================================================
! Copyright (C) VMware, Inc. 2010-2012.  All Rights Reserved.
!
! $Id: AbstractException2.gs $
!
!=========================================================================

! fix 41428
category: 'Private'
method: AbstractException
_errNumPrefix
  ^ '(error '
%

method: AbstractException
_description: subclassDetails

  "Create the full message text if needed, save it in messageText,
   and return it.
   Should be invoked by #description in a subclass. "

  messageText ifNotNil:[ :s | ^ s ].
  ^ [ | str num |
      (str := String withAll:'a ') add: self class name ; add: ' occurred ' .
      str add: self _errNumPrefix , (num := gsNumber) asString ; add: $) . 
      gsReason ifNotNil:[ :r| str add:', reason:' ; add: r asString ].
      subclassDetails ifNotNil:[
        "details instVar already included in subclassDetails if desired"
         str add:', '; add: subclassDetails .
      ] ifNil:[ | d |
        (d := gsDetails) ifNil:[
          d := (LegacyErrNumMap atOrNil: num) ifNotNil:[:m | m atOrNil: 3].
        ].
	d ifNotNil:[ 
	  d _isArray ifTrue:[ "details is a legacy message template"
	    str add:', ' ; add: (self _legacyDetails: d).
	  ] ifFalse:[
	    str add:', ' ; add: d asString "ANSI details" .
	  ]
        ].
      ].
      messageText := str .
      str
    ] onException: Error do:[ :ex | | s |
      s := 'error during Exception>>description'.
      [  
        messageText := s 
      ] onSynchronous: AbstractException do:[ :exB | 
        exB return "ignore errors during assignment"
      ].
      ex return: s
    ].
% 

classmethod: AbstractException
legacyNumberToClasses: aSmallInt
  "returns nil , a Class or an Array of Classes"
  |arr|
  arr := LegacyErrNumMap atOrNil: aSmallInt .
  "arr is { aClass . aReasonString . legacyTemplateStringOrNil }
      or { firstClass . aReasonString . legacyTemplateStringOrNil.
                { list of split classes...} } 
      per code in errordict.gs "

  arr ifNotNil:[ 
    (arr atOrNil: 4) ifNotNil:[ :arrayOfClasses | ^ arrayOfClasses] .
    ^ arr atOrNil:1  "a single class"
  ].
  ^ nil
%

classmethod: AbstractException
legacyNumberToClass: aSmallInt
  "returns a Class"
  |arr|
  arr := LegacyErrNumMap atOrNil: aSmallInt .
  "arr is { aClass . aReasonString . legacyTemplateStringOrNil }
      or { firstClass . aReasonString . legacyTemplateStringOrNil.
                { list of split classes...} } 
      per code in errordict.gs "

  arr ifNotNil:[ 
    ^ (arr atOrNil:1 ) ifNil:[ Error ].
  ].
  ^ Error 
%

! v3.0.1 more careful checking of gsNumber
method: AbstractException
_matchesExpectedLegacyNumber: aSmallInt
  "used by topaz EXPECTERROR command, returns true if 
   receiver is one of the classes produced by specified error number,
   false otherwise. "
^ [ | arr |
    (arr := LegacyErrNumMap atOrNil: aSmallInt) ifNil:[ ^ false ]. 
    (arr atOrNil: 4) ifNotNil:[ :inner | 
      (arr atOrNil: 1) ifNotNil:[ :c | (self isKindOf: c) ifTrue:[ ^ true ]].
      inner do:[:aClass | (self isKindOf: aClass) ifTrue:[ ^ true ]]
    ].
    (arr atOrNil: 1) ifNotNil:[ :c | 
       ^ (self isKindOf: c) and: [ self gsNumber == aSmallInt ]
    ].
    false
  ] onException: Error do:[:ex | ex return: false ]
%

! fixed 41414 , 41409 , 41814
method: CompileError
asString
  self errorDetails . "ensure error details are are populated"
  ^ self _description: nil
%
method: CompileError
errorDetails
  "Fill in the error message for each element of the error details,
   as a side effect of returning the error details."
  gsArgs ifNotNil:[ :arr |
    (arr atOrNil: 1) ifNotNil:[ :errorArrays | | numErrs n |
       numErrs := errorArrays size .
       n := 1 .
       [ n <= numErrs ] whileTrue:[ | anErr |
         anErr := errorArrays at: n . 
         (anErr atOrNil: 3) ifNil:[ "message not yet filled in" | errNum |
            (errNum := anErr atOrNil: 1 ) _isSmallInteger ifTrue:[ | msg |
              msg := LegacyErrNumMap atOrNil: errNum .
              msg _isOneByteString ifFalse:[ msg := 'unknown compiler error'].
              anErr at: 3 put: msg .
            ]
          ] ifNotNil:[ "messages already filled in"
            n := numErrs "early exit"
          ].
          n := n + 1 .
       ].
       ^ errorArrays
    ].
  ].
  ^ { { 1075"STDB_MALFORMED_COMPILER_ERROR" . 1 . 'malformed CompilerError'  } }
%

method: CompileError
gsArguments
  | ed |
  ed := self errorDetails "ensure error details are populated" .
  ^ gsArgs ifNil:[ { ed . nil"no source string" } ].
%
