! ========================================================================
! Copyright (C) by VMware, Inc. 1991-2012.  All Rights Reserved
! ========================================================================

expectvalue %String
run
(Object
	subclass: 'CDeclaration'
	instVarNames: #( header name storage
		type count pointer fields
		parameters enumTag isConstant
		includesCode isVaryingArgCount bitCount
		source file line)
	classVars: #()
	classInstVars: #()
	poolDictionaries: { }
	inDictionary: Globals
	options: #( #logCreation )) definition
%

! set class category
run
CDeclaration category: 'FFI'.
true
%

! Remove existing behavior from CDeclaration
doit
CDeclaration removeAllMethods.
CDeclaration class removeAllMethods.
true
%
! ------------------- Class methods for CDeclaration
category: 'Documentation'
set compile_env: 0
classmethod: CDeclaration
comment
^'This class is internal to the FFI implementation and 
represents a declaration from a C header file.'
%
category: 'other'
set compile_env: 0
classmethod: CDeclaration
header: aHeader
	"Return a CDeclaration (might not be self)"

	^(self readTypeFrom: aHeader) readDeclaration
%
category: 'other'
set compile_env: 0
classmethod: CDeclaration
readTypeFrom: aHeader
	"Return a CDeclaration (might not be self)"

	^self basicNew readTypeFrom: aHeader
%
! ------------------- Instance methods for CDeclaration
category: 'CByteArray get'
set compile_env: 0
method: CDeclaration
accessorForOffset: anInteger

	1 < count size ifTrue: [self error: 'Multi-dimensional arrays not supported!'].
	0 == pointer ifTrue: [
		type == #'class' ifTrue: [^self structAccessorForOffset: anInteger].
		(type isKindOf: CDeclaration) ifTrue: [^self structAccessorForOffset: anInteger].
		count == nil ifTrue: [^self simpleAccessorForOffset: anInteger].
		type == #'uint8' ifTrue: [^self stringAccessorForOffset: anInteger].
		^self arrayAccessorForOffset: anInteger.
	].
	1 < pointer ifTrue: [self error: 'Multi-level pointers not supported!'].
	count == nil ifTrue: [^self singlePointerAccessorForOffset: anInteger].
	type == #'uint8' ifTrue: [^self stringPointerAccessorForOffset: anInteger].
	^self arrayPointersAccessorForOffset: anInteger.
%
category: 'CCallout'
set compile_env: 0
method: CDeclaration
addSourceTo: aStream

	aStream 
		lf; tab; nextPut: $";
		nextPutAll: (file ifNil: ['test'] ifNotNil: [file]);
		nextPutAll: ' line ';
		nextPutAll: line printString;
		yourself.
	source do: [:char | aStream nextPut: char. char == $" ifTrue: [aStream nextPut: char]].
	aStream nextPut: $"; lf.
%
category: 'CCallout'
set compile_env: 0
method: CDeclaration
argumentType

	(self isSimple and: [self knownArgumentTypes includes: type]) ifTrue: [^type].
	(type == #'uint8' and: [isConstant]) ifTrue: [
		(count size == 1 and: [pointer == 0]) ifTrue: [^#'const char*'].		"const char foo[]"
		(count == nil and: [pointer == 1]) ifTrue: [^#'const char*'].			"const char * foo"
	].
	^#'ptr'.
%
category: 'CByteArray get'
set compile_env: 0
method: CDeclaration
arrayAccessorForOffset: anInteger

	^(WriteStream on: String new)
		nextPutAll: name; lf; lf; tab;
		nextPutAll: '| array offset |'; lf; tab;
		nextPutAll: 'array := Array new: ';
		nextPutAll: count first printString;
		nextPut: $.; lf; tab;
		nextPutAll: 'offset := ';
		nextPutAll: anInteger printString;
		nextPut: $.; lf; tab;
		nextPutAll: '1 to: array size do: [:i | '; lf; tab; tab;
		nextPutAll: 'array at: i put: (self ';
		nextPutAll: self getSelector;
		nextPutAll: ' offset).'; lf; tab; tab;
		nextPutAll: 'offset := offset + ';
		nextPutAll: self baseByteSize printString;
		nextPut: $.; lf; tab;
		nextPutAll: '].'; lf; tab;
		nextPutAll: '^array.'; lf;
		contents.
%
category: 'CByteArray get'
set compile_env: 0
method: CDeclaration
arrayPointersAccessorForOffset: anInteger

	^(WriteStream on: String new)
		nextPutAll: name; lf; lf; tab;
		nextPutAll: '| array offset |'; lf; tab;
		nextPutAll: 'array := Array new: ';
		nextPutAll: count first printString;
		nextPut: $.; lf; tab;
		nextPutAll: 'offset := ';
		nextPutAll: anInteger printString;
		nextPut: $.; lf; tab;
		nextPutAll: '1 to: array size do: [:i | '; lf; tab; tab;
		nextPutAll: 'array at: i put: (self pointerAt: offset resultClass: CByteArray).'; lf; tab; tab;
		nextPutAll: 'offset := offset + 8.'; lf; tab;
		nextPutAll: '].'; lf; tab;
		nextPutAll: '^array.'; lf;
		contents.
%
category: 'CByteArray put'
set compile_env: 0
method: CDeclaration
arrayPointersUpdatorForOffset: anInteger

	^(WriteStream on: String new)
		nextPutAll: name;
		nextPutAll: ': anArrayOfCByteArrayInstances'; lf; lf; tab;
		nextPutAll: '| offset |'; lf; tab;
		nextPutAll: 'offset := ';
		nextPutAll: anInteger printString;
		nextPut: $.; lf; tab;
		nextPutAll: '1 to: anArrayOfCByteArrayInstances size do: [:i | '; lf; tab; tab;
		nextPutAll: 'self'; lf; tab; tab; tab;
		nextPutAll: 'pointerAt: offset'; lf; tab; tab; tab; 
		nextPutAll: 'put: (anArrayOfCByteArrayInstances at: i).'; lf; tab; tab;
		nextPutAll: 'offset := offset + 8.'; lf; tab;
		nextPutAll: '].'; lf;
		contents.
%
category: 'CByteArray put'
set compile_env: 0
method: CDeclaration
arrayUpdatorForOffset: anInteger

	^(WriteStream on: String new)
		nextPutAll: name;
		nextPutAll: ': aSequenceableCollection'; lf; lf; tab;
		nextPutAll: '| offset |'; lf; tab;
		nextPutAll: 'offset := ';
		nextPutAll: anInteger printString;
		nextPut: $.; lf; tab;
		nextPutAll: '1 to: aSequenceableCollection size do: [:i | '; lf; tab; tab;
		nextPutAll: 'self '; lf; tab; tab; tab;
		nextPutAll: self getSelector; space;
		nextPutAll: 'offset'; lf; tab; tab; tab;
		nextPutAll: 'put: (aSequenceableCollection at: i).'; lf; tab; tab;
		nextPutAll: 'offset := offset + ';
		nextPutAll: self baseByteSize printString;
		nextPut: $.; lf; tab;
		nextPutAll: '].'; lf;
		contents.
%
category: 'CCallout'
set compile_env: 0
method: CDeclaration
asClassVarName

	^'Func' , name.
%
category: 'other'
set compile_env: 0
method: CDeclaration
atEnd

	^header atEnd.
%
category: 'CByteArray get'
set compile_env: 0
method: CDeclaration
baseAccessorForOffset: anInteger

	^'self ' , self getSelector , ' ' , anInteger printString.
%
category: 'size'
set compile_env: 0
method: CDeclaration
baseByteSize

	0 < pointer ifTrue: [^8].
	bitCount ifNotNil: [^bitCount].
	(type == #'int8' or: [type == #'uint8']) ifTrue: [^1].
	(type == #'int16' or: [type == #'uint16']) ifTrue: [^2].
	(type == #'int32' or: [type == #'uint32']) ifTrue: [^4].
	(type == #'int64' or: [type == #'uint64']) ifTrue: [^8].
	(type == #'void' and: [0 < pointer]) ifTrue: [^8].
	type == #'class' ifTrue: [^self baseStructByteSize].
	type == #'struct' ifTrue: [^self baseStructByteSize].
	type == #'union' ifTrue: [^self baseUnionByteSize].
	type == #'enum' ifTrue: [^4].
	type == #'function' ifTrue: [^8].
	type == #'double' ifTrue: [^8].
	(type isKindOf: CDeclaration) ifTrue: [^type baseByteSize].
	self halt.
%
category: 'size'
set compile_env: 0
method: CDeclaration
baseStructByteSize

	| bits |
	bits := 0.
	fields do: [:each |  
		each bitCount ifNil: [
			bits := bits + (each byteSize * 8).
		] ifNotNil: [
			each count ifNotNil: [self error: 'Array of bits!?'].
			bits := bits + each bitCount.
		].
	].
	^(bits / 8) ceiling.
%
category: 'size'
set compile_env: 0
method: CDeclaration
baseUnionByteSize

	^fields 
		inject: 0
		into: [:max :each | max max: each baseByteSize].
%
category: 'other'
set compile_env: 0
method: CDeclaration
beExternStorage

	storage := #'extern'.
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
bitCount

	^bitCount.
%
category: 'size'
set compile_env: 0
method: CDeclaration
byteSize

	| x |
	x := self baseByteSize.
	count ifNil: [^x].
	count do: [:y | x := x * y].
	^x.
%
category: 'CCallout'
set compile_env: 0
method: CDeclaration
cCalloutForLibrary: aCLibrary

	^CCallout
		library: aCLibrary 
		name: name 
		result: self resultType 
		args: (parameters collect: [:each | each argumentType])
		varArgsAfter: (isVaryingArgCount ifTrue: [parameters size] ifFalse: [-1]).
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
count

	^count.
%
category: 'CCallout'
set compile_env: 0
method: CDeclaration
createFunctionInitializerCode

	| stream space |
	stream := WriteStream on: String new.
	stream
		nextPutAll: self initializerFunctionName;
		nextPutAll: ' cLibrary'; lf; lf; tab;
		nextPutAll: self asClassVarName;
		nextPutAll: ' := CCallout'; lf; tab; tab;
		nextPutAll: 'library: cLibrary'; lf; tab; tab;
		nextPutAll: 'name: ''';
		nextPutAll: name; 
		nextPut: $'; lf; tab; tab;
		nextPutAll: 'result: #''';
		nextPutAll: self resultType;
		nextPutAll: ''''; lf; tab; tab;
		nextPutAll: 'args: #(';
		yourself.
	space := ''.
	parameters do: [:each | 
		stream 
			nextPutAll: space;
			nextPutAll: each argumentType printString;
			yourself.
		space := ' '.
	].
	^stream
		nextPutAll: ')'; lf; tab; tab;
		nextPutAll: 'varArgsAfter: ';
		nextPutAll: self varArgsAfter printString;
		nextPut: $.; lf;
		contents.
%
category: 'querying'
set compile_env: 0
method: CDeclaration
derivedType

	storage == #'typedef' ifFalse: [^self].
	type == #'class' ifTrue: [^self].
	self isFunction ifTrue: [^self].
	^type
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
enumTag

	^enumTag.
%
category: 'other'
set compile_env: 0
method: CDeclaration
errorExpectedIdentifier

	self error: 'Expected an identifier'.
%
category: 'other'
set compile_env: 0
method: CDeclaration
expectCloseParenthesisCount: anInteger

	anInteger timesRepeat: [
		| token |
		(token := self next) isCloseParenthesisToken ifFalse: [self error: 'Expected close parenthesis'].
	].
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
fields

	^type == #'struct'
		ifTrue: [fields]
		ifFalse: [nil].
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
file

	^file.
%
category: 'CByteArray get'
set compile_env: 0
method: CDeclaration
getSelector

	(#(
		#'int8' #'int16' #'int32' #'int64'
		#'uint8' #'uint16' #'uint32' #'uint64'
	) includes: type) ifTrue: [^type , 'At:'].
	type == #'enum' ifTrue: [^'int32At:'].
	type == #'function' ifTrue: [^'uint64At:'].
	self halt.
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
header

	^header.
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
includesCode

	^includesCode.
%
category: 'type'
set compile_env: 0
method: CDeclaration
initializeAsStructOrUnion

	| token |
	storage := #'auto'.
	(token := self peek) isOpenCurlyBracketToken ifFalse: [^self].
	self next.
	fields ifNotNil: [self error: 'Fields already defined!'].
	fields := Array new.
	[
		(token := self peek) isCloseCurlyBracketToken not.
	] whileTrue: [
		| field |
		field := self class header: header.
		field storage == #'auto' ifFalse: [self error: 'Unexpected storage!'].
		[
			fields add: field.
			(token := self next) isCommaToken.
		] whileTrue: [
			field := field copy.
			field readDeclaration.
		].
		token isSemicolonToken ifFalse: [self error: 'Unexpected end-of-field!'].
	].
	self next.
%
category: 'CCallout'
set compile_env: 0
method: CDeclaration
initializerFunctionName
	^'init_', name, '_inLibrary:'
%
category: 'type'
set compile_env: 0
method: CDeclaration
interpretTypeSpecifier

	type class == Array ifFalse: [^self].
	(type isEmpty and: [storage == #'extern' and: [name = 'C' or: [name = 'C++']]]) ifTrue: [type := nil. ^self].
	(type size == 1 and: [#(#'void' #'enum' #'float' #'class') includes: type first]) ifTrue: [type := type first. ^self].

	(type includes: #'char') ifTrue: [
		type removeIdentical: #'char'.
		type = #(#'signed') ifTrue: [type := #'int8'. ^self].
		(type isEmpty or: [type = #(#'unsigned')]) ifTrue: [type := #'uint8'. ^self].
		self halt.
	].
	(type includes: #'double') ifTrue: [
		type removeIdentical: #'double'.
		type isEmpty ifTrue: [type := #'double'. ^self].
		type = #(#'long') ifTrue: [type := #'longDouble'. ^self].
		self halt.
	].
	(type isEmpty or: [#(#'int' #'unsigned' #'long' #'signed' #'short') anySatisfy: [:each | type includes: each]]) ifTrue: [
		| isUnsigned |
		(type includes: #'int') ifTrue: [type removeIdentical: #'int'].
		(isUnsigned := type includes: #'unsigned') ifTrue: [type removeIdentical: #'unsigned'].
		(type includes: #'signed') ifTrue: [type removeIdentical: #'signed'].
		type = #(#'long') ifTrue: [
			type := isUnsigned ifTrue: [#'uint64'] ifFalse: [#'int64'].
			^self.
		].
		type = #(#'long' #'long') ifTrue: [
			type := isUnsigned ifTrue: [#'uint64'] ifFalse: [#'int64'].
			^self.
		].
		type = #(#'short') ifTrue: [
			type := isUnsigned ifTrue: [#'uint16'] ifFalse: [#'int16'].
			^self.
		].
		type isEmpty ifTrue: [
			type := isUnsigned ifTrue: [#'uint32'] ifFalse: [#'int32'].
			^self.
		].
		self halt.
	].
	(type size == 1 and: [#(#'uint64' #'int64' #'uint32' #'int32' #'uint16' #'int16' #'uint8' #'int8') includes: type first]) ifTrue: [
		type := type first.
		^self.
	].
	(type size == 1 and: [type first class == CDeclaration]) ifTrue: [type := type first. ^self].
	self halt.
%
category: 'CCallout'
set compile_env: 0
method: CDeclaration
invokeFunctionCode

	| args stream delimiter |
	args := parameters collect: [:each | each name].
	1 to: args size do: [:i | 
		(args at: i) ifNil: [args at: i put: 'arg' , i printString].
	].
	stream := (WriteStream on: String new)
		nextPutAll: name;
		yourself.
	args do: [:each | 
		stream 
			nextPutAll: '_: ';
			nextPutAll: each;
			space;
			yourself.
	].
	self addSourceTo: stream.
	stream
		lf; tab; 
		nextPutAll: '"Interpreted as #';
		nextPutAll: self resultType;
		nextPutAll: ' from #(';
		yourself.
	parameters do: [:each | stream space. each argumentType printOn: stream].
	stream 
		nextPutAll: ' )';
		nextPut: $"; lf; lf; 
		tab; nextPut: $^;
		nextPutAll: self asClassVarName;
		nextPutAll: ' callWith: { ';
		yourself.
	delimiter := ''.
	args do: [:each | 
		stream 
			nextPutAll: delimiter;
			nextPutAll: each;
			yourself.
		delimiter := '. '.
	].
	isVaryingArgCount ifTrue: [stream nextPutAll: '"varArgs array should be appended to this array"'].
	^stream 
		nextPutAll: ' }'; lf;
		contents.
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
isConstant

	^isConstant.
%
category: 'querying'
set compile_env: 0
method: CDeclaration
isEmptyVoid

	^name == nil and: [
		storage == #'auto' and: [
		type == #'void' and: [
		count == nil and: [
		pointer == 0 and: [
		fields == nil and: [
		isConstant == false and: [
		self isFunction == false and: [
		bitCount == nil]]]]]]]].
%
category: 'querying'
set compile_env: 0
method: CDeclaration
isFunction

	^parameters notNil.
%
category: 'CCallout'
set compile_env: 0
method: CDeclaration
isSimple

	^count isNil and: [pointer = 0].
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
isVaryingArgCount

	^isVaryingArgCount.
%
category: 'CCallout'
set compile_env: 0
method: CDeclaration
knownArgumentTypes

	^#( #int64 #uint64 #int32 #uint32 #int16 #uint16 #int8 #uint8 #bool #double #float #ptr #'char*' "#'&ptr'" #'const char*' ).
%
category: 'CCallout'
set compile_env: 0
method: CDeclaration
knownResultTypes

	^#( #int64 #uint64 #int32 #uint32 #int16 #uint16 #int8 #uint8 #bool #double #float #ptr #'char*' #void ).
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
line

	^line.
%
category: 'declarator'
set compile_env: 0
method: CDeclaration
mayHaveCode

	storage == #'static' 	ifTrue: [^true].
	storage == #'inline' 	ifTrue: [^true].
	type == #'class'		ifTrue: [^true].
	(type class == CDeclaration and: [type type == #'class']) ifTrue: [^true].
	^false
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
members

	^type == #'union'
		ifTrue: [fields]
		ifFalse: [nil].
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
name

	^name.
%
category: 'other'
set compile_env: 0
method: CDeclaration
next

	^self peek; _next.
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
parameters

	^parameters.
%
category: 'other'
set compile_env: 0
method: CDeclaration
peek

	| next wasIdentifierInSystemNamespace |
	wasIdentifierInSystemNamespace := false.
	[
		| wasEmpty |
		next := header peek.
		next ifNil: [^nil].
		wasEmpty := next type == #'empty'.
		wasEmpty ifFalse: [
			| value |
			(wasIdentifierInSystemNamespace and: [next isOpenParenthesisToken]) ifTrue: [
				self skipFunction.
				^self peek.
			] ifFalse: [
				wasIdentifierInSystemNamespace := next isIdentifierToken and: [2 < (value := next value) size and: 
					[(value at: 1) == $_ and: [(value at: 2) == $~]]].
			].
		].
		wasEmpty or: [wasIdentifierInSystemNamespace].
	] whileTrue: [
		self _next.
	].
	^next.
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
pointer

	^pointer.
%
category: 'other'
set compile_env: 0
method: CDeclaration
printFunctionParametersOn: aStream

	| comma |
	(name isNil or: [name last ~~ $(]) ifTrue: [
		aStream nextPut: $(.
	].
	(parameters isNil or: [parameters isEmpty]) ifTrue: [aStream nextPutAll: 'void)'. ^self].
	comma := ''.
	parameters do: [:each | 
		aStream nextPutAll: comma.
		each printOn: aStream.
		comma := ', '.
	].
	aStream nextPut: $).
%
category: 'other'
set compile_env: 0
method: CDeclaration
printOn: aStream

	storage ifNotNil: [
		(storage == #'auto' or: [storage == #'typedef' and: [type == #'class']]) ifFalse: [aStream nextPutAll: storage; space].
	].
	isConstant ifTrue: [aStream nextPutAll: 'const '].
	((type isKindOf: CDeclaration) and: [type isFunction]) ifTrue: [
		type printOn: aStream from: self.
		^self.
	].
	type ~= #() ifTrue: [
		(type isKindOf: Symbol)
			ifTrue: [aStream nextPutAll: type]
			ifFalse: [type printOn: aStream]. 
		aStream space.
	].
	pointer timesRepeat: [aStream nextPut: $*].
	enumTag ifNotNil: [aStream nextPutAll: enumTag; space].
	(storage == #'typedef' and: [self isFunction and: [0 < pointer]]) ifTrue: [aStream nextPut: $(].
	name ifNotNil: [aStream nextPutAll: name].
	count ifNotNil: [
		count do: [:each | 
			aStream nextPut: $[.
			0 < each ifTrue: [each printOn: aStream].
			aStream nextPut: $].
		].
	].
	(storage == #'typedef' and: [self isFunction and: [0 < pointer]]) ifTrue: [aStream nextPut: $)].
	self isFunction ifTrue: [self printFunctionParametersOn: aStream].
%
category: 'other'
set compile_env: 0
method: CDeclaration
printOn: aStream from: aCDeclaration

	| open close |
	(type isKindOf: Symbol) ifTrue: [
		aStream nextPutAll: type.
	] ifFalse: [
		aStream nextPutAll: '<???>'.
	].
	aStream space.
	0 == aCDeclaration pointer ifTrue: [
		open := ''.
		close := ''.
	] ifFalse: [
		open := '('. 
		close := ')'.
	].
	pointer timesRepeat: [aStream nextPut: $*].
	aStream nextPutAll: open.
	aCDeclaration pointer timesRepeat: [aStream nextPut: $*].
	(name notNil and: [storage ~~ #'typedef']) ifTrue: [aStream nextPutAll: name].
	aStream nextPutAll: close.
	self printFunctionParametersOn: aStream.
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
properties

	^type == #'class'
		ifTrue: [fields]
		ifFalse: [nil].
%
category: 'attributes'
set compile_env: 0
method: CDeclaration
readAttributes
"
	http://tigcc.ticalc.org/doc/gnuexts.html
"
	| token flag |
	self peek isOpenParenthesisToken ifFalse: [self error: 'Expected parenthesis'].
	self next.
	self peek isOpenParenthesisToken ifFalse: [self error: 'Expected parenthesis'].
	self next.
	(token := self peek) key == #'identifier' ifFalse: [^self].
	token value = 'deprecated'		ifTrue: [self next. ^self readDeprecatedAttribute].
	token value = '__mode__' 		ifTrue: [self next. ^self readModeAttributes].
	token value = '__nonnull__' 	ifTrue: [self next. ^self readNonnullAttribute].
	token value = '__format__'		ifTrue: [self next. ^self readFormatAttribute].
	token value = '__sentinel__'	ifTrue: [self next. ^self readNonnullAttribute].
	flag := #(
		'__const__'
		'__deprecated__'
		'__malloc__'
		'__noreturn__'
		'__nothrow__'
		'__pure__'
		'__transparent_union__'
		'__warn_unused_result__'
	) includes: token value.
	flag ifTrue: [
		self next. 
		self expectCloseParenthesisCount: 2. 
		^self
	].
%
category: 'other'
set compile_env: 0
method: CDeclaration
readBitFields
	"We parse bit fields, but we don't handle them properly!"

	bitCount := CPreprocessor evaluateHeader: header.
	(bitCount isKindOf: SmallInteger) ifFalse: [self error: 'Expected a SmallInteger'].
%
category: 'type'
set compile_env: 0
method: CDeclaration
readComplexType
"
>>--+--struct--+--+-----identifier---------------------------+--><
      |              |   |                        .----------------.     |
      '--union----'    |                       V                   |     |
                          '--+----------+--{----member--;--+--}-'
                             '-identifier-'
"
	self peek isIdentifierToken ifTrue: [
		| dictionary |
		name := self next value.
		dictionary := type == #'struct'
			ifTrue: [header structs]
			ifFalse: [header unions].
		type := dictionary
			at: name
			ifAbsentPut: [self copy].
	] ifFalse: [
		type := self copy.
	].
	type fields ifNil: [type initializeAsStructOrUnion].
	name := nil.
%
category: 'declarator'
set compile_env: 0
method: CDeclaration
readConstantExpression

	| token |
	[
		self atEnd not and: [self peek isOpenSquareBracketToken].
	] whileTrue: [
		| value |
		self next.
		count ifNil: [count := Array new].
		value := self peek isCloseSquareBracketToken 
			ifTrue: [0]
			ifFalse: [CPreprocessor evaluateHeader: header].
		(value isKindOf: SmallInteger) ifFalse: [self error: 'Expected a SmallInteger'].
		count add: value.
		(token := self next) isCloseSquareBracketToken ifFalse: [self halt].
	].
%
category: 'declarator'
set compile_env: 0
method: CDeclaration
readDeclaration
	"Return a CDeclaration (might not be self)
	   .-,------------------------------.
	  V                                      |
	----declarator--+-------------+-+--;--><
						  '--initializer--'

	We handle only one declarator; the caller (in CHeader>>#'readOne') handles the comma.
"
	| declaration token1 token2 |
	token1 := self peek.	"for debugging"
	(declaration := self readDeclarator) ifNil: [
		^self.
	].
	declaration includesCode ifTrue: [^declaration].
	token2 := self peek.	"for debugging"
	^declaration 
		readInitializer;
		readPostlude;
		yourself
%
category: 'declarator'
set compile_env: 0
method: CDeclaration
readDeclarator
	"Return a CDeclaration (might not be self)
	>>--+-------------------------+--directDeclarator--><
	      |  .---------------------.  |
	      |  V                        |  |
	      '----pointerOperator--+-'
"
	| declaration token1 token2 |
	token1 := self peek.	"for debugging"
	(declaration := self readPointerOperator) ifNotNil: [
		token2 := self peek.	"for debugging"
		^declaration readDeclarator.
	].
	^self readDirectDeclarator.
%
category: 'attributes'
set compile_env: 0
method: CDeclaration
readDeprecatedAttribute

	| closeCount |
	closeCount := self peek isOpenParenthesisToken
		ifTrue: [self next. 3]
		ifFalse: [2].
	[
		self next isCloseParenthesisToken.
	] whileFalse: [].
	self expectCloseParenthesisCount: closeCount.
%
category: 'declarator'
set compile_env: 0
method: CDeclaration
readDirectDeclarator
	"Return a CDeclaration (might not be self)
	>>--+-----------------------------------------------------------------------------------------------------------------+-><
	      +--declaratorName---------------------------------------------------------------------------------------------+
	      +--directDeclarator--(--parameterDeclarationList--)--+----------------+---+----------------------------+--+
	      |                                                                       '--cvQualifiers--'     '--exceptionSpecification--'    |
	      +--directDeclarator--[--+-----------------------+--]-----------------------------------------------------------+
	      |                                '-constantExpression-'                                                                           |
	      '--(--declarator--)-----------------------------------------------------------------------------------------------'
"
	| token declarator |
	(token := self peek) isNil ifTrue: [^self].
	token isIdentifierToken ifTrue: [
		name := token value.
		(token := self next; peek) isNil ifTrue: [^self].
	] ifFalse: [
		((type class == CDeclaration and: [type type == #'class']) not and: [token isOpenParenthesisToken]) ifTrue: [
			self next.	"skip the open parenthesis token"
			declarator := CDeclaration header: header.
			(token := self next) isCloseParenthesisToken ifFalse: [self syntaxError].
			(token := self peek) isNil ifTrue: [self syntaxError].
		].
	].
	token isOpenParenthesisToken ifTrue: [
		self readParameterDeclarationList.
		declarator isNil ifTrue: [^self].
		declarator 
			setStorage: storage
			type: self.
		storage := #'auto'.
		^declarator
	].
	token isOpenSquareBracketToken ifTrue: [
		declarator notNil ifTrue: [self syntaxError].
		self readConstantExpression.
		^self
	].
	declarator isNil ifTrue: [^self].
	"this is a specialized check for a function argument of the form 'struct dirent *(*[])'."
	(declarator storage == #'auto' and: [
		declarator type == #'int32' and: [
		declarator name isNil and: [
		declarator count = #(0)]]]) ifTrue: [
			pointer := pointer + declarator pointer + 1.
		].
	^self
%
category: 'type'
set compile_env: 0
method: CDeclaration
readEnumeration

	| token integralConstant |
	token := self peek.
	token isIdentifierToken ifTrue: [
		enumTag := self next value.
		token := self peek.
	].
	(header enumTags includesKey: enumTag) ifTrue: [
		token isIdentifierToken ifTrue: [
			name := self next value.
			^self.
		].
	].
	token isOpenCurlyBracketToken ifFalse: [^self].
	self next.
	integralConstant := -1.
	[
		(token := self next) isCloseCurlyBracketToken.
	] whileFalse: [
		| identifier |
		token isIdentifierToken ifFalse: [self errorExpectedIdentifier].
		identifier := token value.
		integralConstant := self peek isEqualsToken ifTrue: [
			self next.
			integralConstant := CPreprocessor evaluateHeader: header.
			(integralConstant isKindOf: SmallInteger) ifFalse: [self error: 'Expected a SmallInteger'].
			integralConstant.
		] ifFalse: [
			integralConstant + 1.
		].
		(header enums includesKey: identifier) ifTrue: [self error: 'Duplicate enum definition!'].
		header enums at: identifier put: integralConstant.
		self peek isCommaToken ifTrue: [self next].
	].
%
category: 'attributes'
set compile_env: 0
method: CDeclaration
readFormatAttribute
"
 ... __attribute__ ( ( __format__ ( __printf__ , 3 , 4 ) ) ) ;
"
	| token |
	self peek isOpenParenthesisToken ifFalse: [self error: 'Expected parenthesis'].
	self next.
	(token := self next) key == #'identifier' ifFalse: [self errorExpectedIdentifier].
	2 timesRepeat: [
		(token := self next) isCommaToken ifFalse: [self error: 'Expected comma'].
		(token := self next) key == #'int16' ifFalse: [self error: 'Expected number'].
	].
	self expectCloseParenthesisCount: 3.
%
category: 'declarator'
set compile_env: 0
method: CDeclaration
readInitializer
	"Since we don't have a real expression parser, we will go till we reach
	a comma, semicolon, or close parenthesis not in a nesting"

	| token parenthesisCount curlyBracketCount |
	(token := self peek) isNil ifTrue: [^self].
	token isEqualsToken ifFalse: [^self].
	self next.
	parenthesisCount := 0.
	curlyBracketCount := 0.
	[true] whileTrue: [
		(token := self peek) ifNil: [self syntaxError].
		(parenthesisCount == 0 and: [curlyBracketCount == 0]) ifTrue: [
			token isCommaToken ifTrue: [^self].
			token isSemicolonToken ifTrue: [^self].
			token isCloseParenthesisToken ifTrue: [^self].
		].
		token isOpenParenthesisToken ifTrue: [parenthesisCount := parenthesisCount + 1] ifFalse: [
		token isCloseParenthesisToken ifTrue: [parenthesisCount := parenthesisCount - 1] ifFalse: [
		token isOpenCurlyBracketToken ifTrue: [curlyBracketCount := curlyBracketCount + 1] ifFalse: [
		token isCloseCurlyBracketToken ifTrue: [curlyBracketCount := curlyBracketCount - 1]]]].
		self next.
	].
%
category: 'attributes'
set compile_env: 0
method: CDeclaration
readModeAttributes
"
	http://tigcc.ticalc.org/doc/gnuexts.html#SEC91_mode
	http://www.linuxtopia.org/online_books/programming_tool_guides/linux_using_gnu_compiler_collection/vector-extensions.html
		QI = 8-bit integer, HI = 16-bit integer, SI = 32-bit integer, DI = 64-bit integer, SF = 32-bit float, DF = 64-bit float
		byte = 8-bit integer, word = 64-bit integer, pointer = 64-bit integer

	typedef int int8_t __attribute__ ( ( __mode__ ( __QI__ ) ) ) ;
"
	| token mode |
	self peek isOpenParenthesisToken ifFalse: [self error: 'Expected parenthesis'].
	self next.
	(token := self next) key == #'identifier' ifFalse: [self error: 'Expected identifier'].
	mode := token value.
	type := 
		mode = '__QI__' ifTrue: [type first == $u ifTrue: [#'uint8'] ifFalse: [#'int8']] ifFalse: [
		mode = '__HI__' ifTrue: [type first == $u ifTrue: [#'uint16'] ifFalse: [#'int16']] ifFalse: [
		mode = '__SI__' ifTrue: [type first == $u ifTrue: [#'uint32'] ifFalse: [#'int32']] ifFalse: [
		mode = '__DI__' ifTrue: [type first == $u ifTrue: [#'uint64'] ifFalse: [#'int64']] ifFalse: [
		mode = '__byte__' ifTrue: [type first == $u ifTrue: [#'uint8'] ifFalse: [#'int8']] ifFalse: [
		mode = '__word__' ifTrue: [type first == $u ifTrue: [#'uint64'] ifFalse: [#'int64']] ifFalse: [
		mode = '__pointer__' ifTrue: [type first == $u ifTrue: [#'uint64'] ifFalse: [#'int64']] ifFalse: [
		self halt]]]]]]].
	self expectCloseParenthesisCount: 3.
%
category: 'attributes'
set compile_env: 0
method: CDeclaration
readNonnullAttribute

	| closeCount |
	self peek isOpenParenthesisToken ifFalse: [self error: 'Expected parenthesis'].
	self next.
	closeCount := self peek isOpenParenthesisToken
		ifTrue: [self next. 3]
		ifFalse: [2].
	[
		self next isCloseParenthesisToken.
	] whileFalse: [].
	self expectCloseParenthesisCount: closeCount.
%
category: 'declarator'
set compile_env: 0
method: CDeclaration
readParameterDeclarationList

	| token |
	self next.	"open parenthesis token"
	parameters := Array new.
	[
		(token := self peek) isCloseParenthesisToken.
	] whileFalse: [
		self peek isDotToken ifTrue: [
			self next.
			2 timesRepeat: [(token := self next) isDotToken ifFalse: [self syntaxError]].
			isVaryingArgCount := true.
			(token := self peek) isCloseParenthesisToken ifFalse: [self syntaxError].
		] ifFalse: [
			parameters add: (CDeclaration header: header).
			(token := self peek) isCommaToken ifTrue: [self next].
		].
	].
	self next.	"close parenthesis token"
	(parameters size = 1 and: [parameters first isEmptyVoid]) ifTrue: [
		parameters := #().
	].
	self mayHaveCode ifFalse: [^self].
	[
		(token := self peek) ifNil: [^self].
		token isSemicolonToken ifTrue: [^self].
		token isOpenCurlyBracketToken.
	] whileFalse: [
		self next.
	].
	self next.
	self skipToEndOfCodeBlock.
%
category: 'declarator'
set compile_env: 0
method: CDeclaration
readPointerOperator
	"Return a CDeclaration (might not be self)
	>>--*--+----------------+--><
	          '-typeQualifiers-' 
"
	| token |
	(token := self peek) isNil ifTrue: [^nil].
	token isStarToken ifFalse: [^nil].
	pointer := pointer + 1.
	(token := self next; peek) isNil ifTrue: [^self].
	token isIdentifierToken ifFalse: [^self].
	(#('const' 'volatile') includes: token value) ifFalse: [^self].
	self next.
	"we don't keep track of const pointers, only the data"
%
category: 'declarator'
set compile_env: 0
method: CDeclaration
readPostlude

	| token1 token2 |
	(token1 := self peek) isNil ifTrue: [^self].
	token1 isIdentifierToken ifTrue: [
		(2 < token1 value size and: [(token1 value at: 1) == $_ and: [(token1 value at: 2) == $_]]) ifTrue: [
			(token2 := self next; peek) isOpenParenthesisToken ifTrue: [
				token1 value = '__attribute__' ifTrue: [
					self readAttributes.
				] ifFalse: [
					self skipFunction.
				].
			].
			^self readPostlude.
		].
		self halt.
	].
	token1 isColonToken ifTrue: [self next. self readBitFields. ^self readPostlude].
	(token1 isOpenCurlyBracketToken and: [type class == CDeclaration and: [type type == #'class']]) ifTrue: [
		self next.
		self skipToEndOfCodeBlock.
	].
%
category: 'type'
set compile_env: 0
method: CDeclaration
readStorageClassSpecifier
	"Return a CDeclaration (might not be self)
	( 'auto' | 'extern' | 'register' | 'static' | 'typedef' | 'inline' )
"
	| token |
	(token := self peek) isNil ifTrue: [^nil].
	token isTildeToken ifTrue: [token := self next; peek].	"indicates a class destructor function, which we can ignore"
	[token isExtensionIdentifierToken] whileTrue: [token := self next; peek].	"safe to ignore"
	token  isIdentifierToken ifFalse: [^nil].
	token value = 'inline' ifTrue: [
		self next.
		storage := #'inline'.
		^self.
	].
	(#('auto' 'extern' 'register' 'static' 'typedef') includes: token value) ifTrue: [
		storage ifNotNil: [self error: 'We support only one storage specifier'].
		storage := self next value asSymbol.
		^self.
	].
	^nil.
%
category: 'type'
set compile_env: 0
method: CDeclaration
readType
	"Return a CDeclaration (might not be self)
"
	| declaration |
	declaration := self readTypeA.
	storage isNil ifTrue: [storage := #'auto'].
	self interpretTypeSpecifier.
	^self.
%
category: 'type'
set compile_env: 0
method: CDeclaration
readTypeA
	"Return a CDeclaration (might not be self)
	        .-----------------------------.  
	        V                                 |
	>>---+----------------------------+--><
           +--storageClassSpecifier--+			( 'auto' | 'extern' | 'register' | 'static' | 'typedef' | 'inline' )
           +--typeSpecifier-----------+			( 'void' | 'char' | 'short' | 'int' | 'long' | 'float' | 'double' | ... )
           '--typeQualifier-------------'				( 'const' | 'volatile' )
"
	| declaration token |
	token := self peek.	"for debugging"
	(declaration := self readStorageClassSpecifier) ifNotNil: [^declaration readTypeA].
	(declaration := self readTypeSpecifier) ifNotNil: [^declaration readTypeA].
	(declaration := self readTypeQualifier) ifNotNil: [^declaration readTypeA].
	^self.
%
category: 'type'
set compile_env: 0
method: CDeclaration
readTypeFrom: aHeader
	"Return a CDeclaration (might not be self)"

	header := aHeader.
	[self peek isEmptyToken] whileTrue: [self next].
	type := Array new.
	pointer := 0.
	isConstant := false.
	includesCode := false.
	isVaryingArgCount := false.
	^self readType.
%
category: 'type'
set compile_env: 0
method: CDeclaration
readTypeQualifier
 	"Return a CDeclaration (might not be self)
	( 'const' | 'volatile' )
"
	| token value |
	(token := self peek) 	ifNil: [^nil].
	token isIdentifierToken 	ifFalse: [^nil].
	value := token value.
	value = 'const' 	ifTrue: [self next. isConstant := true. ^self].
	value = 'volatile' 	ifTrue: [self next. ^self].
	^nil.
%
category: 'type'
set compile_env: 0
method: CDeclaration
readTypeSpecifier
 	"Return a CDeclaration (might not be self)
	( 'void' | 'char' | 'short' | 'int' | 'long' | 'float' | 'double' | 'signed' | 'unsigned'
		| struct_or_union_spec | enum_spec | typedef_name
"
	| token value declaration |
	(token := self peek) isNil ifTrue: [^nil].
	value := token value.
	(token type = #'string' and: [value = 'C']) ifTrue: [
		(token := self next; peek) isOpenCurlyBracketToken ifFalse: [self error: 'Unexpected token!'].
		name := 'C'.
		^self.
	].
	(token type = #'string' and: [value = 'C++']) ifTrue: [
		(token := self next; peek) isOpenCurlyBracketToken ifFalse: [self error: 'Unexpected token!'].
		name := 'C++'.
		^self.
	].
	token isIdentifierToken ifFalse: [^nil].
	(#('void' 'char' 'short' 'int' 'long' 'float' 'double' 'signed' 'unsigned') includes: value) ifTrue: [
		type add: value asSymbol.
		self next.
		^self.
	].
	value = 'ptr'					ifTrue: [type := #'ptr'.		self next. ^self].
	value = 'enum'					ifTrue: [type := #'enum'.	self next. self readEnumeration. 		^self].
	value = 'struct' 				ifTrue: [type := #'struct'.	self next. self readComplexType. 	^self].
	value = 'union' 				ifTrue: [type := #'union'.	self next. self readComplexType. 	^self].
	value = 'class' 				ifTrue: [type := #'class'.	self next. storage := #'typedef'.		^self].
	type = #() ifTrue: [
		declaration := 
			header types at: value ifAbsent: [
			header structs at: value ifAbsent: [
			header unions at: value ifAbsent: [
			header enums at: value ifAbsent: [nil]]]].
		declaration notNil ifTrue: [type := declaration derivedType. self next. ^self].
	].
	"These are a bit troubling. Sometimes the are expected to be built-in, and other times they are typedefs!"
	(value = '__builtin_va_list' and: [type = #()])	ifTrue: [type := #'ptr'. 		self next. ^self].
	(value = '__gnuc_va_list' and: [type = #()])		ifTrue: [type := #'ptr'. 		self next. ^self].	
	^nil.
%
category: 'CCallout'
set compile_env: 0
method: CDeclaration
resultType

	self isFunction ifFalse: [self error: 'Not a function!'].
	(self isSimple and: [self knownResultTypes includes: type]) ifTrue: [^type].
	type == #'enum' ifTrue: [^#'int32'].
	type == #'uint8' ifTrue: [
		(count size == 1 and: [pointer == 0]) ifTrue: [^#'char*'].		"const char foo[]"
		(count == nil and: [pointer == 1]) ifTrue: [^#'char*'].			"const char * foo"
	].
	^#'ptr'.
%
category: 'other'
set compile_env: 0
method: CDeclaration
setClassProperties: anArray

	fields := anArray.
%
category: 'Updating'
set compile_env: 0
method: CDeclaration
setStorage: aSymbol type: aCDeclaration

	storage := aSymbol.
	type := aCDeclaration.
%
category: 'CByteArray get'
set compile_env: 0
method: CDeclaration
simpleAccessorForOffset: anInteger

	^(WriteStream on: String new)
		nextPutAll: name; lf; lf; tab;
		nextPutAll: '^self ';
		nextPutAll: self getSelector; space;
		nextPutAll: anInteger printString;
		nextPut: $.; lf;
		contents.
%
category: 'CByteArray put'
set compile_env: 0
method: CDeclaration
simpleUpdatorForOffset: anInteger

	^(WriteStream on: String new)
		nextPutAll: name;
		nextPutAll: ': anObject'; lf; lf; tab;
		nextPutAll: 'self '; lf; tab; tab;
		nextPutAll: self getSelector; space;
		nextPutAll: anInteger printString; lf; tab; tab;
		nextPutAll: 'put: anObject.'; lf;
		contents.
%
category: 'CByteArray get'
set compile_env: 0
method: CDeclaration
singlePointerAccessorForOffset: anInteger

	^(WriteStream on: String new)
		nextPutAll: name; lf; lf; tab;
		nextPutAll: '^self'; lf; tab; tab;
		nextPutAll: 'pointerAt: ';
		nextPutAll: anInteger printString; lf; tab; tab;
		nextPutAll: 'resultClass: CByteArray.'; lf;
		contents.
%
category: 'CByteArray put'
set compile_env: 0
method: CDeclaration
singlePointerUpdatorForOffset: anInteger

	^(WriteStream on: String new)
		nextPutAll: name; 
		nextPutAll: ': aCByteArray'; lf; lf; tab;
		nextPutAll: 'self'; lf; tab; tab;
		nextPutAll: 'pointerAt: ';
		nextPutAll: anInteger printString; lf; tab; tab;
		nextPutAll: 'put: aCByteArray.'; lf;
		contents.
%
category: 'other'
set compile_env: 0
method: CDeclaration
skipFunction

	| parenthesisCount |
	parenthesisCount := 1.
	self next.
	[
		0 < parenthesisCount.
	] whileTrue: [
		| token |
		token := self next.
		token isCloseParenthesisToken ifTrue: [parenthesisCount := parenthesisCount - 1] ifFalse: [
		(token isIdentifierToken and: [token value last == $(]) ifTrue: [parenthesisCount := parenthesisCount + 1] ifFalse: [
		token isOpenParenthesisToken ifTrue: [parenthesisCount := parenthesisCount + 1]]].
	].
%
category: 'declarator'
set compile_env: 0
method: CDeclaration
skipToEndOfCodeBlock

	| depth |
	includesCode := true.
	depth := 1.
	[
		0 < depth.
	] whileTrue: [
		| token |
		token := self next.
		token isOpenCurlyBracketToken ifTrue: [depth := depth + 1].
		token isCloseCurlyBracketToken ifTrue: [depth := depth - 1].
	].
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
source

	^source.
%
category: 'Updating'
set compile_env: 0
method: CDeclaration
source: aString

	source ifNotNil: [^self].
	source := aString.
	(type isKindOf: CDeclaration) ifTrue: [type source: aString].
	fields ifNotNil: [
		fields do: [:each | 
			each source ifNil: [
				each source: aString.
			].
		].
	].
	parameters ifNotNil: [
		parameters do: [:each | 
			each source ifNil: [
				each source: aString.
			].
		].
	].
%
category: 'CByteArray get'
set compile_env: 0
method: CDeclaration
sourceStringsForAccessors

	| list offset |
	list := OrderedCollection new.
	offset := 0.
	fields do: [:each | 
		list add: (each accessorForOffset: offset).
		offset := offset + each byteSize.
	].
	^list.
%
category: 'CByteArray put'
set compile_env: 0
method: CDeclaration
sourceStringsForUpdators

	| list offset |
	list := OrderedCollection new.
	offset := 0.
	fields do: [:each | 
		list add: (each updatorForOffset: offset).
		offset := offset + each byteSize.
	].
	^list.
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
storage

	^storage.
%
category: 'other'
set compile_env: 0
method: CDeclaration
stream

	^header stream.
%
category: 'CByteArray get'
set compile_env: 0
method: CDeclaration
stringAccessorForOffset: anInteger

	^(WriteStream on: String new)
		nextPutAll: name; lf; lf; tab;
		nextPutAll: '^self _stringFromBytes: (self '; lf; tab; tab;
		nextPutAll: 'byteArrayFrom: ';
		nextPutAll: anInteger printString; lf; tab; tab;
		nextPutAll: 'to: ';
		nextPutAll: (anInteger + count first - 1) printString;
		nextPutAll: ').'; lf;
		contents.
%
category: 'CByteArray get'
set compile_env: 0
method: CDeclaration
stringPointerAccessorForOffset: anInteger

	1 == count ifTrue: [
		^(WriteStream on: String new)
			nextPutAll: name; lf; lf; tab;
			nextPutAll: '^self stringFromCharStarAt: ';
			nextPutAll: anInteger printString; 
			nextPut: $.; lf; tab;
			nextPutAll: '"^self'; lf; tab; tab;
			nextPutAll: 'pointerAt: ';
			nextPutAll: anInteger printString; lf; tab; tab;
			nextPutAll: 'resultClass: CByteArray."'; lf;
			contents.
	] ifFalse: [
		^(WriteStream on: String new)
			nextPutAll: name; lf; lf; tab;
			nextPutAll: '| array offset |'; lf; tab;
			nextPutAll: 'array := Array new: ';
			nextPutAll: count first printString;
			nextPut: $.; lf; tab;
			nextPutAll: 'offset := ';
			nextPutAll: anInteger printString;
			nextPut: $.; lf; tab;
			nextPutAll: '1 to: array size do: [:i | '; lf; tab; tab;
			nextPutAll: 'array at: i put: (self stringFromCharStarAt: offset).'; lf; tab; tab;
			nextPutAll: '"array at: i put: (self pointerAt: offset resultClass: CByteArray)."'; lf; tab; tab;
			nextPutAll: 'offset := offset + 8.'; lf; tab;
			nextPutAll: '].'; lf; tab;
			nextPutAll: '^array.'; lf;
			contents.
	].
%
category: 'CByteArray put'
set compile_env: 0
method: CDeclaration
stringPointerUpdatorForOffset: anInteger

	1 == count ifTrue: [
		^(WriteStream on: String new)
			nextPutAll: name;
			nextPutAll: ': aCByteArray'; lf; lf; tab;
			nextPutAll: '"self'; lf; tab; tab;
			nextPutAll: 'pointerAt: ';
			nextPutAll: anInteger printString; lf; tab; tab;
			nextPutAll: 'put: aCByteArray."'; lf;
			contents.
	] ifFalse: [
		^self arrayPointersUpdatorForOffset: anInteger.
	].
%
category: 'CByteArray put'
set compile_env: 0
method: CDeclaration
stringUpdatorForOffset: anInteger

	^(WriteStream on: String new)
		nextPutAll: name;
		nextPutAll: ': aByteArray'; lf; lf; tab;
		nextPutAll: 'self'; lf; tab; tab; 
		nextPutAll: 'replaceFrom: ';
		nextPutAll: anInteger printString; lf; tab; tab;
		nextPutAll: 'to: ';
		nextPutAll: (anInteger - 1) printString;
		nextPutAll: ' + (';
		nextPutAll: count first printString;
		nextPutAll: ' min: aByteArray size)'; lf; tab; tab; 
		nextPutAll: 'with: aByteArray'; lf; tab; tab;
		nextPutAll: 'startingAt: 1.'; lf;
		contents.
%
category: 'CByteArray get'
set compile_env: 0
method: CDeclaration
structAccessorForOffset: anInteger

	^(WriteStream on: String new)
		nextPutAll: name; lf; lf; tab;
		nextPutAll: '^self'; lf; tab; tab;
		nextPutAll: 'newFrom: ';
		nextPutAll: anInteger printString; lf; tab; tab;
		nextPutAll: 'numBytes: ';
		nextPutAll: self baseByteSize printString;
		nextPut: $.; lf;
		contents.
%
category: 'CByteArray put'
set compile_env: 0
method: CDeclaration
structUpdatorForOffset: anInteger

	^(WriteStream on: String new)
		nextPutAll: name;
		nextPutAll: ': aByteArray'; lf; lf; tab;
		nextPutAll: 'self'; lf; tab; tab; 
		nextPutAll: 'replaceFrom: ';
		nextPutAll: anInteger printString; lf; tab; tab;
		nextPutAll: 'to: ';
		nextPutAll: (anInteger - 1) printString;
		nextPutAll: ' + (';
		nextPutAll: self baseByteSize printString;
		nextPutAll: ' min: aByteArray size)'; lf; tab; tab; 
		nextPutAll: 'with: aByteArray'; lf; tab; tab;
		nextPutAll: 'startingAt: 1.'; lf;
		contents.
%
category: 'other'
set compile_env: 0
method: CDeclaration
syntaxError

	self error: 'Syntax error!'.
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
type
	"For many types, the value is a Symbol (e.g., #'uint64' or #'struct').
	For other types, the value is a CDeclaration for a function or a class"

	^type.
%
category: 'CByteArray put'
set compile_env: 0
method: CDeclaration
updatorForOffset: anInteger

	0 == pointer ifTrue: [
		type == #'class' ifTrue: [^self structUpdatorForOffset: anInteger].
		(type isKindOf: CDeclaration) ifTrue: [^self structUpdatorForOffset: anInteger].
		count == nil ifTrue: [^self simpleUpdatorForOffset: anInteger].
		1 < count size ifTrue: [self error: 'Multi-dimensional arrays not supported!'].
		type == #'uint8' ifTrue: [^self stringUpdatorForOffset: anInteger].
		^self arrayUpdatorForOffset: anInteger.
	].
	1 < pointer ifTrue: [self error: 'Multi-level pointers not supported!'].
	count == nil ifTrue: [^self singlePointerUpdatorForOffset: anInteger].
	type == #'uint8' ifTrue: [^self stringPointerUpdatorForOffset: anInteger].
	^self arrayPointersUpdatorForOffset: anInteger.
%
category: 'CCallout'
set compile_env: 0
method: CDeclaration
varArgsAfter

	^ isVaryingArgCount ifTrue: [parameters size] ifFalse: [-1].
%
category: 'accessors'
set compile_env: 0
method: CDeclaration
_fields

	^fields.
%
category: 'other'
set compile_env: 0
method: CDeclaration
_next

	| next |
	(next := header next) ifNil: [^nil].
	file ifNil: [
		file := next file.
		line := next line.
	].
	^next.
%
