!=========================================================================
! Copyright (C) VMware, Inc. 1986-2012.  All Rights Reserved.
!
! $Id: GsObjectSecurityPolicy.gs 29610 2012-11-20 05:02:00Z otisa $
!
! Superclass Hierarchy:
!   GsObjectSecurityPolicy, Object.
!
!=========================================================================

removeallmethods GsObjectSecurityPolicy
removeallclassmethods GsObjectSecurityPolicy

run
	"upgrade to Gs64 v2.2 should have renamed instvar spare1 to itsId"
| nam |
nam := (GsObjectSecurityPolicy _instVarNames) at: 7 .
nam == #itsId ifFalse:[ nil error: 'bad ivName ' , nam asString ].
^ true
%

category: 'For Documentation Installation only'
classmethod: GsObjectSecurityPolicy
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'Each Repository is composed of an integral number of ObjectSecurityPolicies.  
 A GsObjectSecurityPolicy has the following properties:

 * Ownership.  This is the smallest unit of ownership for accounting and
   authorization purposes.  Each GsObjectSecurityPolicy is owned by one and 
   only one user.

 * Authorization unit.  A user may control access to objects by placing them in
   a GsObjectSecurityPolicy with a known authorization.  Prior to reading or 
   writing an object in a GsObjectSecurityPolicy, users must be authorized to 
   perform the desired action.  
   However, during a transaction, once a user is authorized to read one object
   in a GsObjectSecurityPolicy, that user can read any object in the same 
   GsObjectSecurityPolicy.

 ObjectSecurityPolicies and Repositories are discussed in more detail in the 
 GemStone Programming Guide.

Text added as of Gs64 v2.2:
  All objects have a objectSecurityPolicyId attribute, which is stored in the 
  object header.  
  References from an object to its GsObjectSecurityPolicy are indirect via the
  SystemRepository object.  
  (See implementation of method Object >> objectSecurityPolicy ) 
  In a Repository that was converted from Gs64 v1.x , v2.0 or v2.1 ,
  all objects created prior to conversion will have objectSecurityPolicyId 0 .

  Read authorization is checked when an object is faulted or refreshed
  into memory , once in memory no further read authorization checks
  are performed. No checks are done for an object with objectSecurityPolicyId 0.

  Write authorization is checked the first time an object is written
  during each transaction.  No checks are done for an object
  with objectSecurityPolicyId 0 .

  Changes to GsObjectSecurityPolicy, UserProfile , or Group objects that affect
  read or write authorization checks are only guaranteed to be noticed
  after the next login.  This is different from Gemston/S 6.x , in which
  authorization changes are supposed to be noticed at the
  next transaction boundary.  Changes to the objectSecurityPolicyId of an 
  object will take effect on the next transaction boundary.

  Creation of ObjectSecurityPolicies and changes to attributes of 
  ObjectSecurityPolicies require write authorization permission to the 
  DataCuratorObjectSecurityPolicy .

  If you are logged in as SystemUser , no authorization errors will occur .

  All instances of Symbol, DoubleByteSymbol must be in a objectSecurityPolicy 
  with World read permission. By default they are on 
  DataCuratorObjectSecurityPolicy .
 
  For logins other than SystemUser, the virtual machine requires read access 
  to both DataCuratorObjectSecurityPolicy and SystemObjectSecurityPolicy 
  in order to be able to bootstrap object memory
  as part of a session login.  Specifically , well known classes
  such as Class, Object, etc are expected to be in SystemObjectSecurityPolicy ,
  and instances of Symbol and objects such as SystemRepository are
  expected to be in DataCuratorObjectSecurityPolicy.  
  To avoid infinite recursion in loading the authorization cache, 
  the virtual machine preloads authorization for SystemObjectSecurityPolicy 
  and DataCuratorObjectSecurityPolicy after validating the userId/password,
  and will generate a fatal error if it does not have read access to both 
  DataCuratorObjectSecurityPolicy and DataCuratorObjectSecurityPolicy.

  All instances of UserProfile,  IdentitySets representing Groups, 
  and ObjectSecurityPolicies must be in a objectSecurityPolicy with World 
  read permission.  By default they are on DataCuratorObjectSecurityPolicy.

  During repository conversion from a prior release to Gs64 v2.2 ,
  if the previous size of SystemRepository was < 20 , and
  a  GsObjectSecurityPolicy will be created with objectSecurityPolicyId 20 
  and world write, to handle the case where an older version of GBS clients
  may have created objects in a Gs64 v2.0 or v2.1 repository
  with objectSecurityPolicyId 20 .
'.
doc documentClassWith: txt.

txt := (GsDocText new) details:
'The Repository containing the GsObjectSecurityPolicy.'.
doc documentInstVar: #itsRepository with: txt.

txt := (GsDocText new) details:
'A UserProfile indicating the owner of the GsObjectSecurityPolicy.'.
doc documentInstVar: #itsOwner with: txt.

txt := (GsDocText new) details:
'An IdentitySet of canonical Strings.  Each String must be an element of
 AllGroups, and represents a group of users who are authorized to access
 the GsObjectSecurityPolicy for reading.'.
doc documentInstVar: #groupsRead with: txt.

txt := (GsDocText new) details:
'An IdentitySet of canonical Strings.  Each String must be an element of
 AllGroups, and represents a group of users who are authorized to access
 the GsObjectSecurityPolicy for writing.'.
doc documentInstVar: #groupsWrite with: txt.

txt := (GsDocText new) details:
'A SmallInteger specifying authorization for the owner to access the GsObjectSecurityPolicy.
 0 = no access, 1 = read access, 2 = write access.'.
doc documentInstVar: #ownerAuthorization with: txt.

txt := (GsDocText new) details:
'A SmallInteger specifying world authorization to access the GsObjectSecurityPolicy.
 0 = no access, 1 = read access, 2 = write access.'.
doc documentInstVar: #worldAuthorization with: txt.

txt := (GsDocText new) details:
'A SmallInteger > 0, 
 the offset of this GsObjectSecurityPolicy in the object SystemRepository. '.
doc documentInstVar: #objectSecurityPolicyId with: txt.

txt := (GsDocText new) details:
'Methods in this category are obsolete and are provided only for compatibility
 with earlier releases of GemStone.  They will be removed in a future release.'.
doc documentCategory: #'Deprecated' with: txt.

self comment: doc asString.
%

category: 'Instance Creation'
classmethod: GsObjectSecurityPolicy
newInRepository: aRepository

"Returns a new GsObjectSecurityPolicy in the Repository aRepository.  If the
 maximum number (65534) of ObjectSecurityPolicies has already been created for 
 aRepository, this generates an error.
 The new GsObjectSecurityPolicy is owned by the UserProfile of the current session, 
 and has the default authorization of W----- (owner can read and write).

 After execution of this method, the current transaction must be committed
 before objects can be created in the GsObjectSecurityPolicy, object assigned to the
 objectSecurityPolicy, or the GsObjectSecurityPolicy used as argument to a 
 defaultObjectSecurityPolicy: keyword in creation of UserProfiles.

 Requires the ObjectSecurityPolicyCreation privilege and WriteAccess to 
 DataCuratorObjectSecurityPolicy.
 "

<primitive: 291>
aRepository _validateClass: Repository.
self _primitiveFailed: #newInRepository: args: { aRepository } .
self _uncontinuableError
%

category: 'Accessing'
method: GsObjectSecurityPolicy
groups

"Returns an IdentitySet, the set of user groups (Strings) that are
 explicitly authorized to read or write in this GsObjectSecurityPolicy."

| result |
groupsRead ~~ nil ifTrue:[ result := groupsRead ] 
                 ifFalse:[ result := IdentitySet new ].
groupsWrite ~~ nil ifTrue:[ result := result + groupsWrite ].
^ result
%

category: 'Accessing'
method: GsObjectSecurityPolicy
number

"Returns the index of the receiver in its Repository"

^ itsId
%

category: 'Deprecated'
method: GsObjectSecurityPolicy
segmentId

self deprecated: 'GsObjectSecurityPolicy>>segmentId is deprecated v3.0 and provided for backwards compatibility.
Use #objectSecurityPolicyId instead.'.

^self objectSecurityPolicyId
%

category: 'Accessing'
method: GsObjectSecurityPolicy
objectSecurityPolicyId

"Returns the index of the receiver in its Repository"

^ itsId
%

category: 'Accessing'
method: GsObjectSecurityPolicy
objectSecurityPolicyId

"Returns the index of the receiver in its Repository"

^ itsId
%

category: 'Accessing'
method: GsObjectSecurityPolicy
owner

"Returns the UserProfile of the receiver's owner."

^ itsOwner
%

category: 'Accessing'
method: GsObjectSecurityPolicy
repository

"Returns the Repository containing the receiver."

^ itsRepository
%


category: 'Updating'
method: GsObjectSecurityPolicy
owner: aUserProfile

"Redefines the receiver's owner to be the user represented by aUserProfile.
 To execute this method, you must be authorized to write in the receiver's
 GsObjectSecurityPolicy (customarily, the DataCurator GsObjectSecurityPolicy)."

itsOwner := aUserProfile
%

category: 'Formatting'
method: GsObjectSecurityPolicy
asString

"Returns a formatted String that contains the following information:

 The name of the Repository containing the receiver and the index of
 the receiver in that Repository.

 * The user ID of the receiver's owner.

 * The Symbol that defines each user group which is authorized to read or
   write in the receiver.

 * Information about whether the receiver's owner, each group, or all
   users are authorized to read or write in the receiver.

 For example, the message 'System myUserProfile defaultObjectSecurityPolicy asString'
 returns a String that resembles the following:

 anObjectSecurityPolicy, Number 1 in Repository SystemRepository, Owner SystemUser write, World read

 If the receiver is not a GsObjectSecurityPolicy in the SystemRepository, the String contains
 'NOT IN REPOSITORY' in place of the index."

| tempString theGroupAuthString |

tempString := String new.  " string to return "
theGroupAuthString := String new.  " string of group authorizations"

"groups and build a single string to describe group authorizations."
groupsRead size ~~ 0 ifTrue:[
  (groupsRead sortAscending:  '' ) do: [:groupName | 
     theGroupAuthString addAll: ' Group ';
                 addAll: (groupName asString);
                 add: ' ';
                 addAll: 'read' ; add: $, .
  ].
].
groupsWrite size ~~ 0 ifTrue:[
  (groupsWrite sortAscending:  '' ) do: [:groupName | 
     theGroupAuthString addAll: ' Group ';
        addAll: (groupName asString);
        add: ' ';
        addAll: 'write' ; add: $, .
  ].
].

tempString " Build the string to return"
             addAll: 'anObjectSecurityPolicy, Number ';
             addAll: self objectSecurityPolicyId  asString ;
             addAll: ' in Repository ';
             addAll: (self repository name); add: $, ;
             addAll: ' Owner ';
             addAll: (self owner userId);   add: $   ; 
             addAll: (self ownerAuthorization); add: $, ;
             addAll: theGroupAuthString;
             addAll: ' World ';
             addAll: (self worldAuthorization).

^tempString
%

category: 'Clustering'
method: GsObjectSecurityPolicy
clusterDepthFirst

"This method clusters the receiver.  (Overrides the inherited method, which
 also clusters all instance variables.)  Returns true if the receiver has 
 already been clustered during the current transaction, false otherwise."

  self cluster ifTrue:[ ^ true ].
  groupsRead cluster .
  groupsWrite cluster .
^ false
%

category: 'Copying'
method: GsObjectSecurityPolicy
copy

"Returns the receiver. Copies of objectSecurityPolicies are not made."

^ self.
%

category: 'Copying'
method: GsObjectSecurityPolicy
_deepCopyWith: copiedObjDict

"Private. Used internally to implement deepCopy."

^ self.
%

category: 'Accessing Authorization'
method: GsObjectSecurityPolicy
authorizationForGroup: aGroupString

"Returns a Symbol that defines whether the specified group is authorized to
 write (and read) in this GsObjectSecurityPolicy (#write), to read only (#read), or neither
 (#none)."

| canonGroup |
canonGroup := AllGroups at: aGroupString otherwise: nil .
canonGroup == nil ifTrue:[ ^ #none ].

groupsWrite size ~~ 0 ifTrue:[
  (groupsWrite includesIdentical: canonGroup) ifTrue:[ ^ #write ]
  ].
groupsRead size ~~ 0 ifTrue:[ 
  (groupsRead includesIdentical: canonGroup) ifTrue:[ ^ #read ]
  ].
^ #none
%

category: 'Accessing Authorization'
method: GsObjectSecurityPolicy
authorizationForUser: aUserProfile

"Returns a Symbol that describes the authorization that the given user has for
 the receiver."

| userGroups curAuth |

"Consider world authorization"
curAuth := worldAuthorization.
curAuth == 2 ifTrue: [ ^ #write ].

"Owner authorization applicable?"
itsOwner == aUserProfile ifTrue: [
  ownerAuthorization == 2 ifTrue:[ ^ #write ].
  curAuth := curAuth bitOr: ownerAuthorization .
  ].

"Consider group authorizations..."
userGroups := aUserProfile groups .
userGroups size ~~ 0 ifTrue:[
  groupsWrite size ~~ 0 ifTrue:[
    (groupsWrite * userGroups) size ~~ 0 ifTrue:[ ^ #write ].
    ].
  groupsRead size ~~ 0 ifTrue:[
    (groupsRead * userGroups) size ~~ 0 ifTrue:[ curAuth := curAuth bitOr: 1 ].
    ].
  ].
   
curAuth == 1 ifTrue:[ ^ #read ].
^#none
%

category: 'Accessing Authorization'
method: GsObjectSecurityPolicy
ownerAuthorization

"Returns a Symbol that defines whether the GsObjectSecurityPolicy's owner is authorized to
 write (and read) in this GsObjectSecurityPolicy (#write), to read only (#read), or neither
 (#none)."

^ #( #none #read #write) at: ownerAuthorization + 1
%

category: 'Accessing Authorization'
method: GsObjectSecurityPolicy
worldAuthorization

"Returns a Symbol that defines whether all system users are authorized to write
 (and read) in this GsObjectSecurityPolicy (#write), to read only (#read), or neither
 (#none)."

^ #( #none #read #write) at: worldAuthorization + 1
%

category: 'Accessing Authorization'
method: GsObjectSecurityPolicy
groupsWithAuthorization: anAccessSymbol

"Returns an IdentitySet of group Strings of all groups with the authorization of
 anAccessSymbol (one of #write, #read, or #none) for the receiver."

| result |

( #( #write #read #none) includesValue: anAccessSymbol)
  ifFalse: [  ^ self _error: #segErrBadAuthKind args: { anAccessSymbol } ].

result := IdentitySet new .
AllGroups valuesDo:[ :aGroup |
  (self authorizationForGroup: aGroup) == anAccessSymbol ifTrue:[
    result add: aGroup
    ].
  ].
^ result
%

! fixed 36635
category: 'Accessing Authorization'
method: GsObjectSecurityPolicy
usersWithAuthorization: anAccessSymbol

"Returns a IdentitySet containing all users with the authorization of
 anAccessSymbol (one of #write, #read, or #none) for the receiver."

| result |

( #( #write #read #none) includesValue: anAccessSymbol)
  ifFalse: [  ^ self _error: #segErrBadAuthKind args: { anAccessSymbol } ].
   
result := IdentitySet new .
AllUsers do:[:aUserProfile |
  (self authorizationForUser: aUserProfile) == anAccessSymbol ifTrue:[
     result add: aUserProfile
     ].
  ].
^ result
%

category: 'Updating Authorization'
method: GsObjectSecurityPolicy
ownerAuthorization: anAuthorizationSymbol

"Redefines the authorization for the GsObjectSecurityPolicy's owner to one of the following
 authorization Symbols:

 * #write (the GsObjectSecurityPolicy's owner can both read and write in this GsObjectSecurityPolicy).

 * #read (read only).

 * #none (neither read nor write permission).

 Generates an error if anAuthorizationSymbol is not one of (#read, #write,
 #none).

 Requires the ObjectSecurityPolicyProtection privilege, if the objectSecurityPolicy is not owned
 by the GemStone UserProfile under which this session is running."

 | authInt |
 authInt := #( #none #read #write ) indexOf: anAuthorizationSymbol .
 authInt == 0 ifTrue:[ 
   ^ self _error: #segErrBadAuthKind args: { anAuthorizationSymbol } 
   ].
 self _authAt: 5 put: (authInt - 1) 
%

category: 'Updating Authorization'
method: GsObjectSecurityPolicy
worldAuthorization: anAuthorizationSymbol

"Redefines the authorization for all users to one of the following
 authorization Symbols:

 * #write (all users can both read and write in this GsObjectSecurityPolicy).

 * #read (read only).

 * #none (neither read nor write permission).

 Generates an error if anAuthorizationSymbol is not one of (#read, #write,
 #none).

 Requires the ObjectSecurityPolicyProtection privilege, if the objectSecurityPolicy is not owned
 by the GemStone UserProfile under which this session is running."

 | authInt |
 authInt := #( #none #read #write ) indexOf: anAuthorizationSymbol .
 authInt == 0 ifTrue:[ 
   ^ self _error: #segErrBadAuthKind args: { anAuthorizationSymbol } 
   ].
 self _authAt: 6 put: (authInt - 1) 
%

!  _authorizationChanged not used in Gs64 v2.2

category: 'Private'
method: GsObjectSecurityPolicy
_authAt: anOffset put: aValue

"offset   action 
 ------   ---------
  5       assign to ownerAuthorization instance variable
  6       assign to worldAuthorization instance variable
  3       add to groupsRead IdentitySet, create set if needed.
  4       add to groupsWrite IdentitySet, create set if needed.
 -3       remove from groupsRead IdentitySet
 -4       remove from groupsWrite IdentitySet
  7       assign new value to groupsRead (for repository conversion)
  8       assign new value to groupsWrite (for repository conversion)

  Gs64 v2.2 ,  offsets 7 and 8 not supported, was used by _migrateGroups.
"

<primitive: 292>
self _primitiveFailed: #_authAt:put: args: { anOffset . aValue }.
self _uncontinuableError
%

category: 'Execution'
classmethod: GsObjectSecurityPolicy
setCurrent: anObjectSecurityPolicy while: aBlock

| oldPolicy |

oldPolicy := System currentObjectSecurityPolicy.
oldPolicy == anObjectSecurityPolicy ifTrue: [ "no need to change/restore policy "
  ^ aBlock value 
].
System currentObjectSecurityPolicy: anObjectSecurityPolicy .
^ [ 
    [
      aBlock value 
    ] onException: Error do:[:ex |
      (oldPolicy == nil or:[ oldPolicy isKindOf: GsObjectSecurityPolicy]) ifTrue: [
        System currentObjectSecurityPolicy: oldPolicy .
      ].
      ex outer
    ]
  ] ensure:[
    System currentObjectSecurityPolicy: oldPolicy
  ].
%

category: 'Execution'
method: GsObjectSecurityPolicy
setCurrentWhile: aBlock

"Sets the receiver to be the current security policy while the given block
 executes.  Catches all errors and reinstalls the current security policy
 to avoid having the receiver be left as the current security policy.  Returns
 the result of evaluating the zero-argument block."

^ self class setCurrent: self while: aBlock
%

category: 'Deprecated'
method: GsObjectSecurityPolicy
groupNo: groupIndex group: aGroupString authorization: anAuthorizationSymbol

self deprecated: 'Obsolete in GemStone/64.  Use the #group:authorization: method instead.'.
^ self group: aGroupString authorization: anAuthorizationSymbol
%

category: 'Updating Authorization'
method: GsObjectSecurityPolicy
group: aGroupString authorization: anAuthorizationSymbol

"Redefines the authorization for the specified group to one of the following
 authorization Symbols:

 * #write (members of the group can both read and write in this GsObjectSecurityPolicy).
 * #read (read only).
 * #none (neither read nor write permission).

 This method generates an error if aGroupString is not an element of
 the global object AllGroups, or if anAuthorizationSymbol is not one of (#read,
 #write, #none).

 Requires the ObjectSecurityPolicyProtection privilege, if the objectSecurityPolicy is not owned
 by the GemStone UserProfile under which this session is running."

| theGroup |

"Test for errInvalidAuthorization "
( #( #write #read #none) includesValue: anAuthorizationSymbol)
ifFalse: [  ^ self _error: #segErrBadAuthKind args: { anAuthorizationSymbol } ].

"Get canonical group string from AllGroups."
theGroup := AllGroups _validateGroupString: aGroupString .
  
"modify the group collections for this objectSecurityPolicy as appropriate."
anAuthorizationSymbol == #write ifTrue:[ 
  self _authAt: -3 put: theGroup "readGroups remove: theGroup ifAbsent:[]" .
  self _authAt: 4 put: theGroup  "writeGroups _addSymbol: theGroup " .
  ]
ifFalse:[
  anAuthorizationSymbol == #read ifTrue:[
    self _authAt: -4 put: theGroup "writeGroups remove: theGroup ifAbsent:[]" .
    self _authAt: 3 put: theGroup  "readGroups _addSymbol: theGroup " .
    ]
  ifFalse:[ "none" 
    self _authAt: -3 put: theGroup "readGroups remove: theGroup ifAbsent:[]" .
    self _authAt: -4 put: theGroup "writeGroups remove: theGroup ifAbsent:[]" .
    ].
  ].
%

! nameOfGroupAt: deleted 

category: 'Private'
method: GsObjectSecurityPolicy
_currentUserAccess

"returns current user's access to objects in the receiver  ,
 0 == none,  1 == read , 2 == write
"
<primitive: 591>
self _primitiveFailed: #_currentUserAccess.
self _uncontinuableError
%


category: 'Accessing Authorization'
method: GsObjectSecurityPolicy
currentUserCanRead

"Returns true if the current user has read authorization for the
 receiver, false if not."

^ self _currentUserAccess >= 1
%

category: 'Accessing Authorization'
method: GsObjectSecurityPolicy
currentUserCanWrite

"Returns true if the current user has write authorization for the
 receiver, false if not."

^ self _currentUserAccess == 2
%

category: 'Storing and Loading'
method: GsObjectSecurityPolicy
writeTo: aPassiveObject

"Instances of GsObjectSecurityPolicy cannot be converted to passive form.  This method writes
 nil to aPassiveObject and stops GemStone Smalltalk execution with a notifier."

  aPassiveObject writeObject: nil.
  self _error: #rtErrAttemptToPassivateInvalidObject.
%

category: 'Repository Conversion'
method: GsObjectSecurityPolicy
_migrateGroups

"Converts the groups collections from SymbolSets to IdentitySets of Strings."

(Globals at:#DataCuratorObjectSecurityPolicy) setCurrentWhile:[
  groupsRead ~~ nil ifTrue:[
    self _authAt:7 put: (AllGroups _migrateSymbolSet: groupsRead ) .
  ].
  groupsWrite ~~ nil ifTrue:[
    self _authAt:8 put: (AllGroups _migrateSymbolSet: groupsWrite) .
  ].
].
%


! --------------- methods disallowed  are in GsObjectSecurityPolicy2.gs


! _oldAuthorizationFor: deleted
! _migrate40to41  deleted

