!=========================================================================
! Copyright (C) VMware, Inc. 2008-2012.  All Rights Reserved.
!
! $Id$
!
! File: Metaclass3.gs
!   
!=========================================================================

! Metaclass3 created by bom.c

! should not need SELF_CAN_BE_SPECIAL, self should always be non-nil
!  in instances of Metaclass3

set class Metaclass3
removeallmethods 
removeallclassmethods 

category: 'Documentation'
classmethod: 
comment
  ^ '
In Smalltalk beginning with Gs64 v3.0, metaclasses of newly created Classes 
will be instances of Metaclass3 rather than instances of ObsoleteMetaclass .

In Ruby kinds of Metaclass3 are used for classes, metaclasses, and singleton classes

InstVars
  destClass   in a metaclass,  is the "thisClass" reference.
              in a Smalltalk Class, is the "migrationDestination"
' 
%

! --------------------
category: 'Browser Methods'
method:
definition
	"We are created as a side-effect of creating the class!"
	
	^self thisClass definition.
%
category: 'Instance Creation'
method: 
new
  "Disallowed.  To create a new class or metaclass, use
   Class | subclass:instVarNames:... instead."
  self shouldNotImplement: #new
%
method: 
new: aSize
  "Disallowed.  To create a new class or metaclass, use
   Class | subclass:instVarNames:... instead."
  self shouldNotImplement: #new:
%

category: 'Accessing'
method: 
isMeta

"Returns true if the receiver is a meta class "

 ^ (format bitAnd: GC_Metaclass3_metaClass) ~~ 0 
%

method: 
thisClass

"If receiver is a meta class, returns the class, else returns self."

(format bitAnd: GC_Metaclass3_metaClass) ~~ 0 ifTrue:["a meta class"  ^ destClass ] .
^ self 
%

method:
theNonMetaClass
  "used by Monticello"

  "inline  self thisClass"
  (format bitAnd: GC_Metaclass3_metaClass) ~~ 0 ifTrue:["a meta class"  ^ destClass ] .
  ^ self
%
  
method: 
name
 self isMeta ifTrue:[ ^ destClass name , ' class' ] 
           ifFalse:[ ^ super name ]
%
method:
_name
 | cls |
 self isMeta ifFalse:[ ^ super _name ].
 (cls := destClass) ifNotNil:[ ^ cls _name , ' class' ].
 (cls := primaryCopy) ifNotNil:[ ^ 'virtual_', cls _name  ].
 ^ 'unknownMetaclass3'
%
method:
classHistory
  self isMeta ifTrue:[ ^ destClass classHistory ].
  ^ super classHistory
%

method:
extraDict
  self isMeta ifTrue:[ ^ destClass extraDict ].
  ^ nil
%

method:
_classCategory
  ^ self thisClass _classCategory
%

method:
instanceString
  self isMeta ifTrue:[ ^ destClass name ] ifFalse:[ ^ name ]
%
method:
instanceSymbol
  self isMeta ifTrue:[ ^ destClass name ] ifFalse:[ ^ name ]
%

method:
methodStampDictName
  self isMeta ifTrue:[ ^ #GSMetaMethodStampDict ].
  ^ super methodStampDictName.
%
method:
pragmaDictName

  self isMeta ifTrue:[ ^ #GSMetaMethodPragmaDict ].
  ^ super pragmaDictName
%
method:
_gbsTraversalCallback

"Private.  When GemBuilder Smalltalk traverses a metaclass, this method
 is called to return a description of the metaclass."

 self isMeta ifTrue:[ ^ self printString ].
 ^ super _gbsTraversalCallback  "probably MNU"
%
method:
_subclasses

"Returns the class variable that is the list of subclasses, or
 nil if the receiver does not keep track of subclasses."

| subclss result |

self isMeta ifTrue:[
  subclss := self thisClass _subclasses .
  subclss == nil ifTrue:[ ^ nil ].
  result := IdentitySet new .
  subclss do:[ :aClass | result add: aClass class ].
  ^ result
].
^ super _subclasses
%

category: 'Updating the Method Dictionary'
method: 
compileAccessingMethodsFor: anArrayOfSymbols

"Reimplemented to treat class instance variables specially,
 if isMeta is true for receiver . 
 The new methods have environmentId == 0 .  "

| allVarNames varName symlst |
self isMeta ifFalse:[
  ^ super compileAccessingMethodsFor: anArrayOfSymbols
].
symlst := SymbolList new.
varName := 'newValue'.
allVarNames := self allInstVarNames.
[allVarNames includesValue: varName] whileTrue: [
  varName := 'z' , varName.
].
anArrayOfSymbols do: [ :var | | methodtext lf |
  lf := Character lf .
  (methodtext := String new) add: var; add: lf ; add: lf ;
     add: '   "Return the value of the instance variable '''; add: var ;
     add: '''."' ; add: lf ; add: '   ^' ; add: var ; add: lf .
  [ self compileMethod: methodtext dictionaries: symlst
       category: #Accessing environmentId: 0
  ] onException: CompileError do:[:ex |
     self _error: #classErrNotAVar args: { var }
  ].
  (allVarNames indexOf: var) > Metaclass3 instSize ifTrue:[
    "compile a method that lets the variable be modified if the user has
     the proper authority"
    (methodtext := String new )
      add: var; add: ': '; add: varName; add: lf ;
      add: '  "changes the value of the receiver''s class instance variable ''';
      add: var; addAll: '''"'; add: lf ;
      add: lf ;
      add: '  self atClassInstVar: #'; add: var;
      add: ' put: '; add: varName; add: lf .
    [ self compileMethod: methodtext dictionaries: symlst
           category: #Updating environmentId: 0
    ] onException: CompileError do:[:ex | 
      self _error: #classErrNotAVar args: { var }
    ].
  ] ifFalse: [
    (methodtext := String new ) add: var; add: ': ' ; add: varName ; add: lf ; add: lf;
      add: '   "Modify the value of the instance variable ''' ; add: var ;
        add: '''."' ; add: lf ;
      add: '   ' ; add: var ; add: ' := ' ; add: varName ; add: lf .
    [ self compileMethod: methodtext dictionaries: symlst
             category: #Updating environmentId: 0
    ] onException: CompileError do:[:ex | 
        self _error: #classErrNotAVar args: { var }
    ].
  ].
].
%

category: 'Modifying Classes'
method: 
_setClassVars: aDict

  classVars ifNil:[ 
    classVars := aDict .
    self thisClass _setClassVars: aDict.
  ] ifNotNil:[ aDict ~~ classVars 
            ifTrue:[ self error:'invalid store to classVars']
  ]
%
! fixed 41174
method: 
_validateNewClassName: aName
  "Return true if aName is a legal class name, otherwise signal an error.
   If execution continued from the error, return false."

(aName _isSymbol) ifTrue:[
  (aName _validIdentifier: 1024"GCI_MAX_SYMBOL_SIZE") ifTrue:[ ^ true ].
  ( Exception _new: 2149 args: { aName } ) signal .
  ^ false
] ifFalse:[
  "primitive will generate not-a-symbol error"
  ^ false
].
%

category: 'Private'
method: 
_subclass: nameSym
instVarNames: anArrayOfStrings
format: anInteger
constraints: aConstraint
classVars: aSymbolDictionary
poolDictionaries: anArrayOfPoolDicts

self _validateNewClassName: nameSym .
^ self _primSubclass: nameSym
  instVarNames: anArrayOfStrings
  format: anInteger
  constraints: aConstraint
  classVars: aSymbolDictionary
  poolDictionaries: anArrayOfPoolDicts
%
method: 
_primSubclass: nameSym
instVarNames: anArrayOfStrings
format: anInteger
constraints: aConstraint
classVars: aSymbolDictionary
poolDictionaries: anArrayOfPoolDicts

"subclass creation primitive.

 if aSymbolDictionary is empty or nil , the resulting class will have
  classVars == nil .
 if anArrayOfPoolDicts is empty or nil , the resulting class will have
   poolDictionaries == nil 

 Gemstone64, this primitive enforces the restriction that
 a class cannot be more than 450 classes below Object when
 following the superClass chain upwards ."

<primitive: 233>
nameSym _validateClass: Symbol .
anInteger _validateClass: SmallInteger.
anArrayOfStrings _validateClass: Array.
((self == IdentityBag) or: [self inheritsFrom: IdentityBag])
   ifTrue: [(aConstraint isKindOf: Class)
      ifFalse: [ ^ aConstraint _error: #classErrConstraintNotClass]]
   ifFalse: [(aConstraint _isArray)
      ifFalse: [ ^ aConstraint _error: #classErrConstraintNotClass]].
aSymbolDictionary ifNotNil:[ aSymbolDictionary _validateClass: SymbolDictionary].
anArrayOfPoolDicts ifNotNil:[ anArrayOfPoolDicts _validateClass: Array ].

^ self _primitiveFailed:
   #_primSubclass:instVarNames:format:constraints:classVars:poolDictionaries:
   args: { nameSym . anArrayOfStrings . anInteger . 
           aConstraint . aSymbolDictionary . anArrayOfPoolDicts }
%

category: 'Class Instance Variables'
method: 
addInstVarNames: instVarNamesArray

"Adds the instance variables specified in the argument to the receiver and any
 of its subclasses.  Generates an error upon encountering a name that is not a
 valid instance variable name or that is already an instance variable of the
 receiver.

 Instance variables that are added to a metaclass are called
 Class Instance Variables."

| ivnames oldEnd tCls |

self isMeta ifFalse:[ ^ self error:'receiver is not a metaClass' ].
tCls  := self thisClass .
(tCls isInvariant not or: [tCls subclassesDisallowed]) ifFalse: [
  ^ self _error: #rtErrClassNotModifiable
  ].
instVarNamesArray size == 0 ifTrue: [ ^self ].
"allow an error handler to proceed from the errors signaled by the validation
 method and thus skip offending instance variable names"
ivnames := { } .
instVarNamesArray do:[ :aStr | ivnames add: aStr asSymbol ].
ivnames := ivnames select: [:e |
  self _validateNewClassInstVar: e
  ].
oldEnd := self instSize + 1.
self _inheritCIVs: ivnames at: oldEnd.
%

method:
_validateNewClassInstVar: ivname

"Returns true if the name passes all tests.  Generates errors if the name fails
 due to a) being an invalid identifier or b) being a duplicate of an existing
 instance variable, either in this metaclass or in one of this metaclass's
 instance's subclasses metaclasses."

| subs |
((ivname size == 0
  or: [ivname isValidIdentifier not])
  or:[ ivname _isSymbol not]) ifTrue: [
  self _error: #classErrBadIdentifier args: { ivname }.
  ^false
  ].
(instVarNames includesIdentical: ivname) ifTrue: [
  self _error: #rtErrAddDupInstvar args: { ivname }.
  ^false
  ].
self thisClass isModifiable ifTrue: [
  subs := self _subclasses .
  subs ~~ nil ifTrue: [
    subs do: [:sub |
      (sub _validateNewClassInstVar: ivname) ifFalse: [ ^false ]
      ].
    ].
  ].
^true
%

category: 'Class Instance Variables'
method: 
_inheritCIVs: ivnames at: inheritedIVIndex

"Adds the instance variables specified in the argument to the receiver at the
 given location.

 Instance variables that are added to a metaclass are called
 Class Instance Variables."

| myConstr newConstr names vname subClss tCls numNewIvs |

self isMeta ifFalse:[ ^ self error:'receiver is not a metaClass' ].
(numNewIvs := ivnames size) == 0 ifTrue: [ ^self ].
(numNewIvs + self instSize) > GEN_MAX_INSTANCE_VARS ifTrue:[
  self error: 'new instVars would exceed max number of instVars'
]. 
myConstr := constraints .
myConstr size == 0 ifTrue:[ myConstr := nil ]
                  ifFalse:[ newConstr := { Object } ].
names := Array withAll: instVarNames.
tCls := self thisClass .
1 to: numNewIvs do: [:i |
  vname := ( ivnames at: i) asSymbol .
  names insertObject: vname at: (inheritedIVIndex + i - 1).
  self _incrementInstVars: 1 .
  self _unsafeAt: 4 put: (InvariantArray withAll: names).
  myConstr ifNotNil:[
    self _unsafeAt: 5 put: constraints , newConstr .
  ].
  "force the class to increase in size by storing one position off the end"
  self _refreshClassCache: false .
  tCls _insertCivAt: (inheritedIVIndex + i - 1).
  ].
tCls _refreshClassCache: false .
tCls subclassesDisallowed ifFalse: [
  subClss := self _subclasses.
  subClss ~~ nil ifTrue:[
    subClss do: [:e |
      e _inheritCIVs: ivnames at: inheritedIVIndex.
      ].
    ].
  ].
"finally, recompile my methods to ensure that the instance variable indexes are
 correct"
self _recompileMethodsAfterNewIvOffset: inheritedIVIndex.
%

