! =========================================================================
! Copyright (C) VMware, Inc. 1986-2012.  All Rights Reserved.
!
! $Id: ProfMonitorTree.gs 19139 2009-06-15 23:48:14Z lalmarod $
!
! Superclass Hierarchy:
!   ProfMonitorTree, ProfMonitor, Object.
!
!=========================================================================

expectvalue %String
run
| cls |
cls := Globals at: #ProfMonitorTree ifAbsent:[
  ProfMonitor subclass: 'ProfMonitorTree'
    instVarNames: #()
    classVars: #()
    classInstVars: #()
    poolDictionaries: { }
    inDictionary: Globals
].
^ cls definition
%
removeallmethods ProfMonitorTree
removeallclassmethods ProfMonitorTree

category: 'For Documentation Installation only'
classmethod: ProfMonitorTree
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'ProfMonitorTree is an enhancement of ProfMonitor, adding a tree based
display of the profiling results.' .
doc documentClassWith: txt.

self comment: doc asString.
%

category: 'Private'
method: ProfMonitorTree
gatherPMEsForTree

	| res rawArray rawArraySiz rawArrayPointer sampleStart sampleEnd rootPME reposCls |

	res := OrderedCollection new. 
	rawArray := self _readSampleFile .	
	(rawArraySiz := rawArray size) < 1 ifTrue:[ ^ self ].
	rawArrayPointer := 1 .
	rootPME := self readFirstPMEFrom: rawArray startingAt: rawArrayPointer.
  reposCls := SystemRepository .
	[ rawArrayPointer < rawArraySiz ] whileTrue:[
			(rawArray at: rawArrayPointer) == reposCls ifFalse:[
         self error: 'data order problem'
      ].
			sampleStart := rawArrayPointer + 2.   "skip count"
			sampleEnd := sampleStart + ((rawArray at: rawArrayPointer + 1) * 2) - 1.
			self tallyPME: rootPME 
					position: sampleEnd 
					sampleStartIndex: sampleStart 
					sampleEndIndex: sampleEnd 
					in: rawArray.
			rawArrayPointer := sampleEnd + 1.
	].

	results at: 3 put: rootPME.

				
%
category: 'Reporting'
method: ProfMonitorTree
gatherResults
	"See superclass comment"
	super gatherResults.
	self gatherPMEsForTree.
%
category: 'Private'
method: ProfMonitorTree
printablePercentFor: aFractionOrOtherNumber
	
	^((aFractionOrOtherNumber asFloat * 100) 
		asStringUsingFormat: #(-6 1 false)) , '%'
%
category: 'Private'
method: ProfMonitorTree
printPME: aPME on: str total: theTotal
	
	str add: (self printablePercentFor: (aPME tally / theTotal)); space.
	str add: '('; add: aPME tally asString; add: ')'; space.

	str add: (aPME asStringWidth:20) ; lf
%
category: 'Private'
method: ProfMonitorTree
readFirstPMEFrom: rawArray startingAt: index
	"trace consistes of SystemRepository, count, and (count*2) slots with method and class.
	The highest point in context is at end, create a PME based on that."

| count |
(rawArray at: index) == SystemRepository 
	ifFalse: [self error: 'Cannot read first PME'].
count := (rawArray at: index + 1).
count _isSmallInteger ifFalse: [self error: 'corrupt prof monitor file'].

^ ProfMonitorEntry newForMethod: (rawArray at: (index + (count*2)))  
	  receiverClass: (rawArray at: (index + (count*2) + 1)).
%
category: 'Reporting'
method: ProfMonitorTree
reportDownTo: tally
	"See superclass comment"
	| rpt |
  	rpt := super reportDownTo: tally.
        rpt lf; lf.
	rpt add: (self _treeReportDownTo: tally); 
		lf; lf.
	^ rpt
%
category: 'Private'
method: ProfMonitorTree
tallyPathPME: aPME position: index sampleStartIndex: startIndex sampleEndIndex: endIndex in: rawData

	"Add a tally point for the PME's child, the one that matches the stack at 
	the given index.  Create child if does not exist. "

	| aMethod aClass matchingPME |
	aMethod := rawData at: index - 1.
  aMethod class == GsNMethod ifFalse:[ aMethod := nil ].
	aClass := rawData at: index .
  aPME ifNotNil:[
	  aPME children do: [:aChildPME | 
		  aChildPME cmethod == aMethod ifTrue:[ matchingPME := aChildPME ].
	  ].
    aMethod ifNotNil:[ aClass ifNotNil:[
	    matchingPME ifNil:[
        matchingPME := ProfMonitorEntry newForMethod: aMethod receiverClass: aClass. 
		    aPME children add: matchingPME.
	    ].
    ]].
  ].
	^ matchingPME
%
category: 'Private'
method: ProfMonitorTree
tallyPME: aPME position: index sampleStartIndex: startIndex sampleEndIndex: endIndex in: rawData
	" go backwards, top of stack is at end"

	| childPME idxMinus1 idxMinus2 mth |

  idxMinus1 := index - 1 .
	mth := rawData at: idxMinus1 .
  mth ifNotNil:[ mth class == GsNMethod ifTrue:[ aPME ifNotNil:[
    mth == aPME cmethod ifTrue: [aPME tally: aPME tally + 1]
  ]]].

	idxMinus1 == startIndex ifTrue: [^self "end of path" ] .

  idxMinus2 := index - 2 .
	childPME := self tallyPathPME: aPME position: idxMinus2
						sampleStartIndex: startIndex sampleEndIndex: endIndex 
						in: rawData.

	self tallyPME: childPME 
			position: idxMinus2
			sampleStartIndex: startIndex sampleEndIndex: endIndex 
			in: rawData.
%

category: 'Private'
method: ProfMonitorTree
treePrintPME: aPME on: str tabs: tabs thisTab: myTab total: theTotal downTo: limit
	| children childrenTab |
	
	tabs do: [:tab | str add: tab].
	self printPME: aPME on: str total: theTotal.
	children := aPME children.
	children isEmpty ifFalse: [
      tabs addLast: myTab.
			children := children asSortedCollection: [:a :b | a tally > b tally].
			1 to: children size do: [:i | | child |
					child := children at: i.
					(child tally >= limit) ifTrue: [
						childrenTab := i < children size ifTrue: ['   |'] ifFalse: ['   '].
						self
							treePrintPME: child
							on: str
							tabs: tabs 
							thisTab: childrenTab
							total: theTotal
							downTo: limit
          ].
			].
			tabs removeLast
    ]
%
category: 'Private'
method: ProfMonitorTree
_treeReportDownTo: tally

| rpt elapsedTime |
  rpt := String new .
  (endTime >= startTime) 
	ifTrue: [elapsedTime := endTime - startTime] 
	ifFalse: [elapsedTime := endTime + (16rffffffff - startTime)].

  rpt add:'================'; lf ; add: 'STACK SAMPLING TREE RESULTS'; lf ;
    	add: 'elapsed CPU time:    '; 
    	add: (elapsedTime) asString; add: ' ms';  lf ;
        add: (self _intervalString: interval ) ; lf ;
   	add: 'reporting down to: '; 
     	add: tally asString; add: ' hits';  lf; lf.

results size < 3 ifTrue: [^rpt].
(results at: 3) children do: [:aPME |
	self 
		treePrintPME: aPME 
		on: rpt 
		tabs: OrderedCollection new  
		thisTab: '  ' 
		total: aPME tally
		downTo: tally
	].
^rpt
%

