!=========================================================================
! Copyright (C) VMware, Inc. 1986-2012.  All Rights Reserved.
!
! $Id: ReadWriteStreamPortable.gs 22222 2009-09-02 00:36:20Z stever $
!
! Based on Pharo-1.2.1 implementation
!
! Superclass Hierarchy:
!   ReadWriteStream, WriteStream, PositionableStream, Stream, Object.
!
!=========================================================================

! class created by installStreamHierarchy.topaz

removeallmethods ReadWriteStream
removeallclassmethods ReadWriteStream


category: 'For Documentation Installation only'
classmethod: ReadWriteStream
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'ANSI compliant implementation of ReadWriteStream is a PositionableStream that allows 
its objects to be read and written. 

See ANSI 5.9.7'.
doc documentClassWith: txt.

self comment: doc asString.
%

category: 'Accessing'
method: ReadWriteStream
contents
"Answer with a copy of my collection from 1 to readLimit."

readLimit := readLimit max: position.
^collection copyFrom: 1 to: readLimit
%

category: 'Accessing'
method: ReadWriteStream
next
"Returns the next object that the receiver can access for reading.
 Returns nil if an attempt is made to read beyond the end of the stream."

position >= readLimit
	ifTrue: [^nil]
	ifFalse: [^collection at: (position := position + 1)]
%

category: 'Accessing'
method: ReadWriteStream
next: amount
"Answer the next amount elements of my collection.  overriden for efficiency"

| ans endPosition |
amount < 0 ifTrue: [ ^self error: 'amount may not be less than 0' ].
readLimit := readLimit max: position.

endPosition := position + amount  min:  readLimit.
ans := collection copyFrom: position + 1 to: endPosition.
position := endPosition.
^ans
%

category: 'Testing'
method: ReadWriteStream
nextMatchFor: anObject

	"The first object is removed from the receiver's future sequence value and appended to the end of
	the receiver's past sequence values. The value that would result from sending #= to the object with
	anObject as the argument is returned.
	The results are undefined if there are no future sequence values in the receiver.

	ANSI 5.9.2.6"

	^self next = anObject.
%
