!=========================================================================
! Copyright (C) VMware, Inc. 1986-2012.  All Rights Reserved.
!
! $Id: bomlastconv.gs 29610 2012-11-20 05:02:00Z otisa $
!
! Description - 
!   This file contains the subset of bomlast.gs used for repository
!   conversion. (bomlast.gs is not used for conversion).
!=========================================================================

! Fix up methods that would have contained forward reference to PrintStream.
!  PrintStream requires execution to create class; execution not
!  possible until filein2 stage, but we need Object>>printString in 
!  filein1 in order to make topaz execution work ... 
category: 'Formatting'
method: Object
printString

"Returns a String whose contents are a displayable representation of the
 receiver."

"This method uses the printOn: method to create the String."

| ws str |

str := String new.

ws := PrintStream printingOn: str.  

self printOn: ws.
^ ws _collection "contents might have been converted to an EUCString."
%


! compile methods that contain forward references to Float if they are compiled
! earlier in the filein when building a fresh GemStone repository.

category: 'Converting'
method: Integer
asDecimalFloat

"Returns a DecimalFloat representing the receiver."

<primitive: 137>

| factor sum oldRoundingMode radix |

oldRoundingMode := Float roundingMode.  "nearest even is only right way"
Float roundingMode: 'nearestEven'.

sum := 0.0F0 .
radix := 4294967295.0F0 .
factor := self sign asFloat.
1 to: self size do: [:i |
  sum := (self _digitAt: i) * factor + sum.
  factor := factor * radix 
].
DecimalFloat roundingMode: oldRoundingMode.
^sum
%

category: 'Truncation and Rounding'
method: Number
ceiling

"Returns the integer that is closest to the receiver, on the same side
 of the receiver as positive infinity.  In particular, returns the
 receiver if the receiver is an integer."

(self <= 0.0)
   ifTrue: [^self truncated]
   ifFalse: [^self negated floor negated truncated]
%

category: 'Arithmetic'
method: Number
degreesToRadians

"Assuming the receiver represents an angle in degrees, returns the angle
 in radians."

^ self / 57.29577951308232
%

category: 'Arithmetic'
method: Number
radiansToDegrees

"Assuming the receiver represents an angle in radians, returns the angle
 in degrees."

^ self * 57.29577951308232
%

method: GsMethodDictionary
rebuildTable: newSize

"Rebuilds the method dictionary by populating a larger method dictionary
 first and doing a (primitive) become:"

"This is a reimplementation of the method in object.gs to handle
instances that have a dependency list."

<primitive: 2001>
| prot |
prot := System _protectedMode .
[
  | newGsMethodDict depList |

  tableSize = newSize ifTrue:[ 
    ^ self "no change in table size" 
  ].
  collisionLimit == 536870911 ifTrue:[
    ^ self 		"avoid recursive rebuild"
  ].

  newGsMethodDict := self class new: (newSize * 2).
  newGsMethodDict valueConstraint: valueConstraint.
  newGsMethodDict keyConstraint: keyConstraint.
  
  self keysAndValuesDo: [ :aKey :aValue |
    newGsMethodDict at: aKey put: aValue.
    ].
  
  (depList := DependencyList for: self) ~~ nil
    ifTrue: [ DependencyList set: nil for: self ].
  
  newGsMethodDict _becomeDictionary: self.
  
  depList ~~ nil
    ifTrue: [ DependencyList set: depList for: self ].
] ensure:[
  prot _leaveProtectedMode
]
%

expectvalue %String
run
  "Per ANSI 5.6.8  , additional names for Float, to fix bug 40876"
| report |
report := String new .
{ #FloatE . #FloatD . #FloatQ } do:[:k |
  (Globals includesKey: k) ifFalse:[
     Globals at: k put: Float .
     (Globals associationAt: k) immediateInvariant .
     report add: 'added ' ; add: k ; lf .
  ].
].
report
%

!--------------------------------------

category: 'Repository Conversion'
method: Class
_unsafeDisallowSubclasses

""

self _unsafeAt: 2 "offset of format"
           put: (format bitOr: 32 "subclasses disallowed bit") .
self _refreshClassCache: false
%

category: 'Repository Conversion'
method: Class
_unsafeMakeClassNp

""

self _unsafeAt: 2 "offset of format"
           put: (format bitOr: 16r800"instancesNP bit" ) .
self _refreshClassCache: false
%

!--------------------------------------
run
| list |
  list := { Behavior . Class . ObsoleteMetaclass . Repository . UserProfileSet . 
     AbstractUserProfileSet . 
     CompiledMethod . Float . DecimalFloat . SmallFloat . SmallDouble . 
     Symbol . DoubleByteSymbol . 
     UnimplementedFloat1 . UnimplementedFloat2 . 
     MethodContext . Block . SelectionBlock . 
     GsObjectSecurityPolicy . SymbolList . System . UndefinedObject . UserProfile . Boolean . 
     Character . SmallInteger . Integer . LargeInteger . 
     JISCharacter . StackSegment . 
     StackBuffer . Activation . Process . VariableContext . BlockClosure . 

     ExecBlock . ExecutableBlock . SimpleBlock . ComplexBlock . ComplexVCBlock . 
     GsMethod . GsNMethod . GsNativeCode . 
     GsProcess . GsMethodDictionary . GsStackBuffer . 

     SelectBlock . BtreeBasicInteriorNode . BtreeBasicLeafNode . BasicSortNode . 
     RcIndexBucket . RcIndexBucketWithCache . RcIndexDictionary . NscBuilder . 

     PathTerm . ConstrainedPathTerm . SetValuedPathTerm  "fix bug 11484"
     }.

  list do: [ :aClass | aClass _unsafeDisallowSubclasses ] .

  list do: [ :aClass |
     aClass subclassesDisallowed ifFalse:[ ^ 'Error on ' , aClass name ].
     ] .
  ^ true
%
run
Semaphore _unsafeMakeClassNp .  "set instancesNonPersistent"
ProcessorScheduler _unsafeMakeClassNp .  "set instancesNonPersistent"
^ true
%

run
Class removeSelector: #_unsafeDisallowSubclasses .
Class removeSelector: #_unsafeMakeClassNp .
^ true
%

expectvalue %SmallInteger
run
"Disallow GCI create/store of instances of Metaclasses.
 We don't patch customer classes.  Newly created metaclasses will get this
 automatically."
| count |
count := 0 .
Class class _structuralUpdatesDisallowed ifFalse:[
  Globals valuesDo:[:v |
    v isBehavior ifTrue:[ | m |
      (m := v class) isMeta ifTrue:[
        m _disallowGciCreateStore . count := count + 1
      ]
    ]
  ].
].
^ count
%

run
| list blk |

" allow adding instance variables to the following classes.  fixes bug 6569"
list := { UserProfile } .

blk := [ :aClass |
          aClass _unsafeAt: 4 "instVarNames"
                 put: (Array withAll: (aClass instVarAt: 4)) .
          aClass _unsafeAt: 5 "constraints"
                     put: (Array withAll: (aClass instVarAt: 5)) .
          aClass _refreshClassCache: false
       ].
list do: blk .
^ true
%

! deleted "Change the constraints on VariableContext..." , deletes fix 15424

! fixup classHistory of Symbol and String for fix 42440
run
| hist ver |
hist := Symbol classHistory.
hist size: 0 ;
  add: DoubleByteSymbol ; add: QuadByteSymbol; add: Symbol  .
hist := DoubleByteSymbol classHistory .
hist size: 0 ;
  add: Symbol ; add: QuadByteSymbol; add: DoubleByteSymbol .
hist := QuadByteSymbol classHistory .
hist size: 0 ;
  add: Symbol ; add: DoubleByteSymbol ; add: QuadByteSymbol .
^ (ver := Symbol currentVersion) == Symbol ifTrue:[ true ] ifFalse:[ ver ].
%
run
| hist ver |
hist := String classHistory.
hist size: 0 ;
  add: MultiByteString ; add: String  .
^ (ver := String currentVersion) == String ifTrue:[ true ] ifFalse:[ ver ].
%


!--------------------------------------

!--------------------------------------
! clear "conversion in progress flag"   
expectvalue false
run
ConversionStatus at: 1 put: false
%

!--------------------------------------
! Create class documentation for kernel classes that are only placeholders.

category: 'For Documentation Installation only'
classmethod: System
installPlaceholderDocumentation

| str |

str :=
' is a placeholder for another related GemStone product that has not currently
  been installed.' .

" list of placeholder classes for other products "
{ GsfClassDefinitionInfo . GsfModificationLog }
  do: [ :aClass |
    | doc txt |
    doc := GsClassDocumentation _newForPrivateGsClass: aClass.
    txt := (GsDocText new) details:
           ((String withAll: (aClass name)) add: str; yourself).
    doc documentClassWith: txt.
    aClass comment: doc asString.
  ].

" list of placeholder classes within the server "
{ UnimplementedFloat1 . UnimplementedFloat2 }
  do: [ :aClass |
    | doc |
    doc := GsClassDocumentation _newForPrivateGsClass: aClass.
    doc documentClassWith: (GsDocText new).
    aClass comment: doc asString.
  ].
%


!--------------------------------------
! Install class documentation for kernel classes only.

run

| classes |

System installPlaceholderDocumentation.

classes := ClassSet new.
(Globals values) do: [ :each |
  ( (each isKindOf: Class)
      and:[(each class includesSelector: #installDocumentation)] )
  ifTrue: [ classes add: each.  ].
  ].

classes do: [ :aClass | 
  aClass installDocumentation .
  GsFile gciLogServer:'installed doc for ', aClass name  . 
] .
classes do: [ :aClass |
  (aClass class) removeCategory: 'For Documentation Installation only'.
  ].
^true
%
commit

!------------------------------------------------------------
! Make various methods in the indexing subsystem 'protected'
!  now done via compilation directives during method compilation
!------------------------------------------------------------

!------------------------------------------------------------
! Make various classes in the indexing subsystem 'protected'.
! This prevents accessing an instance through structural access
! 31162: also protect classes for code security.
!------------------------------------------------------------
!
run
" list of classes which will be protected "
{ GsObjectSecurityPolicySet . 
   SetValuedPathTerm . 
   SetValuedPathEvaluator . UserSecurityData . 
"31162"
   GsMethod . Class . GsMethodDictionary . SymbolDictionary . SymbolList . 
   UserProfile
 }
    do: [ :aClass | aClass _makeProtected ].

^ true
%
lev 1
obj AbstractCharacter
lev 0
commit

!------------------------------------------------------------
! Build the dictionary of configuration file parameters
!------------------------------------------------------------
run
Globals at:#ConfigurationParameterDict put:
(System _configFileParameterDict) .
^ true
%
!------------------------------------------------------------
! Build the dictionary of software version information parameters
!------------------------------------------------------------
run
"dictionary to translate names to internal identifiers"
Globals at:#VersionParameterDict put:
(System _versionParameterDict) .
^ true
%
run
"dictionary of image version information"
Globals at:#ImageVersion put: SymbolDictionary new .
^ true
%
input $upgradeDir/version.topaz

! build the set of classes for which instances are disallowed
run
Globals at:#InstancesDisallowed put:
 ( IdentitySet withAll:
  { Activation . Character . Behavior . Block . BlockClosure . Boolean . 
      Class . GsMethod . ComplexBlock . ComplexVCBlock . DateTime . 
      ExecBlock . ExecutableBlock . Float . DecimalFloat . SmallFloat . BinaryFloat . 
      Integer . JISCharacter . 
      LargeInteger . ObsoleteMetaclass . MethodContext . Number . Process . 
      Repository . GsObjectSecurityPolicy . SelectBlock . SelectionBlock . SimpleBlock . 
      SmallInteger . StackBuffer . StackSegment . System . UndefinedObject . 
      UserProfile . VariableContext }) .
^ true
%
commit

!------------------------------------------------------------
! Create collating tables for raw ASCII collating order for use with
! String compare methods  in category 'Other comparisons'
!------------------------------------------------------------
run
| a |
a := String new .
0 to: 255 do:[:j| a add: (Character withValue: j )].
Globals at:#AsciiCollatingTable put: a .
a immediateInvariant .
^ a size == 256
%
run
| a |
a := DoubleByteString new .
0 to: 255 do:[:j| a add: (Character withValue: j )].
Globals at:#DoubleByteAsciiCollatingTable put: a .
a immediateInvariant .
^ a size == 256
%
run
| a |
a := QuadByteString new .
0 to: 255 do:[:j| a add: (Character withValue: j )].
Globals at:#QuadByteAsciiCollatingTable put: a .
a immediateInvariant .
^ a size == 256
%

!-----------
!  ensure bom has created the old canonicalized float zeros

! old oop 1873
expectvalue %Float
obj @239617

! old oop 1875
expectvalue %SmallFloat
obj @239873


! reset weak refs array 
expectvalue true
run
Globals at: #GcWeakReferences put: Array new.
^ true
%

! indexing  not supported, so clear SharedDependencyLists
expectvalue true
run
Globals at: #SharedDependencyLists put: DepListTable new .
^true
%

!-------------
! removal of _unsafesetoop method done later in filein


!------------------------------------------------------------
! fixes 31329 , 31540 - Remove references to Geode
!------------------------------------------------------------

expectvalue %String
run
" remove references to geode without creating Symbols unnecessarily"
| result geodeKeys globalsKeys removeList report |
result := String new .
geodeKeys := { 'Geode' . 'GeodeGlobals' . 'GeodeLauncher' . 
		'GeodeMaint' . 'GeODEDbfHistory' . 'ObsoleteGeode' }.
globalsKeys := Globals keys . "collection of Symbols"
removeList := Array new .
geodeKeys do:[ :aGeod |
  globalsKeys do:[ :aGlb | 
     aGlb asString = aGeod ifTrue:[ removeList add: aGeod ]
  ].
].
report := String new .
removeList do:[ :aKey |
  Globals removeKey: aKey asSymbol .
  report add: ' removed ' , aKey , ' ; ' .
  ].
^ report .
%


!------------------------------------------------------------
! Fix 31942, disallow gci store to instances of SymbolList .
! If customers have created subclasses of SymbolList , those are not
! changed .
expectvalue true
run
SymbolList _disallowGciCreateStore .
^ SymbolList _structuralUpdatesDisallowed
%

commit

expectvalue true
run
 "Check that certain kernel classes have only one version defined.
  The virtual machine contains optimizations which assume that
  the classes checked below have only one version, and thus the
  virtual machine can skip looking at the classHistory when doing
  certain isKindOf checks on arguments to primitives.  
  See fix 32339"

| clsList |
clsList := { AbstractCharacter . 
  Array . 
  ExecBlock . 
  ExecutableBlock . 
  GsNMethod . 
  GsProcess . 
  GsSocket . 
  IdentityBag . 
  Interval . 
  SequenceableCollection . 
  SmallFloat . 
  SmallDouble . 
  VariableContext  }.

clsList do:[ :aCls | | hist val sz |
  hist := aCls classHistory .
  sz := hist size .
  sz == 1 ifFalse:[ 
    nil error:'classHistory for ' , aCls name , ' too big' 
  ].
  val := hist at: 1 .
  val == aCls ifFalse:[
    nil error:'classHistory for ' , aCls name , ' is corrupt'
  ].
].   
"primitives for Float do not check classHistory, but explicitly
 check arguments for both SmallDouble and Float ."
true ifTrue:[ | hist val sz |
  hist := Float classHistory .
  sz := hist size .
  sz == 1 ifFalse:[
    nil error: 'classHistory for Float has wrong size.'
  ].
  val := hist at: 1 .
  val == Float ifFalse:[
    nil error:'classHistory for Float, bad value at:1' . 
  ].
].
^ true
%

expectvalue %String
run
| dcSeg |
dcSeg := Globals at:#DataCuratorObjectSecurityPolicy .
AllUsers objectSecurityPolicy == dcSeg ifTrue:[
  ^ 'objectSecurityPolicy ok'
] ifFalse:[
  AllUsers objectSecurityPolicy: dcSeg .
  AllUsers do:[ :aUp | 
    aUp objectSecurityPolicy == dcSeg ifFalse:[ aUp objectSecurityPolicy: dcSeg].
  ].
  ^ 'changed objectSecurityPolicy'
]
%
lev 1
obj AbstractCharacter
lev 0

expectvalue %String
run
	"ensure Association for FdcResults is in DataCurator objectSecurityPolicy"
| res dcSeg |
dcSeg := Globals at:#DataCuratorObjectSecurityPolicy .
(Globals includesKey:#FdcResults) ifTrue:[
  (Globals associationAt:#FdcResults) objectSecurityPolicy: dcSeg .
  res := 'changed objectSecurityPolicy of FdcResults' .
] ifFalse:[
  dcSeg setCurrentWhile:[
    Globals at:#FdcResults put: nil .
    res := 'created FdcResults' . 
  ]
].
^ res 
%

! adjust existing UserProfiles to for new inverse privileges
!  is now done at end of nameless.gs

! v3.1 deleted code for fix 38848, UserSecurityData class is fixed up by bom.c

expectvalue %String
run
Globals at: #GemStoneRCLock ifAbsent:[
  Globals at: #GemStoneRCLock put: Object new .
  ^ 'created'
].
^ 'found'
%

!------------------------------------------------------------
! Remove behavior from obsolete classes if repository originated
!  on Gs64 v2.3 or prior.  bom.c code run in first conversion session
!  has already renamed them !  to Obsolete*  if they existed.
!------------------------------------------------------------
expectvalue %String
run
 | names report eol |
 names := { 
  'ObsoleteExceptionA' . 
  'ObsoleteError' . 
  'ObsoleteMessageNotUnderstood' . 
  'ObsoleteZeroDivide' . 
  'ObsoleteNotification' . 
  'ObsoleteWarning' . 
  'ObsoleteTestFailure' . 
  'ObsoleteResumableTestFailure' . 
  'ObsoleteExceptionHandler' . 
  'ObsoleteExceptionSet' . 
  'ObsoleteFailedMessage' }.
 "Obsolete subclasses of  Number handled in respective files previous to here"
 report := String new .
 eol := Character lf .
 names do:[ :aName | | aSym aCls |
   (aSym := Symbol _existingWithAll:aName) ~~ nil ifTrue:[
     aCls := Globals at: aSym otherwise: nil .
     aCls ~~ nil ifTrue:[
       aCls removeAllMethods .
       aCls class removeAllMethods .
       report addAll: aName , ' removed all methods' .
     ] ifFalse:[
       report addAll: aName , ' not found in Globals'.
     ].
   ] ifFalse:[
     report addAll: aName , ' symbol does not exist'.
   ].
   report add: eol .
 ].
^ report 
%

! 42240: Code to convert description fields into comment methods
! deleted as part of 3.1 comment redesign.

expectvalue %String
run
  "Move GsCompilerIRNode and subclasses to a GsCompilerClasses dictionary
   and/or remove them from Globals."
  | dict rpt |
  rpt := String new .
  (dict := Globals at: #GsCompilerClasses otherwise: nil) ifNotNil:[
    rpt add: 'GsCompilerClasses exists' ; lf .
  ] ifNil:[ | irNodeCls |
    rpt add: 'created GsCompilerClasses' ; lf .
    (dict := SymbolDictionary new) name: #GsCompilerClasses .
    irNodeCls := Globals at:#GsCompilerIRNode .
    Globals associationsDo:[ : assoc | | val |
      val := assoc _value .
      val isBehavior ifTrue:[
        ((val isSubclassOf: irNodeCls) 
           or:[ val name == #IndentingStream]) ifTrue:[ 
          dict addAssociation: assoc 
        ].
      ].
    ]. 
    Globals at:#GsCompilerClasses put: dict .
  ].
  dict keys do:[ :aKey |
    aKey ~~ #GsCompilerClasses ifTrue:[ 
      (Globals includesKey: aKey) ifTrue:[
         Globals removeKey: aKey .
         rpt add: 'removed ' , aKey ; lf . 
      ].
    ].
  ]. 
  (Globals includesKey:#IndentingStream) ifTrue:[
    dict addAssociation: (Globals associationAt:#IndentingStream).
    Globals removeKey:#IndentingStream.
    rpt add: 'removed #IndentingStream'; lf .
  ].
  Globals removeKey:#isInternedNeedsSymbolCheck ifAbsent:[
    rpt add: 'isInternedNeedsSymbolCheck not found'; lf 
  ].
  ^ rpt.
%

! fix 42260
expectvalue %String
run
  "Remove GsRdbConnection if not being used by GemConnect"
  | rpt |
  rpt := String new .
  rpt add: 'Checking GsRdbConnection status..'; lf .
  (Globals includesKey: #GsRdbConnection) ifTrue: [
    (Globals includesKey: #GsOracleConnection)
    ifTrue: [ 
      rpt add: 'GemConnect installed: GsRdbConnection retained.'; lf ]
    ifFalse: [
      Globals removeKey: #GsRdbConnection.
      rpt add: 'Class GsRdbConnection not used: removed.'; lf .
  ]].
^ rpt
%

expectvalue %String
run
  "Move selected obsolete classes to ObsoleteClasses, to fix 40985"
  | rpt dict |
  rpt := String new .
  (dict := Globals at: #ObsoleteClasses otherwise: nil) ifNotNil:[
    rpt add: 'ObsoleteClasses exists' ; lf .
  ] ifNil:[
    rpt add: 'created ObsoleteClasses' ; lf .
    (dict := SymbolDictionary new ) name: #ObsoleteClasses .
    #( Block
       MethodContext
       Obsolete23ClampSpecification
       Obsolete23GsFile
       ObsoleteClampSpecification
       ObsoleteDateTime
       ObsoleteDictionary
       ObsoleteGsFile
       ObsoleteIdentityCollisionBucket
       ObsoleteIdentityDictionary
       ObsoleteLanguageDictionary
       ObsoleteRcCollisionBucket
       ObsoleteSymbol
       ObsoleteSymbolAssociation
       ObsoleteSymbolDictionary
       ObsoleteSymbolKeyValueDictionary
       ObsoleteSymbolListDictionary
       ObsoleteSymbolSet
       ObsoleteVariableContext
       ReenterBlock
       SelectionBlock
       ObsoleteTestCase
       ObsoleteTestResource
       ObsoleteExampleSetTest
       ObsoleteResumableTestFailureTestCase
       ObsoleteSimpleTestResourceTestCase
       ObsoleteSUnitTest
      ObsoleteError ObsoleteException ObsoleteExceptionHandler ObsoleteExceptionSet 
      ObsoleteFailedMessage ObsoleteMessageNotUnderstood ObsoleteNotification 
      ObsoleteResumableTestFailure 
      ObsoleteTestFailure ObsoleteWarning ObsoleteZeroDivide
      ) do:[:sym | 
         (Globals associationAt: sym otherwise: nil ) ifNotNil:[:assoc|
           dict addAssociation: assoc .
           rpt add: 'added ' , sym ; lf .
         ]
      ].
    Globals at:#ObsoleteClasses put: dict . 
  ].
  dict keys do:[ :aKey |
    aKey ~~ #ObsoleteClasses ifTrue:[
      (Globals includesKey: aKey) ifTrue:[
         Globals removeKey: aKey .
         rpt add: 'removed from Globals:  ' , aKey ; lf .
      ].
    ].
  ].
  ^ rpt
%
run
"Install class variable used to support recompile of instances of GsMethod
 to produce instances of GsNMethod .  This dictionary names the classes
 for which it is NOT valid to reuse the existing literal variables in a method."

| cNames dict |
 cNames := #(  "classes for which want to use the current definition when recompiling"
   "triples:  key in association after 3.x upgrade, sourceName, 
		3.x lookup key if different than sourceName"
   "classes normally moved to ObsoleteClasses"
   #Obsolete23ClampSpecification #ClampSpecification nil
   #ObsoleteError #Error nil
   #ObsoleteException #Exception nil
   #ObsoleteExceptionHandler #ExceptionHandler nil
   #ObsoleteExceptionSet #ExceptionSet nil
   #ObsoleteFailedMessage #FailedMessage nil
   #Obsolete23GsFile #GsFile nil
   #ObsoleteGsProcess #GsProcess nil
   #ObsoleteGsSocket #GsSocket nil
   #ObsoleteMessageNotUnderstood #MessageNotUnderstood nil
   #ObsoleteNotification #Notification nil
   #ObsoleteResumableTestFailure #ResumableTestFailure nil
   #ObsoleteTestFailure #TestFailure nil
   #ObsoleteVariableContext #VariableContext nil
   #ObsoleteWarning #Warning nil
   #ObsoleteZeroDivide #ZeroDivide  nil
   "classes in Globals"

   #ObsQuadByteString  #QuadByteString nil
   #ObsDoubleByteSymbol #DoubleByteSymbol nil
   #ObsDoubleByteString #DoubleByteString nil
   #ObsLargeNegativeInteger #LargeNegativeInteger  #LargeInteger
   #ObsLargePositiveInteger #LargePositiveInteger  #LargeInteger
   #ObsSmallFloat  #SmallFloat nil
   #ObsFloat       #Float   nil
   #FixedPoint     #ScaledDecimal  #FixedPoint "use FixedPoint in oldCode" 

   "ObsoleteTimeZone   not handled"
   "ObsoleteMetaclass  not handled"
   "ObsoleteIgnoredException  not handled"
).
 dict := SymbolDictionary new .
 1 to: cNames size by: 3 do:[:j | | obsName newName vArr srcName oldCls |
   obsName := cNames at: j .         "key in 3.x association for obsolete class"
   newName :=   cNames at: j + 1 .   "key to lookup in 3.x Globals"
   srcName :=  (cNames at: j + 2) ifNil:[ newName ].
   oldCls := Globals at:obsName ifAbsent:[ 
               (Globals at:#ObsoleteClasses) at: obsName otherwise:nil ] .
   oldCls ifNotNil:[ vArr := { oldCls . newName . srcName }.
     dict at: obsName put:  vArr .
     dict at: newName put: vArr .
   ].
 ].
 dict at:#ExceptionA put: { Globals at:#ObsoleteExceptionA otherwise: nil . 
	 		     #Exception } .
 (GsMethod _resolveClassVar: #ObsoleteClassesDict) value: dict .
 true
%

! fix 42259
expectvalue %String
run
| seg |
(seg := Globals at:#PublishedSegment otherwise:nil) ifNil:[
  Globals at:#PublishedSegment put: PublishedObjectSecurityPolicy.
  ^ 'Installed key PublishedSegment'
] ifNotNil:[
  (seg == PublishedObjectSecurityPolicy) ifFalse:[  
    ^ 'PublishedSegment refers to some other object'.
  ].
  ^ 'PublishedSegment already present'
].
%

! fix 42240 -- upgrade comment handling of Global classes to 3.1 protocol

expectvalue true
run

  | errors lf |
  errors := false.
  lf := Character lf.
  GsFile gciLogServer: 
    'Converting comments in Global classes to 3.1 standard.'.

  Globals classesDo: [:aUserProfile :aSymbolDictionary :aClass |

    [ aClass upgradeComments ] 
    on: Error do: [:ex | | lf |
        lf := Character lf.
        errors := true.
        GsFile gciLogServer:
           'ERROR: Class ' + aClass name asString + 
           ' failure during comment upgrade: ' + lf +
           ex description asString + lf +
           '---------------------------------' + lf +
           (GsProcess stackReportToLevel: 300) + lf +
           '---------------------------------' + lf.
  ]].

  " Trigger a final internal error if any errors seen during scan.."
  errors ifTrue: [ 
    InternalError signal: 'Problem with comment conversion - check log' ].

  GsFile gciLogServer: ' ' + lf + ' '.
^ true
%

commit

! end of bomlastconv.gs
