!=========================================================================
! Copyright (C) VMware, Inc. 1986-2012.  All Rights Reserved.
!
! $Id: criticalsection.gs 28927 2012-08-22 22:26:20Z monty $
!
! Superclass Hierarchy:
!   CriticalSection, Object
!
!=========================================================================

expectvalue %String
run
(Globals at:#CriticalSection otherwise: nil) ifNotNil:[ :cls | |constr|
  constr := cls _constraints .
  (constr at: 1) == ObsoleteGsProcess ifTrue:[
    constr _unsafeAt: 1 put: GsProcess . "bypass invariance"
    ^ 'changed constraint'
  ].
].
^ 'no change'
%    

expectvalue %String
run
^ Object _newKernelSubclass: #CriticalSection
  instVarNames: #(#owner #semaphore)
  classVars: #()
  classInstVars: #()
  poolDictionaries: { }
  inDictionary: Globals
  constraints: { { #owner . GsProcess } .
                 { #semaphore . Semaphore } }
      options: #()
  reservedOop: 917
%

! Remove existing behavior from CriticalSection
removeallmethods CriticalSection
removeallclassmethods CriticalSection

! ------------------- Class methods for CriticalSection
category: 'Documentation'
classmethod: CriticalSection
comment
^'CriticalSection is a class from the Blue Book API that
provides one way of using semaphores.'
%

category: 'Instance Creation'
classmethod: CriticalSection
new
  "Answer a new CriticalSection"
  ^self basicNew _initialize
%

category: 'Private'
method: CriticalSection
_scheduler

"Returns   ProcessorScheduler scheduler "
<primitive: 457>

self _primitiveFailed: #_scheduler
%


! ------------------- Instance methods for CriticalSection
category: 'Blue Book Protocol'
method: CriticalSection
critical: aBlock
  "Wait until no other processes are sending critical to the receiver
   and then execute aBlock and return its value."

  | value activeProcess |

  activeProcess := self _scheduler activeProcess.
  (owner == activeProcess) ifFalse: [
    semaphore wait.
    owner := activeProcess.
    value := aBlock ensure: [
      owner := nil.
      semaphore signal.
    ].
  ] ifTrue: [
    value := aBlock value.
  ].
  ^value
%

category: 'Private'
method: CriticalSection
_initialize

  owner := nil.
  semaphore := Semaphore forMutualExclusion.
%
