!=========================================================================
! Copyright (C) VMware, Inc. 1986-2012.  All Rights Reserved.
!
! $Id: gspackage.gs 28927 2012-08-22 22:26:20Z monty $
!
! Superclass Hierarchy:
!   GsPackage, Object
!
!=========================================================================

expectvalue %String
run
| oldCls |
oldCls := Globals at:#GsPackage otherwise: nil .
oldCls == nil ifTrue:[
 Object subclass: 'GsPackage'
  instVarNames: #( sessionMethods prereqs symbolDict)
  classVars: #()
  classInstVars: #()
  poolDictionaries: { }
  inDictionary: Globals .
  ^ 'created class'
] ifFalse:[
  ^ 'existing class definition:  ' + oldCls definition
].
%

! Remove existing behavior from GsPackage
removeallmethods GsPackage
removeallclassmethods GsPackage
set class GsPackage

! ------------------- Class methods for GsPackage
category: 'Accessing'
classmethod:
globalName

  ^#GsPackage_Current
%
category: 'Instance Creation'
classmethod:
installIn: aSymbolDictionary

  | package |
  package := self new.
  aSymbolDictionary at: self globalName put: package.
    package symbolDict: aSymbolDictionary.
  ^package
%
category: 'Instance Creation'
classmethod:
new

  ^self basicNew initialize
%
! ------------------- Instance methods for GsPackage
category: 'Initialize-Release'
method:
initialize

  sessionMethods := IdentityKeyValueDictionary new.
  prereqs := { } .
%
category: 'Enumerating'
method:
behaviorAndMethodDictDo: aBlock
  self sessionMethods keysAndValuesDo: [:beh :ar |
      aBlock value: beh value: (ar at: 1)
  ].
%
category: 'Compiling'
method:
methodAndCategoryDictionaryFor: aBehavior 

  "Returns a 2 element array { methodDict . categoryDict }"
  | arr |
  arr := self sessionInfoFor: aBehavior.
  ^ { arr at: 1 . arr at: 2 }
%

category: 'Compiling'
method:
setPragmas: pragmaArray
forBehavior: aBehavior
forMethod: selector

    | ar methodPragmaDict |
    ar := self sessionMethods at: aBehavior otherwise: nil .
    ar ifNil:[ ^ nil ].
    ar size < 4 ifTrue: [ar at: 4 put: IdentityKeyValueDictionary new].
    methodPragmaDict := (ar at: 4) at: aBehavior otherwise: nil.
    methodPragmaDict ifNil:[
      methodPragmaDict := IdentityKeyValueDictionary new.
      (ar at: 4) at: aBehavior put: methodPragmaDict.
    ].
  pragmaArray ifNil:[ methodPragmaDict removeKey: selector asSymbol otherwise: nil ]
         ifNotNil:[ methodPragmaDict at: selector asSymbol put: pragmaArray ].
  ^self
%
category: 'Compiling'
method:
setStamp: aStamp 
forBehavior: aBehavior
forMethod: selector

    | ar methodStampDict |
    ar := self sessionMethods at: aBehavior otherwise: nil .
    ar ifNil:[ ^ nil ].
    methodStampDict := (ar at: 3) at: aBehavior otherwise: nil.
  methodStampDict == nil 
    ifTrue: [ 
      methodStampDict := IdentityKeyValueDictionary new.
      (ar at: 3) at: aBehavior put: methodStampDict.
     ].
  aStamp == nil
    ifTrue: [ methodStampDict removeKey: selector asSymbol otherwise: nil ]
    ifFalse: [ methodStampDict at: selector asSymbol put: aStamp ].
  ^self
%
category: 'Compiling'
method:
methodPragmaDictFor: aBehavior

    | ar |
    ar := self sessionMethods at: aBehavior otherwise: nil .
    ar ifNil:[ ^ nil ].
    ar size < 4 ifTrue: [ar at: 4 put: IdentityKeyValueDictionary new].
   ^(ar at: 4) at: aBehavior otherwise: nil
%
category: 'Compiling'
method:
methodStampDictFor: aBehavior

    | ar |
    ar := self sessionMethods at: aBehavior otherwise: nil .
    ar ifNil:[ ^ nil ].
   ^(ar at: 3) at: aBehavior otherwise: nil
%
category: 'Accessing'
method:
sessionMethods

  ^ sessionMethods
%
category: 'Categories'
method:
addCategory: aSymbol for: aBehavior
  "returns the newly added category (a SymbolSet) or nil if not newly added"
  | ar catDict |
  ar := self sessionInfoFor: aBehavior.
  catDict := (ar at: 2).
  (catDict includesKey: aSymbol)
    ifFalse: [ | aSet |
      aSet := SymbolSet new .
      catDict add: (SymbolAssociation newWithKey: aSymbol value: aSet ).
      ^aSet].
  ^ nil
%
category: 'Categories'
method:
removeCategory: aSymbol for: aBehavior

    | ar catDict |
  ar := self sessionMethods at: aBehavior otherwise: nil .
  ar ifNil:[ ^ self ].
  catDict := (ar at: 2).
  catDict removeKey: aSymbol otherwise: nil 
%
category: 'Categories'
method:
categoryNamesFor: aBehavior into: anArray

    | ar catDict |
    ar := self sessionMethods at: aBehavior otherwise: nil .
    ar ifNil:[ ^ self ].
    catDict := (ar at: 2).
    catDict keysAndValuesDo: [:aKey :aValue |
      anArray add: aKey
    ].
%
category: 'Categories'
method:
selectorsIn: categoryName for: aBehavior into: anArray

    | ar catDict |
    ar := self sessionMethods at: aBehavior otherwise: nil .
    ar ifNil:[ ^ self ].
    catDict := (ar at: 2).
    anArray addAll: (catDict at: categoryName asSymbol otherwise: #()).
%
category: 'Methods'
method:
copyCategoryDictFor: aBehavior into: aGsMethodDictionary 
 
    | ar catDict | 
    ar := self sessionMethods at: aBehavior otherwise: nil .
    ar ifNil:[ ^ self ].
    catDict := (ar at: 2). 
    catDict keysAndValuesDo: [:key :values | | coll | 
        coll := aGsMethodDictionary at: key otherwise: nil. 
        coll ifNil:[
            coll := SymbolSet new. 
            aGsMethodDictionary at: key put: coll
        ].
        coll canBeWritten ifFalse: [ 
            coll := coll copy. 
            aGsMethodDictionary at: key put: coll
        ].  
        coll addAll: values.
    ].
%
category: 'Methods'
method:
selectorsFor: aBehavior into: anArray

    | ar methDict |
    ar := self sessionMethods at: aBehavior otherwise: nil .
    ar ifNil:[ ^ self ].
    methDict := (ar at: 1).
    anArray addAll: methDict keys.
%
category: 'Methods'
method:
categoryOfSelector: aSymbol for: aBehavior

    | ar catDict |
    ar := self sessionMethods at: aBehavior otherwise: nil .
    ar ifNil:[ ^ nil ].
    catDict := (ar at: 2).
    catDict keysAndValuesDo: [:aKey :aValue | 
      (aValue includesIdentical: aSymbol ) ifTrue:[ ^ aKey ].
    ].
    ^ nil
%
category: 'Methods'
method:
compiledMethodAt: aSymbol for: aBehavior

    | ar methDict |
    ar := self sessionMethods at: aBehavior otherwise: nil .
    ar ifNil:[ ^ nil ].
    methDict := (ar at: 1).
    ^methDict at: aSymbol otherwise: nil
%
category: 'Methods'
method:
includesSelector: aSymbol for: aBehavior

    | ar methDict |
    ar := self sessionMethods at: aBehavior otherwise: nil .
    ar ifNil:[ ^ false ].
    methDict := (ar at: 1).
    ^methDict includesKey: aSymbol
%
category: 'Methods'
method:
copyMethodDictFor: aBehavior into: aGsMethodDictionary

    | ar methDict |
    ar := self sessionMethods at: aBehavior otherwise: nil .
    ar ifNil:[ ^ self ].
    methDict := (ar at: 1).
    methDict keysAndValuesDo: [:key :value |
        aGsMethodDictionary at: key put: value
    ].
%
category: 'Accessing'
method:
symbolDict

    ^symbolDict
%
category: 'Accessing'
method:
symbolDict: aSymDict

    symbolDict := aSymDict
%
category: 'Accessing'
method:
prereqs

    ^prereqs
%
category: 'Accessing'
method:
addPrereq: aSymbolDict

    prereqs add: aSymbolDict
%
category: 'Methods'
method:
removeAllMethodsFor: aBehavior

    self sessionMethods removeKey: aBehavior otherwise: nil .  
%
category: 'Methods'
method:
removeMethodAt: aSymbol for: aBehavior

  | ar methDict meth catDict stampDict pragmaDict sel |
  ar := self sessionMethods at: aBehavior otherwise: nil .
  ar ifNil: [ ^nil ]. 
  methDict := (ar at: 1).
  meth := methDict removeKey: aSymbol otherwise: nil .
  meth ifNil: [ ^nil ].
  stampDict := (ar at: 3).
  stampDict removeKey: aSymbol otherwise: nil .
  ar size == 4 ifTrue: [
      pragmaDict := (ar at: 4).
      pragmaDict removeKey: aSymbol otherwise: nil .
  ].
  catDict := (ar at: 2).
  sel := meth selector .
  catDict keysAndValuesDo: [:aKey :aValue | 
    (aValue remove: sel otherwise: nil ) ifNotNil:[ ^ meth ].
  ].
  ^meth
%
category: 'Methods'
method:
sessionInfoFor: aBehavior

  | ar |
  ar := self sessionMethods at: aBehavior otherwise: nil.
  ar == nil
    ifTrue: [ | methodDict categoryDict stampDict pragmaDict |
      methodDict := GsMethodDictionary new.
      categoryDict := GsMethodDictionary new.
            stampDict := IdentityKeyValueDictionary new.
            pragmaDict := IdentityKeyValueDictionary new.
      ar := { methodDict . categoryDict . stampDict . pragmaDict } .
      self sessionMethods at: aBehavior put: ar.
    ].
    ^ar
%
category: 'Methods'
method:
removeAllSubclassCodeFor: aBehavior
    | ar methDict |
    ar := self sessionMethods at: aBehavior otherwise: nil .
    ar ifNil:[ ^ self ].
    methDict := (ar at: 1).
    methDict keysAndValuesDo: [ :aKey :aMethod |
      methDict at: aKey put: ((methDict at: aKey) _copyToForceRecompilation).
    ].
%
category: 'Categories'
method:
removeSelector: aSelector fromCategoriesFor: aBehavior

  | ar catDict removed |
  ar := self sessionMethods at: aBehavior otherwise: nil .
  ar ifNil:[ ^ false ].
  catDict := (ar at: 2).
  removed := false .
  catDict keysAndValuesDo: [:aKey :aValue | 
    (aValue remove: aSelector otherwise: nil ) ifNotNil:[ removed := true ].
  ].
  ^ removed
%
category: 'Categories'
method:
addSelector: aSelector toCategory: categoryName for: aBehavior
    "new category is expected to exist"
  | ar catDict |
  ar := self sessionMethods at: aBehavior otherwise: nil .
  ar ifNil:[ ^ self ].
  catDict := (ar at: 2).
    (catDict at: categoryName) add: aSelector.
%

method:
moveSelector: aSelector toCategory: newCat for: aBehavior
    "new category is expected to exist"
  (self removeSelector: aSelector fromCategoriesFor: aBehavior) ifTrue:[
     self addCategory: newCat for: aBehavior ;
         addSelector: aSelector toCategory: newCat for: aBehavior .
  ].
%

method: 
categorysDo: aBlock for: aBehavior
  "evaluates aBlock for each method category of receiver
   in specified environment. Returns the receiver.

   aBlock should be a two argument block
   expecting the args  categoryNameSymbol ,   selectorsSet.
   aBlock may be invoked more than once for each category name.
   The iteration is done directly over the receiver's categories."

    | ar |
  ar := self sessionMethods at: aBehavior otherwise: nil .
  ar ifNil:[ ^ self ].
  (ar at: 2) keysAndValuesDo: aBlock
%
