!=========================================================================
! Copyright (C) VMware, Inc. 1986-2012.  All Rights Reserved.
!
! $Id: indexmgrpolicy.gs 28927 2012-08-22 22:26:20Z monty $
!
! Superclass Hierarchy:
!  IndexManagerAutoCommitPolicy, Object.
!
!=========================================================================

! class created in idxclasses.topaz

removeallmethods IndexManagerAutoCommitPolicy
removeallclassmethods IndexManagerAutoCommitPolicy

category: 'For Documentation Installation only'
classmethod: IndexManagerAutoCommitPolicy
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'IndexManagerAutoCommitPolicy implements the behavior for determining when commits should be
 performed when the IndexManager autoCommit feature is enabled. Commits will be performed 
 whenever the number of persistent objects modified exceeds the dirtyObjectCommitThreshold or
 when the tempObjSpacePercentUsed crosses the percentTempObjSpaceCommitThreshold.'.
doc documentClassWith: txt.

self comment: doc asString.
%

! ------------------- Class methods for IndexManagerAutoCommitPolicy
category: 'Instance Creation'
classmethod: IndexManagerAutoCommitPolicy
on: anIndexManager

^ self new indexManager: anIndexManager
%
! ------------------- Instance methods for IndexManagerAutoCommitPolicy
category: 'Accessing'
method: IndexManagerAutoCommitPolicy
dirtyObjectCommitThreshold

^ dirtyObjectCommitThreshold
%
category: 'Evaluation'
method: IndexManagerAutoCommitPolicy
evaluate: aBlock for: anIndexManager

self percentTempObjSpaceCommitThreshold ifNil:[
    "no % memory threshold protection"
  aBlock value.
] ifNotNil:[
  | errNumAlmostOutOfMemory |
  errNumAlmostOutOfMemory := ErrorSymbols at: #rtErrSignalAlmostOutOfMemory .
 "Enable AlmostOutOfMemory signal"
  System signalAlmostOutOfMemoryThreshold: self percentTempObjSpaceCommitThreshold .
  [ 
    aBlock onException: AlmostOutOfMemory do:[ :ex | 
      ex number == errNumAlmostOutOfMemory ifTrue:[ | systm | 
        anIndexManager _doCommit .
        (systm := System) _vmMarkSweep .
        (systm _tempObjSpacePercentUsedLastMark 
             <  self percentTempObjSpaceCommitThreshold) ifTrue: [ 
           "dropped below the threshold, reenable the signal"
           systm enableAlmostOutOfMemoryError.
        ] ifFalse:[
          "likely started already above the threshold
           and commits won't necessarily help with out of memory problems"
        ].
        ex resume .
      ] ifFalse:[
        ex outer
      ]
    ]
  ] ensure: [
    "disable AlmostOutOfMemory signal"
    System signalAlmostOutOfMemoryThreshold: -1.
  ].
]
%
category: 'Initialization'
method: IndexManagerAutoCommitPolicy
indexManager: anIndexManager

dirtyObjectCommitThreshold := anIndexManager dirtyObjectCommitThreshold.
percentTempObjSpaceCommitThreshold := anIndexManager percentTempObjSpaceCommitThreshold.
%
category: 'Accessing'
method: IndexManagerAutoCommitPolicy
percentTempObjSpaceCommitThreshold

^ percentTempObjSpaceCommitThreshold
%
category: 'Testing'
method: IndexManagerAutoCommitPolicy
shouldCommit
  "Answer true if enough dirty objects have been modified, or memory is close to full."

 | systm  |
 systm := System . 
 ^ (systm _numPersistentObjsModified > self dirtyObjectCommitThreshold) or:[
    systm _tempObjSpacePercentUsedLastMark >= self percentTempObjSpaceCommitThreshold
 ] 
%
