
category: 'Message Digests'
method:
_asMessageDigestKind: opCode

"
opCode   Digest Algorithm   Digest Bits
=======================================
  1          md5            128
  2          sha1           160
  3          sha256         256
  4          sha512         512
=======================================

"
<primitive: 666>
self _primitiveFailed: #_asMessageDigestKind:
%

category: 'Message Digests'
method:
md5sum

"Return the 128 bit MD5 checksum of the receiver as a LargeInteger. 

 Computation is per RFC 1321 , http://www.ietf.org/rfc/rfc1321.txt,
 using L. Peter Deutsch's C implementation from
 http://sourceforge.net/projects/libmd5-rfc/

 For DoubleByteString and QuadByteString, the computation is based
 on viewing the string as a ByteArray holding big-endian characters.
" 

^ self _asMessageDigestKind: 1
%

category: 'Message Digests'
method:
asMd5String

"Compute the 128 bit MD5 message digest for the receiver and return it as
 a 32 character string of hexadecimal characters."

^ self md5sum asHexStringWithLength: 32
%

category: 'Message Digests'
method:
sha1Sum

"Return the 160 bit SHA1 checksum of the receiver as a LargeInteger. 

 Computation is per FIPS PUB 180-3:
   http://csrc.nist.gov/publications/fips/fips180-3/fips180-3_final.pdf

 For DoubleByteString and QuadByteString, the computation is based
 on viewing the string as a ByteArray holding big-endian characters.
" 

^ self _asMessageDigestKind: 2
%

category: 'Message Digests'
method:
asSha1String

"Compute the 160 bit SHA1 message digest for the receiver and return it as
 a 40 character string of hexadecimal characters."

^ self sha1Sum asHexStringWithLength: 40
%

category: 'Message Digests'
method:
sha256Sum

"Return the 256 bit SHA256 checksum of the receiver as a LargeInteger. 

 Computation is per FIPS PUB 180-3:
   http://csrc.nist.gov/publications/fips/fips180-3/fips180-3_final.pdf

 For DoubleByteString and QuadByteString, the computation is based
 on viewing the string as a ByteArray holding big-endian characters.
" 

^ self _asMessageDigestKind: 3
%

category: 'Message Digests'
method:
asSha256String

"Compute the 256 bit SHA1 message digest for the receiver and return it as
 a 64 character string of hexadecimal characters."

^ self sha256Sum asHexStringWithLength: 64
%

category: 'Message Digests'
method:
sha512Sum

"Return the 512 bit SHA512 checksum of the receiver as a LargeInteger. 

 Computation is per FIPS PUB 180-3:
   http://csrc.nist.gov/publications/fips/fips180-3/fips180-3_final.pdf

 For DoubleByteString and QuadByteString, the computation is based
 on viewing the string as a ByteArray holding big-endian characters.
" 

^ self _asMessageDigestKind: 4
%

category: 'Message Digests'
method:
asSha512String

"Compute the 512 bit SHA1 message digest for the receiver and return it as
 a 128 character string of hexadecimal characters."

^ self sha512Sum asHexStringWithLength: 128
%
