!=========================================================================
! Copyright (C) VMware, Inc. 1986-2012.  All Rights Reserved.
!
! $Id: rcbtreebasicinter.gs 28927 2012-08-22 22:26:20Z monty $
!
! Superclass Hierarchy:
!   RcBtreeBasicInteriorNode, BtreeBasicInteriorNode, BtreeInteriorNode, BtreeNode, Array,
!   SequenceableCollection, Collection, Object.
!
!=========================================================================

! class created in idxclasses.topaz

removeallmethods RcBtreeBasicInteriorNode
removeallclassmethods RcBtreeBasicInteriorNode

category: 'For Documentation Installation only'
classmethod: RcBtreeBasicInteriorNode
installDocumentation

| doc txt |
doc := GsClassDocumentation _newForPrivateGsClass: self.

txt := (GsDocText new) details:
'A RcBtreeBasicInteriorNode represents an interior node of a B-tree for which a
 binary search can be done for a key. Updates to instances of this class are 
 logged and can  be replayed in the case of commit conflicts.' .
doc documentClassWith: txt.

self comment: doc asString.
%

! ------------------- Class methods for RcBtreeBasicInteriorNode
! ------------------- Instance methods for RcBtreeBasicInteriorNode
category: 'Updating'
method: RcBtreeBasicInteriorNode
at: aKey put: aValue

self shouldNotImplement: #at:put:
%
category: 'Updating'
method: RcBtreeBasicInteriorNode
btreeAt: aKey put: aValue

""

self shouldNotImplement: #btreeAt:put:
%
category: 'Updating'
method: RcBtreeBasicInteriorNode
btreeAt: aKey put: aValue for: anIndexObj selectiveAbortSet: selectiveAbortSetOrNil

| res |
(selectiveAbortSetOrNil ~~ nil and: [ (selectiveAbortSetOrNil includes: self) not ])
        ifTrue: [ 
                selectiveAbortSetOrNil add: self.
                self _selectiveAbort ].
res := self _at: aKey put: aValue for: anIndexObj selectiveAbortSet: selectiveAbortSetOrNil.
System _addRootObjectToRcReadSet: self.
System redoLog addConflictObject: self for: anIndexObj.
^res
%
category: 'Constants'
method: RcBtreeBasicInteriorNode
parentNodeClass

"Returns the class of node to be created as the parent when a split occurs."

^ RcBtreeBasicInteriorNode
%
category: 'Removing'
method: RcBtreeBasicInteriorNode
removeKey: aKey value: aValue

self shouldNotImplement: #removeKey:value:
%
category: 'Removing'
method: RcBtreeBasicInteriorNode
removeKey: aKey value: aValue for: anIndexObj selectiveAbortSet: selectiveAbortSetOrNil

| res |
(selectiveAbortSetOrNil ~~ nil and: [ (selectiveAbortSetOrNil includes: self) not ])
        ifTrue: [ 
                selectiveAbortSetOrNil add: self.
                self _selectiveAbort ].
res := self _removeKey: aKey value: aValue for: anIndexObj selectiveAbortSet: selectiveAbortSetOrNil.
System _addRootObjectToRcReadSet: self.
System redoLog addConflictObject: self for: anIndexObj.
^res
%
