!=========================================================================
! Copyright (C) VMware, Inc. 1986-2012.  All Rights Reserved.
!
! $Id: readstr.gs 28927 2012-08-22 22:26:20Z monty $
!
! Superclass Hierarchy:
!   ReadStream, PositionableStream, Stream, Object.
!
!=========================================================================

removeallmethods ReadStream
removeallclassmethods ReadStream

category: 'Documentation'
classmethod:
comment

^ 'ReadStream / ReadStreamLegacy is a PositionableStream that allows its objects to 
 be read but not written.'.
%

category: 'Instance Creation'
classmethod: ReadStream
new

"Disallowed.  To create a new ReadStream, use the class method on: instead."

self shouldNotImplement: #new
%

category: 'Accessing'
method: ReadStream
next

"Returns the next object that the receiver can access for reading.  Generates
 an error if an attempt is made to read beyond the end of the stream."
| pos |
(self atEnd) ifTrue: [ ^ EndOfStream signal ].
pos := position .
position := pos + 1 .
^ itsCollection at: pos 
%

method: ReadStream
nextOrNil
 "Returns the next object that the receiver can access for reading.
  Returns nil if an attempt is made to read beyond the end of the stream.
  For use with Streams whose collections do not contain nil .
"
 | pos coll res |
 pos := position .
 pos > (coll := itsCollection) size ifTrue:[ ^ nil ]. "inline atEnd check"
 res := coll at: pos .
 position := pos + 1 .
 ^ res
%


category: 'Accessing'
method: ReadStream
nextElements: count into: anObject

"Stores the next count elements that the receiver can access for reading
 into anObject.  The receiver's collection and anObject must be compatible
 SequenceableCollections.  Returns the count of elements read.

 Generates an error if an attempt is made to read beyond the end of the stream."

"Used by PassiveObject."

count <= 0 ifTrue:[ ^ 0 ].
(self atEnd)
   ifTrue: [ ^ EndOfStream signal ].

anObject replaceFrom: 1 to: 1 + count - 1 with: itsCollection startingAt: position .
position := position + count.
^ count
%

category: 'Adding'
method: ReadStream
nextPut: anObject

"Disallowed.  You cannot write to a ReadStream."

self shouldNotImplement: #nextPut:
%

category: 'Accessing'
method: ReadStream
_fastNext

"Returns the next item on the input stream without end of stream checks.
 This may result in a bad access error instead of an end of stream error."

| res |
res := itsCollection at: position.
position := position + 1.
^res
%

category: 'Positioning'
method: ReadStream
_fastPosition: pos

"Sets the position of the receiver without checking the validity of the
 argument 'pos'.  This may result in bad index errors instead of end
 of stream errors in subsequent access of the stream."

position := pos
%

