!=========================================================================
! Copyright (C) VMware, Inc. 1986-2012.  All Rights Reserved.
!
! $Id: reflection.gs 21891 2009-07-08 23:43:35Z bretlb $
!
! Superclass Hierarchy:
!   Reflection, Object.
!
!=========================================================================

expectvalue %String
run
(Globals includesKey: #Reflection) ifFalse: [
  ^ Object 
	  _newKernelSubclass: 'Reflection'
	  instVarNames: #()
	  classVars: #()
	  classInstVars: #()
	  poolDictionaries: { }
	  inDictionary: Globals
          options: #()
	  reservedOop: nil.
] ifTrue:[
  ^ 'class exists'
]  
%

removeallmethods Reflection
removeallclassmethods Reflection
set class Reflection

category: 'For Documentation Installation only'
classmethod: Reflection
installDocumentation
| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'Reflection is an abstract class that has no instances.  It implements class
 methods for object so that it is safe to get information about the object
 without sending it a message.'.
doc documentClassWith: txt.

self comment: doc asString.
%

category: 'Private'
classmethod: Reflection
_reflect: anObject opCode: anInt

<primitive: 837 >
anInt _validateClass: SmallInteger.
anInt > 4 ifTrue: [ anInt _error: #rtErrArgOutOfRange "bad opcode" ].
self _primitiveFailed: #_reflect:opcode: args: { anObject . anInt } .
%

category: 'Private'
classmethod: Reflection
_reflect: anObject opCode: opcodeInt arg: anArg

<primitive: 838 >
opcodeInt _validateClass: SmallInteger.
opcodeInt > 6 ifTrue: [ opcodeInt _error: #rtErrArgOutOfRange "bad opcode" ].
opcodeInt == 6 ifTrue:[ anArg _validateClass: Symbol ].
self _primitiveFailed: #_reflect:opcode:withArg: 
     args: { anObject . opcodeInt . anArg } .
%

category: 'Private'
classmethod: Reflection
_reflect: anObject opCode: opcodeInt arg1: anArg1 arg2: anArg2

<primitive: 839 >
opcodeInt _validateClass: SmallInteger.
opcodeInt > 3 ifTrue: [ opcodeInt _error: #rtErrArgOutOfRange "bad opcode" ].
opcodeInt <= 2 ifTrue:[ anArg1 _validateClass: SmallInteger ]
          ifFalse:[     anArg1 _validateClass: Symbol ].
self _primitiveFailed: #_reflect:opcode:withArg: 
     args: { anObject . opcodeInt . anArg1 . anArg2 } .
%

category: 'Accessing'
classmethod: Reflection
oopOf: anObject
"Answers the oop of the object, consistent with Object>>asOop."

^self _reflect: anObject opCode: 1
%

category: 'Accessing'
classmethod: Reflection
classOf: anObject
"Answers the class of the object, consistent with GciFetchClass."

^self _reflect: anObject opCode: 2
%

category: 'Accessing'
classmethod: Reflection
implementationOf: anObject
"Answers #pointers, #bytes, #NSC, or #special, consistent with GciFetchObjImpl."

| impl |
impl := self _reflect: anObject opCode: 3.
(impl == 0) ifTrue: [ ^ #pointers ].
(impl == 1) ifTrue: [ ^ #bytes ].
(impl == 2) ifTrue: [ ^ #NSC ].
(impl == 3) ifTrue: [ ^ #special ].
^ #InvalidImplementation
%

category: 'Accessing'
classmethod: Reflection
sizeOf: anObject

" Answers 0 if #implementationOf: answers #special. 
  Otherwise, answers the number of pointers (for #pointers or #NSC)
    or bytes (#bytes) that the object currently holds. 

  For #pointers objects (small and large), this includes both named instance variables
    and the object's varying size.
  Consistent with GciFetchSize. "

| impl |
impl := self _reflect: anObject opCode: 3.
(impl == 3) ifTrue: [ ^ 0 ].
^self _reflect: anObject opCode: 4
%

category: 'Accessing'
classmethod: Reflection
namedSizeOf: anObject

" Answers 0 if #implementationOf: answers #special. 
  Otherwise, answers the number of named instance variables in the object.

  Consistent with GciFetchNamedSize."

| impl |
impl := self _reflect: anObject opCode: 3.
(impl == 3) ifTrue: [ ^ 0 ].
^self _reflect: anObject opCode: 5
%

category: 'Updating'
classmethod: Reflection
setSizeOf: anObject to: newSize

" Sets the indexable size of anObject so that it will subsequently answer newSize
  when sent sizeOf:. Consistent with GciSetVaryingSize.
  Not valid for #NSC or #special objects."

self _reflect: anObject opCode: 1 arg: newSize
%

category: 'Accessing'
classmethod: Reflection
fetchFrom: anObject at: anIndex
" Answers the object or byte found at the given index in the object. 
  Legal values for anIndex are 1 to sizeOf: anObject. Not valid for 0-size objects.
  
  Consistent with GciFetchOop and GciFetchByte."

^self _reflect: anObject opCode: 2 arg: anIndex - 1
%

category: 'Updating' 
classmethod: Reflection
storeTo: anObject at: anIndex put: aValue
" Stores aValue, which may be any object for #pointers objects, but which
  must be 0-255 for #byte objects, at the given index in the object.
  Consistent with GciStoreOop and GciStoreByte. Legal indices are 1 to
  sizeOf: anObject."

self _reflect: anObject opCode: 1 arg1: anIndex - 1 arg2: aValue
%


category: 'Updating' 
classmethod: Reflection
to: anNsc addObject: anObject

" Add a new object to the NSC. Consistent with GciAddOopToNsc."

self _reflect: anNsc opCode: 3 arg: anObject
%


category: 'Updating' 
classmethod: Reflection
from: anNsc removeObject: anObject

" Removes an existing object from the NSC. Consistent with
  GciRemoveOopFromNsc.
  Answer true if the object was in the NSC and has been removed, false if the
  object was not present."

^self _reflect: anNsc opCode: 4 arg: anObject
%

category: 'Accessing'
classmethod: Reflection
fetchTagOf: anObject tagNum: tagNumber

" Answer the object referred to by the given tag number of the object.
  tagNumber must be 1 or 2. Similar to prim 87."

^self _reflect: anObject opCode: 5 arg: tagNumber
%
classmethod: Reflection
fetchDynamicInstVar: anObject named: aSymbol

 "Return the value of specified dynamic instance variable, or nil if anObject
  has no such dynamic instance variable.  Does not consider instVars defined
  by the anObject's class.  See Object>>dynamicInstVarAt: "
 

^self _reflect: anObject opCode: 6 arg: aSymbol
%

category: 'Updating' 
classmethod: Reflection
storeTagTo: anObject tagNum: tagNumber newValue: aValue

" Store the given value into the specified tag number of the object.
  tagNumber must be 1 or 2. Similar to prim 280."

self _reflect: anObject opCode: 2 arg1: tagNumber arg2: aValue
%
classmethod: Reflection
storeDynamicInstVar: anObject named: aSymbol newValue: aValue

 "Store the value of specified dynamic instance variable, 
  If aValue ~~ _remoteNil, creates the dynamic instance variable if needed.
  aValue == _remoteNil effectively removes the instance variable.
  Does not consider instVars defined by the anObject's class.
  Generates an error if anObjectreceiver is an ExecBlock, Behavior, GsNMethod, 
  or special object.  See  Object>>dynamicInstVarAt:put: "

self _reflect: anObject opCode: 3 arg1: aSymbol arg2: aValue
%

