!=========================================================================
! Copyright (C) VMware, Inc. 1986-2012.  All Rights Reserved.
!
! $Id: reposit.gs 29610 2012-11-20 05:02:00Z otisa $
! Superclass Hierarchy:
!   Repository, Collection, Object.
!
!=========================================================================

removeallmethods Repository
removeallclassmethods Repository

category: 'For Documentation Installation only'
classmethod: Repository
installDocumentation

| doc txt |
doc := GsClassDocumentation newForClass: self.

txt := (GsDocText new) details:
'A Repository is an object that represents a virtual storage into which users
 can place their data.  Each Repository includes an Array of up to 65534
 GsObjectSecurityPolicies, which represent authorization regions for the data.  
 Repositories are described in the GemStone Programming Guide.'.
doc documentClassWith: txt.

txt := (GsDocText new) details:
'The name of the Repository.'.
doc documentInstVar: #name with: txt.

txt := (GsDocText new) details:
'(not used)'.
doc documentInstVar: #dataDictionary with: txt.

txt := (GsDocText new) details:
'Backups and restoration are ordinarily performed while using the GemStone
 DataCurator login.  It is possible to use another login that also has the
 FileControl privilege.  However, for restorations, it is recommended that you
 use only the DataCurator or the SystemUser logins.  If you use another login,
 and that login disappears as a result of the restoration, you will see a
 fatal error.

 The GemStone System Administration Guide discusses backups and restoration
 in more detail.'.
doc documentCategory: #'Backup and Restore' with: txt.

txt := (GsDocText new) details:
'Methods in this category are obsolete and are provided only for compatibility
 with earlier releases of GemStone.  They will be removed in a future release.'.
doc documentCategory: #'Deprecated' with: txt.

self comment: doc asString.
%

expectvalue %String
run
| reposCls |
reposCls := Globals at:#Repository .
reposCls superClass == Array ifTrue:[
  reposCls changeNameTo: #OldRepository .
  Globals at: #Repository put: nil .
  Globals at:#OldRepository put: reposCls .
  ^  Collection _newKernelIndexableSubclass: 'Repository' 
    instVarNames:#( name dataDictionary )
    classVars: #()
    classInstVars: #()
    poolDictionaries: { }
    inDictionary: Globals 
    options: #( disallowGciStore ) 
    reservedOop: 973  
] ifFalse:[
  reposCls superClass == Collection ifTrue:[
    ^ 'class already exists as subclass of Collection'
  ] ifFalse:[
    ^ -1 "error, invalid super class"
  ].
]
%
run
GsObjectSecurityPolicy _disallowGciCreateStore .
^ true
%

category: 'Accessing'
method: Repository
dataDictionary

"Accesses the user-defined data dictionary. (Reserved)"

^ dataDictionary
%

category: 'Accessing'
method: Repository
name

"Returns the logical name of the receiver (a Symbol)."

^ name
%

category: 'Accessing'
method: Repository
fileNames

  <primitive: 429>

"Returns an Array containing the filenames for the extents
 of the receiver.

 Each element within the returned Array contains a String representing
 the filename of the Nth extent (where N is the index into the
 returned Array)."

  self _primitiveFailed: #fileNames .
  self _uncontinuableError
%

! fix 41673, shouldn't rename SystemRepository
! category: 'Updating'
! method: Repository
! name: aString
! 
! "Redefines the logical name of the receiver to be aString."
! 
! name := aString asSymbol
! %

category: 'Updating'
method: Repository
_setVersion

"This method writes the Stone executable version string to the receiver."

<primitive: 289>

self _primitiveFailed: #_setVersion .
self _uncontinuableError
%

! fix 33706
category: 'Clustering'
method: Repository
extentForPage: aPageId

"Returns a SmallInteger specifying an offset into the result from the
 fileNames method.  

 The argument aPageId is an Integer, such as the result from the Object | page
 method, specifying a page in the receiver."

^ (aPageId bitShift: -31 ) + 1
%


category: 'Repository Usage Reporting'
method: Repository
_dbfStatistics: extentFilename

"Returns an Array of size 2 containing the number of bytes the given extent
 requires for physical size and the number of bytes of free space in the
 extent."

<primitive: 283>

extentFilename _validateClass: String .
^ self _primitiveFailed: #_dbfStatistics: args: { extentFilename }
%

! v2.3: edits to obtain stats for multiple extents in one round trip to stone
category: 'Repository Usage Reporting'
method: Repository
_extentStatistics: extentIndex

"If extentIndex > 0 , 
 Returns an Array of size 2 containing the number of bytes the
 given extent requires for physical size and the number of bytes of
 free space in the extent.

 If extentIndex == -1,
 Returns an Array of Arrays. The outer Array has size equal to the
 number of extents; the inner Arrays each have size 2 and contain
 extent szie and extent free space.
"

<primitive: 286>

(extentIndex _isSmallInteger)
  ifTrue:[ ^ self _errorIndexOutOfRange: extentIndex]
  ifFalse:[ extentIndex _validateClass: SmallInteger ].
^ self _primitiveFailed: #_extentStatistics: args: { extentIndex }
%

category: 'Transaction Logging'
method: Repository
_logInfo: logDirId

"If logDirId is within the range of 1 to number of log directories, returns an
 Array containing the following instance variables.  Otherwise returns nil.

 1: logDirId (SmallInteger)
 2: numLogDirs (SmallInteger)
 3: isActiveLog (Boolean)
 4: fileId (SmallInteger) (if active log )
 5: fileSize (SmallInteger) in units of MBytes (if active log)
 6: maxFileSize (SmallInteger) in units of MBytes
 7: logOriginTime (Integer)  (if active log)
 8: fileName (String) (fileOrDevice if active log, else directoryOrDevice)
 9: replicateName (no longer used, always nil)
10: fileId of begin record of oldest transaction in current checkpoint,
	 if active log, otherwise nil
11: fileSize (SmallInteger) in units of 512byte blocks for active log"

<primitive: 397>

logDirId _validateClass: SmallInteger.
^ self _primitiveFailed: #_logInfo: args: { logDirId }
%

category: 'Transaction Logging'
method: Repository
_currentLogInfo

"Returns the result specified for Repository | _logInfo: for the
 currently active transaction log."

| numLogs logInfo |

numLogs := (System configurationAt: #STN_TRAN_LOG_DIRECTORIES) size .
2 timesRepeat:[ "cope with changing to the next log"
  1 to: numLogs do:[ :j |
    logInfo := self _logInfo: j .
    (logInfo at: 3) == true ifTrue:[
      ^ logInfo 
      ].
    ].
  ].
self _halt:'Unable to determine the current log'.
^ { }
%

! fix bug 11283
category: 'Transaction Logging'
method: Repository
currentTranlogSizeMB

"Returns an Integer that is the size of the currently active transaction log in
 units of Megabytes."

^ self _currentLogInfo at: 5
%

category: 'Repository Usage Reporting'
method: Repository
freeSpaceInExtent: extentFilename

"Returns the number of bytes of free space in the extent with the given name.
 This quantity is equal to the number of free pages in the extent times the
 size of a page.

 If the given file is not an active member of the logical Repository
 represented by the receiver, then this method generates an error."

^(self _dbfStatistics: extentFilename) at: 2
%

category: 'Repository Usage Reporting'
method: Repository
fileSizeOfExtent: extentFilename

"Returns the physical size, in bytes, of the extent with the given name.

 If the given file is not an active member of the logical Repository
 represented by the receiver, then this method generates an error."

^(self _dbfStatistics: extentFilename) at: 1
%

category: 'Repository Usage Reporting'
method: Repository
fileSizeReport

"Returns a string which reports on the name, size, and amount of free space
 for each extent and the size and free space of the entire logical Repository."

| result lf stats repositoryFiles extentFile |

    "Initialize"

result := String new.
lf := Character lf.

    "For each extent, add the information about it."

repositoryFiles := self fileNames.
stats := self _extentStatistics: -1 .
1 to: self numberOfExtents do: [:i |
    extentFile := repositoryFiles at: i.
    result addAll: 'Extent #';
           addAll: i asString; add: lf.
    result addAll: '-----------'; add: lf.
    result addAll: '   Filename = ';
           addAll: extentFile; add: lf.
    result addAll: lf  ;
           addAll: '   File size =       ';
           addAll: (self numToMByteString:((stats at:i) at: 1)) ; add: lf;
           addAll: '   Space available = ';
           addAll: (self numToMByteString:((stats at:i) at: 2)) ; add: lf;
           add: lf .
    ].

    "Add totals"

result addAll: 'Totals'; add: lf.
result addAll: '------'; add: lf.
result addAll: '   Repository size = ';
       addAll: (self numToMByteString: self fileSize) ; add: lf;
       addAll: '   Free Space =      ';
       addAll: (self numToMByteString: self freeSpace ) ; add: lf .

^ result
%

category: 'Repository Usage Reporting'
method: Repository
numToMByteString: aNumber

"Convert a number representing a file size in bytes to a formatted
 string reporting the size in megabytes."

| result |
result := String new .
aNumber < 102400 ifTrue: [ ^ '0 Megabytes' ].
result addAll:
  ((aNumber asFloat / 1048576) asStringUsingFormat:#(3 2 false));
  addAll: ' Megabytes' .
^ result
%

category: 'Repository Usage Reporting'
method: Repository
freeSpace

"Returns an integer that gives the number of bytes of free space in the logical
 Repository.  This quantity is equal to the number of free pages in the
 Repository times the size of a page."

| total stats |

total := 0.
stats := self _extentStatistics: -1.
1 to: stats size do:[:j |
  total := total + ((stats at: j) at: 2)
].
^ total
%

category: 'Repository Usage Reporting'
method: Repository
fileSize

"Returns an integer giving the total physical size, in bytes, of all
 of the physical extents that compose the logical Repository."

| total stats |

total := 0.
stats := self _extentStatistics: -1.
1 to: stats size do:[:j |
  total := total + ((stats at: j) at: 1)
].
^ total
%

! fixed 33546
category: 'Repository Usage Reporting'
method: Repository
pageSize

"Returns size in bytes of a disk page in the Repository."

^ System _zeroArgPrim:50
%

category: 'Transaction Logging'
method: Repository
startNewLog

"Causes the most current transaction log to be closed and a new transaction log
 to be opened for writing.  The location of the new log is controlled by the
 STN_TRAN_LOG_DIRECTORIES configuration file parameter.

 If GemStone is running in partial logging mode, then a preceding transaction
 log may be deleted.  See documentation on the STN_TRAN_FULL_LOGGING
 configuration parameter for more details.

 If a checkpoint is in progress, or the repository is in restore mode,
 the operation is NOT performed and a -1 is returned.  
 If it is successful the SmallInteger fileId of the new log is returned.

 This method requires the FileControl privilege."

<primitive: 432>

self _primitiveFailed: #startNewLog .
self _uncontinuableError
%

! deleted replicate argument from addTransactionLog
category: 'Deprecated'
method: Repository
addTransactionLog: deviceOrDirectory  replicate: replicateSpec size: aSize

self deprecated: 'Obsolete, replicate arguments is ignored'.
^ self addTransactionLog: deviceOrDirectory size: aSize
%

category: 'Transaction Logging'
method: Repository
addTransactionLog: deviceOrDirectory  size: aSize
 
"Add deviceOrDirectory to the configuration parameter STN_TRAN_LOG_DIRECTORIES.
 
 The aSize argument must be a positive SmallInteger; it is added to the value
 of the STN_TRAN_LOG_SIZES configuration parameter.

 This method requires the FileControl privilege."

<primitive: 337>

deviceOrDirectory _validateClass: String .
aSize _validateClass: SmallInteger .
aSize > 0 ifFalse:[ aSize _error: #errArgTooSmall args:{ 1 } ] .
^ self _primitiveFailed: #addTransactionLog:size: 
       args: { deviceOrDirectory . aSize }
%

category: 'Transaction Logging'
method: Repository
currentLogFile

"Returns a String containing the file name of the transaction log file to which
 log records are being appended.  If the result is of size 0, then a log has
 failed and a replicate is being used."

 ^ System stoneConfigurationAt:#StnCurrentTranLogNames
%

category: 'Transaction Logging'
method: Repository
allTranlogDirectories

"Returns an Array of Strings, each of which represents 1 element of the 
 STN_TRAN_LOG_DIRECTORIES configuration parameter."

| result |
result := System stoneConfigurationAt: #STN_TRAN_LOG_DIRECTORIES .
^ result class == Array ifTrue:[ result ] ifFalse:[ Array with: result ].
%

category: 'Transaction Logging'
method: Repository
allTranlogSizes

"Returns an Array of SmallIntegers, each of which represents 1 element of the 
 STN_TRAN_LOG_SIZES  configuration parameter."

| result |
result := System stoneConfigurationAt: #STN_TRAN_LOG_SIZES .
^ result class == Array ifTrue:[ result ] ifFalse:[ Array with: result ].
%

category: 'Transaction Logging'
method: Repository
currentLogDirectoryId

"Returns a positive SmallInteger, which is the one-based offset specifying the
 element of the configuration list STN_TRAN_LOG_DIRECTORIES for the current
 transaction log.  (See also the currentLogFile method.)"

^ (System stoneConfigurationAt:#StnCurrentTranLogDirId) + 1 
%

category: 'Transaction Logging'
method: Repository
currentLogFileId

"Returns a positive SmallInteger, which is the internal fileId of 
 the current transaction log."

^ (self _logInfo: (self currentLogDirectoryId)) at: 4
%

category: 'Transaction Logging'
method: Repository
oldestLogFileIdForRecovery

"Returns a positive SmallInteger, which is the internal fileId of the oldest
 transaction log needed to recover from the most recent checkpoint, if the
 Stone were to fail right now."

^ self _currentLogInfo at: 10
%

category: 'Transaction Logging'
method: Repository
logOriginTime

"Returns the log origin time of the receiver. This is the time when a Stone
 started a new sequence of log files for the receiver. A new sequence of
 logs is started if one of the following occurs:

 * Stone is started using extents that were cleanly shutdown, and without
   any log files being present.

 * Stone is started using extents and no existing logs 
   using the 'startstone -N' command.

 * The commitRestore method is executed, and during preceding 
   restore operations we restored a log file with fileId greater
   than the fileId of the log file being written to during the restore."

^ System stoneConfigurationAt:#StnTranLogOriginTime 
%


category: 'Private'
method: Repository
_writeFdcArray: anArray toFile: aFileNameString
"Given an array of objects produced by the one of the findDisconnectedObjects*
 methods, write the array of objects to a file which can be loaded by the
 markGcCandidatesFromFile: method.

 anArray must be an array object and contain objects sorted by object ID.
 aFileNameString must be an instance of String that refers to a file which
 does not yet exist."

<primitive: 557>
anArray _validateClass: Array.
aFileNameString _validateClass: String.
self _primitiveFailed: #_writeFdcArray:toFile: 
     args: { anArray . aFileNameString } .
self _uncontinuableError.
%

category: 'Garbage Collection'
method: Repository
writeFdcArrayToFile: aFileNameString
"Write the array of dead objects found by the last findDisconnectedObjects
 (FDC) to the specified file.  The file is created by this method and
 must not exist when it is called.

 anArray must be an array object and contain objects sorted by object ID.
 The primitive will fail if the objects are not in order.

 aFileNameString must be an instance of String that refers to a file which
 does not yet exist.

 Returns true if successful, false if the entire file could not be written,
 or nil if the results of the last FDC could not be found."

 |array theSize result|
 array := Globals at: #FdcResults otherwise: nil.
 (array == nil)
  ifTrue:[^nil].
 (array size < 3)
  ifTrue:[^nil].
 array := array at: 3.
 (array class == Array)
    ifFalse:[^nil].
 theSize := array size.
 (theSize == 0)
   ifTrue:[^nil].
 result := self _writeFdcArray: array toFile: aFileNameString.
 ^ result == theSize.
%

category: 'Garbage Collection'
method: Repository
findDisconnectedObjectsAndWriteToFile: aFileNameString pageBufferSize: anInt
 saveToRepository: saveToRep
"Perform a markForCollection on the system and write the list of
 dead objects to given file.  If saveToRep is true, also store the results
 in the array under the key #FdcResults in Globals.

 To save the dead objects to an array but not write them to a file, set
 aFileNameString to nil and saveToRep to true.  It is illegal for 
 aFileNameString to be nil and for saveToRep to be false.

 The pageBufSize, which must be a power of two, specifies the number 
 of pages to buffer per thread.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 This method will return an error if the repository is in restore mode.

 *** WARNING *** WARNING *** WARNING *** WARNING *** WARNING ***
 This method is intended to complete the FDC operation as quickly as possible
 and will start 2 threads per CPU core.  Therefore it may consume all available 
 CPU and disk I/O host system resources.

 The FDC operation can be run less aggressively by calling the
 #basicFindDisconnectedObjectsAndWriteToFile: method with a smaller number
 of threads.
 *** WARNING *** WARNING *** WARNING *** WARNING *** WARNING ***

 See the comments in the method:
   Repository>>_fdcWithMaxThreads:waitForLock:pageBufSize:
               percentCpuActiveLimit:toFile:resultArray:

 for a desciption of the result array and a more detailed description of
 this method."

^ self basicFindDisconnectedObjectsAndWriteToFile: aFileNameString 
       pageBufferSize: anInt 
       saveToRepository: saveToRep 
       withMaxThreads: self _aggressiveMaxThreadCount
       maxCpuUsage: 95
%

category: 'Private'
method: Repository
_validatePercentage: anInt
anInt _validateClass: SmallInteger .
((anInt < 1) _or:[ anInt > 100 ])
  ifTrue:[anInt _error: #rtErrArgOutOfRange args:{ 1 . 100 } ] .
%

category: 'Garbage Collection'
method: Repository
findDisconnectedObjectsAndWriteToFile: aFileNameString withMaxThreads: maxThreads
 maxCpuUsage: aPercentage

"Start an FDC operation where the results will be save to the given file but 
 not stored in the repository.  Starts maxThreads on the host system and
 allows the host to run up to aPercentage percent CPU usage.  A page
 buffer of 128 pages (2 MB) is allocated per thread.

 See the comments in the method:
   Repository>>_fdcWithMaxThreads:waitForLock:pageBufSize:
               percentCpuActiveLimit:toFile:resultArray:

 for a desciption of the result array and a more detailed description of
 this method."

^ self basicFindDisconnectedObjectsAndWriteToFile: aFileNameString 
       pageBufferSize: 128 "must be a power of 2"
       saveToRepository: false 
       withMaxThreads: maxThreads
       maxCpuUsage: aPercentage
%

category: 'Private'
method: Repository
_markGcCandidatesFromFile: aFileNameString forceOnError: aBoolean
 resultArray: aCommittedArray errorLimit: errLim readOnly: aBool

"Performs a garbage collection analysis on the objects in the given file by
 scanning the Repository for other references to them.  Objects are marked for
 subsequent reclaiming of storage space if the only references to candidate
 objects are from other candidates.

 aFileNameString is expected to be a String contaning the full path to a file
 containing dead objects.  This file is produced by the method one of the
 findDisconnectedObjects* methods.

 If aFileNameString is nil, no file is loaded and the list of candidates
 is instead assumed to reside in the GcCandidates hidden set.

 All candidate objects are first validated.  A candidate object is valid if
 the following conditions are all true:
   -the object is not a reserved object (not part of the GemStone kernel).
   -the object exists in the object table.
   -the object is not in the stone's list of dead not reclaimed objects.

 If aBoolean is true, then the markGcCandidates method is run even if one
 or more objects fail validation.  If aBoolean is false, then the
 markGcCandidates operation is not run unless all objects are successfully
 validated.

 Setting aBoolean to true also causes the MGC to proceed with the second
 phase of the MGC if any objects were found to be not dead in the first
 phase.  This will happen if any candidate object is referenced by any
 other object that is not a candidate.  If aBoolean is false, the MGC
 will exit if any candidate object is found to be not dead in the first
 phase.  In this case, no dead objects sent to the stone.

 If aBool is true, the MGC runs in read-only mode.  In this mode, 
 the candidates determined to be dead objects are not sent to stone and
 no objects will be reclaimed as a result of this operation.  
 All other MGC functions are performed as described.  In read-only
 mode, the MGC stores the dead objects that would have been passed to the
 stone into hidden set 34 (GcCandidates).

 aCommittedArray must be a committed instance of Array with a size of 7.
 Elements 4 and 7 must be committed instances of Array, both with a size of
 zero.  When the method completes successfully, aCommittedArray is returned
 and will contain the following values:

	Array[1] = number of candidate objects.
	Array[2] = number of candidates objects that passed validation.
	Array[3] = number of candidates objects that failed validation.
	Array[4] = array of Integer objects.  These are the object ID's
                   of candidate objects that failed validation (subject
                   to errLim; see below).
	Array[5] = number of candidates objects that were recorded dead.
                   This is the number of possible dead objects reported
                   to the stone.
	Array[6] = number of candidates objects that were determined to be
                   not dead because they were referenced by an object which
                   was not a candidate object.  This is the number of not
                   dead candidate objects found during the first phase of
                   the MGC.
	Array[7] = array of not dead candidates objects found during the
                   first phase of the MGC (subject to errLim; see below).
                   Objects in this array are candidate objects directly
                   referenced by one or more objects that are not
                   candidates.
		
 The maximum number of candidate objects returned in elements 3 and 7 is
 determined by the value of errLim.  For example, to store no more than
 100 error objects, set errLim to be 100.  To enable returning all objects,
 set the errLim arg to -1.  A setting of zero will prevent any error
 objects from being stored in the arrays.

 This method may fail with an error if a markForCollection or
 other garbage caCommittedArrayollection operation is in progress.  It will 
 also fail if outstanding possible dead objects from previous garbage 
 collection operations exist.  If this happens the markGcCandidates should
 be retried later.

 This method requires the GarbageCollection privilege.

 This method aborts the current transaction.
	
 A GciHardBreak during this method will terminate the session."

<primitive: 324>
(aFileNameString == nil)
  ifFalse:[aFileNameString _validateClass: String].
aBoolean _validateClass: Boolean.
aBool _validateClass: Boolean.
aCommittedArray _validateClass: Array.
errLim _validateClass: SmallInteger.
(aCommittedArray size == 7)
  ifFalse:[aCommittedArray _error: #rtErrArgOutOfRange ].
(aCommittedArray at: 4)  _validateClass: Array.
(aCommittedArray at: 7)  _validateClass: Array.
self _primitiveFailed: 
     #_markGcCandidatesFromFile:forceOnError:resultArray:errorLimit: 
     args: { aFileNameString . aBoolean . 
             aCommittedArray . errLim . aBool } .
self _uncontinuableError
%

category: 'Private'
method: Repository
cleanupFdcResultsForMgc: forMgc
"Delete the results produced by a previous markGcCandidates operation.
 If the caller intends to run a markGcCandidates, then forMgc should
 be true."

|t |
t := Globals at: #FdcResults otherwise: nil.
(t == nil)
  ifTrue:[
	t := Array new: 4.
        t at: 3 put: { } .
        Globals at: #FdcResults put: t.
  ].

t at: 1 put: 0.
t at: 2 put: 0.
self cleanupBigArray: (t at: 3) forMgc: forMgc.
t at: 4 put: false.
^true
%

! changes to fix 31955 
category: 'Private'
method: Repository
cleanupBigArray: array forMgc: forMgc

"Empty the array.  If forMgc is true, we must set every 
 element to nil to avoid stray references in the 
 repository.  Otherwise we can just size the array to
 zero."

| maxTempBytes elemPerCommit count |
(array == nil) ifFalse:[
  forMgc ifTrue:[
     maxTempBytes := System _tempObjSpaceMax .

        " maxTempBytes * 0.30 / (8 bytes per in-memory element) "
     elemPerCommit := ((maxTempBytes // 10) * 3) // 8 .

     count := 0 .
     1 to: array size do:[:n|
        array at: n put: nil.
        count := count + 1 .
        count >= elemPerCommit ifTrue:[
          System commitTransaction ifFalse:[ self _error: #rtErrGcCommitFailure ].
          count := 0 .
        ]
     ].
     count > 0 ifTrue:[
       System commitTransaction ifFalse:[ self _error: #rtErrGcCommitFailure ].
     ].
  ].
  array size: 0.
].
%

category: 'Private'
method: Repository
cleanupMgcResultsForMgc: forMgc

|t|
t := Globals at: #MgcResults otherwise: nil.
(t == nil)
  ifTrue:[
	t := Array new: 7.
        t at: 4 put: { } .
        t at: 7 put: { } .
        Globals at: #MgcResults put: t.].

t at: 1 put: 0.
t at: 2 put: 0.
t at: 3 put: 0.
self cleanupBigArray: (t at: 4) forMgc: forMgc.
t at: 5 put: 0.
t at: 6 put: 0.
self cleanupBigArray: (t at: 7) forMgc: forMgc.
^true.
%

category: 'Garbage Collection'
method: Repository
basicMarkGcCandidatesFromFile: aFileNameString forceOnError: aBoolean
 readOnly: aBool

"Performs a garbage collection analysis on the objects in the given file by
 scanning the Repository for other references to them.  Objects are marked for
 subsequent reclaiming of storage space if the only references to candidate
 objects are from other candidates.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 See the method Repository>>_markGcCandidatesFromFile: aFileNameString
                            forceOnError: aBoolean
                            resultArray: aCommittedArray
                            errorLimit: errLim
                            readOnly: aBool

 for more information on this method."

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

(System transactionMode == #autoBegin)
  ifFalse:[System transactionMode: #autoBegin].

self cleanupMgcResultsForMgc: true.
System commitTransaction
  ifFalse:[^ self _error: #rtErrGcCommitFailure].
self cleanupFdcResultsForMgc: true.
System commitTransaction
  ifFalse:[^ self _error: #rtErrGcCommitFailure].

^ self _markGcCandidatesFromFile: aFileNameString
      forceOnError: aBoolean
      resultArray: (Globals at: #MgcResults)
      errorLimit: 100
      readOnly: aBool.
%

category: 'Garbage Collection'
method: Repository
markGcCandidatesFromFile: aFileNameString forceOnError: aBoolean

"Performs a garbage collection analysis on the objects in the given file by
 scanning the Repository for other references to them.  Objects are marked for
 subsequent reclaiming of storage space if the only references to candidate
 objects are from other candidates.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 See the method Repository>>_markGcCandidatesFromFile: aFileNameString
                            forceOnError: aBoolean
                            resultArray: aCommittedArray
                            errorLimit: errLim
                            readOnly: aBool

 for more information on this method."

^ self basicMarkGcCandidatesFromFile: aFileNameString forceOnError: aBoolean
 readOnly: false.
%

category: 'Garbage Collection'
method: Repository
markGcCandidatesReadOnlyFromFile: aFileNameString
"Perform a markGcCandidates operation, but do not send the resultant
 collection of dead objects to the stone for reclamation.  In read-only
 mode, the MGC stores the dead objects that would have been passed to the
 stone in hidden set 34 (GcCandidates).
 
 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 For more information on this method, see the comments in the method:
 Repository>>_markGcCandidatesFromFile:
             forceOnError:
             resultArray:
             errorLimit:
             readOnly:
"

^ self basicMarkGcCandidatesFromFile: aFileNameString forceOnError: true
 readOnly: true.
%

category: 'Garbage Collection'
method: Repository
markGcCandidatesFromFile: aFileNameString
"Performs a garbage collection analysis on the objects in the given file by
 scanning the Repository for other references to them.  Objects are marked for
 subsequent reclaiming of storage space if the only references to candidate
 objects are from other candidates.

 Does not proceed if any candidate objects fail validation or were found
 to be not dead in the first phase of the MGC.


 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.

 See the method Repository>>_markGcCandidatesFromFile: aFileNameString
                            forceOnError: aBoolean
                            resultArray: aCommittedArray
                            errorLimit: errLim
                            readOnly: aBool

 for more information on this method."

 ^ self markGcCandidatesFromFile: aFileNameString forceOnError: false.
%

category: 'Repository Usage Reporting'
method: Repository
pagesWithPercentFree: aPercent

"This method returns an Array containing the following statistics:

 1.  The total number of data pages processed.
 2.  The sum in bytes of unused space in all data pages.  This quantity is a
     measure of data fragmentation in the receiver.
 3.  The number of bytes in a page.
 4.  The number of data pages that have at least the specified percentage of
     unused space.
 5.  The number of data pages that have at least the specified percentage of
     unused space and contain only 1 object.
 6.  The number of data pages that contain only 1 object.

 Do not confuse unused space on a page with free space (unused pages) in a
 Repository or Extent.  See the freeSpace and freeSpaceInExtent: methods
 for more information.

 This method requires the GarbageCollection privilege.

 A GciHardBreak during this method will terminate the session."

^ self _pagesWithPercentFree: aPercent doScavenge: false
%

category: 'Repository Usage Reporting'
method: Repository
_pagesWithPercentFree: aPercent doScavenge: aBoolean

"This method returns an Array containing the following statistics:

 1.  The total number of data pages processed.
 2.  The sum in bytes of unused space in all data pages.  This quantity is a
     measure of data fragmentation in the receiver.
 3.  The number of bytes in a page.
 4.  The number of data pages that have at least the specified percentage of
     unused space.
 5.  The number of data pages that have at least the specified percentage of
     unused space and contain only 1 object.
 6.  The number of data pages that contain only 1 object.

 If aBoolean is true, then the primitive will FAIL .  (Gs 64 v2.2 and above)

 Do not confuse unused space on a page with free space (unused pages) in a
 Repository or Extent.  See the 'freeSpace' and 'freeSpaceInExtent:' methods
 for more information.

 In order to execute this method you must have the GarbageCollection
 privilege.

 A GciHardBreak during this method terminates the session."

<primitive: 390>
self _validatePercentage: aPercent .
aBoolean _validateClass: Boolean .
^ self _primitiveFailed: #_pagesWithPercentFree:doScavenge:
       args: { aPercent . aBoolean }
%

category: 'Garbage Collection'
method: Repository
_setGcConfigAt: aSymbol put: aValue

"try 10 times to change the value of the specified GC configuration
 parameter.  If cannot successfully commit within 10 tries, raises an error. 
 Aborts current transaction . Returns the previous
 value of the parameter.  aSymbol must be resolvable in the UserGlobals
 for GcUser. "

| triesLeft commitResult oldVal gcUserUg |

triesLeft := 10 .
commitResult := false .
gcUserUg := ((AllUsers userWithId:'GcUser' ) resolveSymbol:#UserGlobals ) value.
[triesLeft > 0 and:[commitResult == false] ] whileTrue:[
  System beginTransaction .
  oldVal := gcUserUg at: aSymbol .
  gcUserUg at: aSymbol put: aValue .

  commitResult := System commitTransaction  .
  triesLeft := triesLeft - 1.
  commitResult 
    ifFalse:[ System abortTransaction .  System sleep: 2 . ]
    ifTrue:[ GsFile gciLogClient:'--setGcConfig: set ' , aSymbol , ' to ' , 
		aValue asString 
    ].
].
commitResult ifFalse:[ ^ self error:'Unable to modify Gc configuration'].
^ oldVal
%

category: 'Garbage Collection'
method: Repository
reclaimAll

"Explicitly triggers the reclamation of all shadowed and dead objects.  
 The caller can disable the reclamation of dead objects by adjusting
 the Gc configuration parameter #reclaimDeadEnabled .

 Returns an Array of saved Gc configuration values (for compatibility) 
 but automatically runs the postReclaimAll: when reclaiming is done.

 The need for running reclaimAll is reduced by the newer multi-threaded
 implementations of backup, restore and objectAudit.

 Each extent must have a reclaim gem assigned to it when this method
 is called.  The Admin GC session must also be running if there are any 
 possible dead objects in the system.  If not, this method will invoke 
 error #rtErrReclaimAllMissingGcGem.  This method may also hang 
 indefinitely if there are sessions logged in that do not vote. 

 This method acquires the GC lock and returns with the lock held.
 The postReclaimAll: method releases the GC lock.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 This method requires the GarbageCollection and SystemControl privileges."

| gcUserUg saveMinPages saveMode saveCrBacklog newMinPages newCrBacklog
  saveGcConfigs includeDead scCount done voteSt extentsNoReclaim 
  lockGcResult haveGcLock saveReclaimSleepTime saveSleepTimeBetweenReclaim |

System myUserProfile _validatePrivilegeName:#SystemControl .
System myUserProfile _validatePrivilegeName:#GarbageCollection .

"All extents must have a reclaim gems"
extentsNoReclaim := System numberOfExtentsWithoutGC.
(extentsNoReclaim > 0)
  ifTrue:[self _error: #rtErrReclaimAllMissingGcGem].

"Admin gem must be present if voteState is not 0"
((System voteState ~~ 0) and:[System adminGcGemSessionId == 0])
  ifTrue:[self _error: #rtErrReclaimAllMissingGcGem].

gcUserUg := ((AllUsers userWithId:'GcUser' ) resolveSymbol:#UserGlobals ) value.

"Get the GC lock to ensure epoch not in progress.  We might not get it if
 voting is in progress so we will try after voting is finished if
 necessary."
haveGcLock := (System _lockGc: true) == true.

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

saveMode := System transactionMode.
System transactionMode: #manualBegin .

saveCrBacklog := System stoneConfigurationAt:#StnSignalAbortCrBacklog .
newCrBacklog := 3 .
System stoneConfigurationAt:#StnSignalAbortCrBacklog put: newCrBacklog .
GsFile gciLogClient:'--reclaimAll: changed StnSignalAbortCrBacklog from ' 
	, saveCrBacklog asString , ' to ' , newCrBacklog asString.

newMinPages := 1 .
saveMinPages := self _setGcConfigAt: #reclaimMinPages put: newMinPages .
GsFile gciLogClient:'--reclaimAll: changed reclaimMinPages from ' 
	, saveMinPages asString , ' to ' , newMinPages asString .

saveReclaimSleepTime := self _setGcConfigAt: #reclaimSleepTime put: 1 .
saveSleepTimeBetweenReclaim := 
  self _setGcConfigAt: #sleepTimeBetweenReclaimMs put: 0 .

saveGcConfigs := { saveMinPages  .
                   saveReclaimSleepTime  .
                   saveSleepTimeBetweenReclaim } .

gcUserUg := ((AllUsers userWithId:'GcUser' ) resolveSymbol:#UserGlobals ) value.
includeDead := gcUserUg at:#reclaimDeadEnabled .
GsFile gciLogClient:'--reclaimAll: using reclaimDeadEnabled ' , 
 includeDead asString .

" wait for GC voting to complete , since possibleDead are 
  not counted as dead objects that need reclaiming. "
scCount := 0 .
done := false .
[ done ] whileFalse:[
  voteSt := System voteState .
  voteSt == 0 ifTrue:[ 
     done := true  "voteState == VOTE_IDLE, so we are done"  
  ] ifFalse:[ 
    (scCount \\ 10) == 0  ifTrue:[ 
      GsFile gciLogClient:
	 '--reclaimAll: waiting for Gc Voting, voteState ' , voteSt asString 
    ].
    System abortTransaction .
    System sleep: 2 .
  ].
  scCount := scCount + 1 .
].

"Get the GC lock to ensure epoch not in progress"
haveGcLock ifFalse:[
  scCount := 0.
  [(lockGcResult := System _lockGc: true) == true] whileFalse:[
     System sleep: 1.
     System abortTransaction.
     scCount := scCount + 1.
    (scCount \\ 10) == 0  ifTrue:[ 
      GsFile gciLogClient:
	 '--reclaimAll: waiting for GC lock.  reason: ' , lockGcResult asString 
    ].
     (scCount > 60)
       ifTrue:[self postReclaimAll: saveGcConfigs.
               ^lockGcResult _error: #abortErrGarbageCollection args: { 0 }].
  ].
].

" do simple commits to ensure CRs are disposed and reclaim completes"
scCount := 0 .
done := false .
[ done ] whileFalse:[
  System abortTransaction .
  System sleep: 2 .
  done := System _simpleCommitForReclaim: includeDead .
  scCount := scCount + 1 .
  (scCount \\ 10) == 0  ifTrue:[ 
    GsFile gciLogClient:
     '--reclaimAll: simpleCommit loopCount ' , scCount asString
  ].
].

System stoneConfigurationAt:#StnSignalAbortCrBacklog put: saveCrBacklog .
System transactionMode: saveMode .

SystemRepository postReclaimAll: saveGcConfigs.

^ saveGcConfigs 
%

category: 'Garbage Collection'
method: Repository
postReclaimAll: savedGcConfigsArray

"This method should be executed after a reclaimAll to restore
 GC configuration values. The argument should be the result
 from reclaimAll ."
 
| saveMinPages saveReclaimSleepTime saveSleepTimeBetweenReclaimMs |

System myUserProfile _validatePrivilegeName:#GarbageCollection .

saveMinPages := savedGcConfigsArray at: 1 .
saveReclaimSleepTime := savedGcConfigsArray at: 2 .
saveSleepTimeBetweenReclaimMs := savedGcConfigsArray at: 3 .
self _setGcConfigAt: #reclaimMinPages put: saveMinPages .
self _setGcConfigAt: #reclaimSleepTime put: saveReclaimSleepTime .
self _setGcConfigAt: #sleepTimeBetweenReclaimMs put: 
 saveSleepTimeBetweenReclaimMs.
GsFile gciLogClient:'--postReclaimAll: restored reclaimMinPages to ' ,
		saveMinPages asString .
"release GC lock"
System _lockGc: false.
%

category: 'Audit and Repair'
method: Repository
objectAuditWithMaxThreads: maxThreads percentCpuActiveLimit: aPercent

"Checks all objects in GemStone for consistency.

 This method should be executed from topaz -l (the linked version of Topaz).

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 This method requires SystemControl privileges.

 A GciHardBreak during this method will terminate the session.

 This method raises the error #3022 if audit detects problems, otherwise
 returns true."

| errCount |

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

errCount := self _objectAuditWithMaxThreads: maxThreads waitForLock: 60 pageBufSize: 8 
                      percentCpuActiveLimit: aPercent
                      csvFile: nil repair: false.
errCount == 0 ifFalse:[
  self _error: #abortErrObjAuditFail args:{ errCount } .
  self _uncontinuableError
].
^ true.
%

category: 'Audit and Repair'
method: Repository
objectAudit
 "See comments in objectAuditWithMaxThreads:percentCpuActiveLimit:"

^ self objectAuditWithMaxThreads: 2 percentCpuActiveLimit: 90
%


category: 'Audit and Repair'
method: Repository
fastObjectAudit
"Same as the #objectAudit method except the audit is performed 
 aggressively in order to complete in as little time as possible.

 This method may consume most or all host system resources while it
 is in progress."

^ self objectAuditWithMaxThreads: self _aggressiveMaxThreadCount
  percentCpuActiveLimit: 95
%

category: 'Deprecated'
method: Repository
auditWithLimit: sizeLimit

" Deprecated, use #objectAudit instead of this method "

^ self objectAudit
%

category: 'Deprecated'
method: Repository
repairWithLimit: sizeLimit

" Deprecated, use #repair instead of this method."
 
^self repair
%

category: 'Audit and Repair'
method: Repository
repair
" This method requires SystemControl privileges.

  See the comments in: 
    _objectAuditWithMaxThreads:waitForLock:pageBufSize:percentCpuActiveLimit:csvFile:repair: 
  for the kinds of repository errors that can be repaired."


System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^self _objectAuditWithMaxThreads: 2 waitForLock: 60 pageBufSize: 8 
                      percentCpuActiveLimit: 90
                      csvFile: nil repair: true.
%

! fix 41773
category: 'Garbage Collection'
method: Repository
markForCollection
"Performs a garbage collection analysis of all permanent objects on disk.  
 Every object in the receiver that cannot be reached from AllUsers is marked 
 for subsequent reclaiming of storage space. 

 This method aborts the current transaction, empties the GcCandidates queue 
 and commits, runs the analysis while outside of a transaction and then reenters
 a transaction if the session was in a transaction at the start of this method.
 If an abort would cause unsaved changes to be lost, it does not execute and 
 signals an error, #rtErrAbortWouldLoseData.

 When this method completes successfully, signals a Warning and then
 returns the Warning.  The Warning has error text of the form
   Successful completion of Garbage Collection
     found <anInt> live objects,
     found <anInt> dead objects, occupying <anInt> bytes
 This specific Warning will satisfy a topaz 
     expectvalue true
 expected result.

 After this method completes, the GcGem process automatically reclaims the space
 occupied by the dead objects.  Space is not reclaimed until other sessions have
 either committed or aborted the transactions that were concurrent with this
 method.

 This method may fail with an error if another garbage collection is
 in progress.  This method waits up to 2 minutes for the other garbage
 collection to finish before issuing the error.
 Another garbage collection can be in progress because:
   1) an epoch garbage collection is in progress in a gc gem.
   2) a markGcCandidates is in progress in another session
   3) a markForCollection is in progress in another session
   4) a previous epoch, markGcCandidates or markForCollection completed the
     mark phase, but the voting on the possibly dead objects has
     not completed. Note that an admin GcGem must be running for the
     voting to complete.  Also, long running sessions that are idle and
     never aborting will prevent voting from completing.

 This method requires the GarbageCollection privilege.
 A GciHardBreak during this method terminates the session.

 This method performs the MFC in a non-aggressive manner and uses only 2 
 threads.  To complete the operation in less time, see the
 #fastMarkForCollection method.

 See also additional comments in the method:
 #_mfcWithMaxThreads:waitForLock:pageBufSize:percentCpuActiveLimit:"

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^ self markForCollectionWithMaxThreads: 2 waitForLock: 60
%

category: 'Garbage Collection'
method: Repository
markForCollectionWait: waitTimeSeconds

"Perform a markForCollection, waiting up to waitTimeSeconds for any
 garbage collection in progress in another session to complete.

 To wait as long as necessary for the other gc to finish, pass the argument -1.
 This should be done with caution, however, as under certain conditions
 the session could appear to wait forever.  To avoid this you need to:

 1.  Make sure that other sessions are committing/aborting to allow
     voting on possible dead to complete.

 2.  Make sure that the Admin Gc gem and at least one Reclaim Gc gem are
     is running to complete processing of dead objects once the vote is
     completed.

 This method performs the MFC in a non-aggressive manner and uses only 2 
 threads.  To complete the operation in less time, see the
 #fastMarkForCollection method.

 For further documentation see markForCollection. "

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^ self markForCollectionWithMaxThreads: 2 waitForLock: waitTimeSeconds
%


! fix 31566
category: 'Extent Operations
method: Repository
shrinkExtents

"Deprecated in 3.0.
 Truncate all Extents of the Repository to remove internal free space between
 the last used page in each extent and the end of the file containing the
 extent.  Has no effect for extents on a raw disk partition.

 WARNING, you should take a full backup of your extents before 
 executing shrinkExtents.  If the machine crashes (such as a power failure)
 while shrinkExtents is executing, the repository may be unusable.

 Requires that no other user sessions be logged in.
 Stops and restarts the symbol creation session and all gc sessions .
 Because it stops the symbol creation session, this method can only
 be run as SystemUser .

 If DBF_PRE_GROW is enabled in the configuration file then this the extents
 will be grown again the next time Stone is restarted, thus cancelling the
 effect of this method. 

 Returns receiver if successful, otherwise returns a String describing
 sessions that prevented the shrink."

<primitive: 2001> "enter protected mode"
| result prot |
prot := System _protectedMode .
[
  | symStopped gcStopped errCount finallyBlock oldRemoveThresh 
    pagesNeedRemoveOfs removalThreshSym finallyDone |
  self deprecated: 'Repository>>shrinkExtents deprecated in v3.0'.
  symStopped := false .
  gcStopped := false .
  errCount := 0 .
  removalThreshSym := #StnPageMgrRemoveMinPages .
  oldRemoveThresh := System stoneConfigurationAt: removalThreshSym .
  System stoneConfigurationAt: removalThreshSym put: 0 .
  pagesNeedRemoveOfs := System cacheStatisticsDescription indexOf:
   'PagesWaitingForRemovalInStoneCount'.
  finallyDone := false .
  finallyBlock := [
      finallyDone ifFalse:[
        finallyDone := true .
        System stoneConfigurationAt: removalThreshSym put: oldRemoveThresh .    
        symStopped ifTrue:[ 					" fix 31630"
	  GsFile gciLogClient:'--shrink: restarting symbol creation session'.
	  System startSymbolCreationSession .
	  symStopped := false .
        ].
        gcStopped ifTrue:[ 
	  GsFile gciLogClient:'--shrink: restarting gc sessions'.
	  System startAllGcSessions .
	  System waitForAllGcGemsToStartForUpToSeconds:10 .
	  gcStopped := false .
        ].
        self __inProtectedMode .  "gciLog operations will have cleared it"
      ].
  ].
  [
    [ | done sesList count otherSes |
      GsFile gciLogClient:'--shrink: stopping symbol creation session'.
      symStopped := true .
      System stopSymbolCreationSession .   "stop first to fix 31606"
      GsFile gciLogClient:'--shrink: stopping gc sessions'.
      gcStopped := true .
      System stopAllGcSessions .
      done := false .
      count := 1 .
      [ done ] whileFalse:[
	sesList := System currentSessions .
	sesList size == 1 ifTrue:[ 
	  done := true .
	] ifFalse:[
	  count <= 10 ifTrue:[
	    count := count + 1 .
	    System sleep: 1 .
	  ] ifFalse:[
	    otherSes := System otherSessionNames .
	    done := true
	  ].
	].
      ].
      otherSes == nil ifTrue:[ | needRemove |
	GsFile gciLogClient:'--shrink: doing simple commits'.
	3 timesRepeat:[
	  System _simpleCommitForReclaim: false .
	  System abortTransaction .
	  System sleep: 3 .
	  System abortTransaction .
	  System sleep: 3 .
	].
	GsFile gciLogClient:'--shrink: attempting startCheckpointSync '.
	System startCheckpointSync ifFalse:[
	  result := 'ERROR, false from startCheckpointSync. ' .
	].  
	System abortTransaction .
	System sleep: 15 . 
	needRemove := 1 .
	count := 0 .
	GsFile gciLogClient:'--shrink: waiting for pagesNeedRemove=0 ' .
	[ needRemove > 0 and:[count < 10]] whileTrue:[ 
	  System sleep: 3 .
	  System abortTransaction .
	  needRemove := (System cacheStatistics: 1) at: pagesNeedRemoveOfs .
	  count := count + 1
	].
	needRemove > 0 ifTrue:[
	  result == nil ifTrue:[ result := String new ].
	  result addAll: ' ERROR, pagesNeedRemove = ' , needRemove asString .
	].
	self __inProtectedMode .  "gciLog operations will have cleared it"
	self _shrinkExtents .
	result == nil ifTrue:[ result := self ].
      ] ifFalse:[
	result := 'ERROR, sessions preventing shrink:' , otherSes .
      ].
    ] ensure: finallyBlock .
  ] onSynchronous: Error do:[:ex |
    errCount := errCount + 1 .
    " GsFile gciLogClient:'--shrink: error count ' , errCount asString . "
    errCount == 1 ifTrue:[ finallyBlock value ] .
    symStopped := false .
    gcStopped := false .
    ex outer
  ].
] ensure: [
  prot _leaveProtectedMode
].
^ result
%

category: 'Private'
method: Repository
_shrinkExtents

<protected primitive: 427>

self _primitiveFailed: #shrinkExtents .
self _uncontinuableError
%

category: 'Extent Operations'
method: Repository
createExtent: extentFilename

"Creates a new Extent with a file named extentFilename (a String).  The new
 Extent has no maximum size.

 This method updates the DBF_EXTENT_NAMES Stone option.  It does not require
 the system to be in single-user mode.

 If the given file already exists, then this method generates an error and the
 given Extent is not added to the logical Repository.

 Creating an extent with this method bypasses any setting the user may have
 specified for the DBF_PRE_GROW option at system startup.  As extents created
 with this method have no maximum size, their files cannot be grown to an
 initial size.

 If GemStone is being run using weighted disk resource allocation, then
 the new Extent will be given a weight equal to the average weight of all
 other extents.

 This method requires the FileControl privilege.  It is recommended
 that it be run by either DataCurator or SystemUser."

^ self createExtent: extentFilename withMaxSize: 0
%

! document bug 6384
! fix 33802
! fix 40859
category: 'Extent Operations'
method: Repository
createExtent: extentFilename withMaxSize: aSize startNewReclaimGem: aBool

"Creates a new Extent with the given extentFilename (aString) and sets the
 maximum size of that Extent to the the given size.

 If aSize == 0 , then the maximum physical size of the extent is limited
 to disk space or 33554G bytes , whichever comes first . If aSize > 0 ,
 it specifies the maximum physical size of the file in megabytes.

 This method updates the DBF_EXTENT_NAMES and DBF_EXTENT_SIZES Stone options.
 It does not require the system to be in single-user mode.

                                  Note:
   The extent is created with the default ownership and permissions of the
   Stone process.  If this is not the same as the ownership and permissions of
   the other extents, then Unix utilities must be used to change the ownership
   or permissions of the new file; such changes may be made without stopping
   the Stone, and should be made a soon as possible, to avoid other sessions
    encountering authorization errors.

 If the given file already exists, then this method generates an error and the
 given Extent is not added to the logical Repository.

 If the Stone option DBF_PRE_GROW is set to true, then this method will cause
 the newly created extent's file to be grown to the given size.  If the grow
 fails, then this method returns an error and the new Extent is not added to the
 logical Repository.

 If GemStone is being run using weighted disk resource allocation, then
 the new Extent will be given a weight equal to the average weight of all
 other Extents.

 If aBool is true, then a new reclaim gem is started to reclaim pages from the
 newly created extent.  If aBool is false, then no reclaim gem is started.

 This method requires the FileControl privilege.  It is recommended
 that it be run by either DataCurator or SystemUser."

<primitive: 284>

extentFilename _validateClass: String.
aSize _validateClass: SmallInteger.
aBool _validateClass: Boolean .
" aSize out of range errors generated in the primitive "
^ self _primitiveFailed: #createExtent:withMaxSize:startNewReclaimGem:
       args: { extentFilename . aSize . aBool }
%

! document bug 6384
! fix 33802
category: 'Extent Operations'
method: Repository
createExtent: extentFilename withMaxSize: aSize

"Create a new extent with the specified file name and maximum size.
 
 See the comments in the method: 
 Repository>>createExtent:withMaxSize:startNewReclaimGem: for more information
 on this method."

^ self createExtent: extentFilename withMaxSize: aSize startNewReclaimGem: true
%

category: 'Extent Operations'
method: Repository
numberOfExtents

  <primitive: 428>

"Returns the number of active extents."

  self _primitiveFailed: #numberOfExtents .
  self _uncontinuableError
%

category: 'Extent Operations'
method: Repository
validateExtentId: anExtentId

"Returns the argument if it is valid, otherwise generates an error."

| maxExtentId |
anExtentId == nil ifTrue:[ ^ anExtentId ] .
anExtentId _validateClass: SmallInteger.
maxExtentId := self numberOfExtents .
(anExtentId < 1 or:[ anExtentId > maxExtentId]) ifTrue:[
   anExtentId _error: #rtErrArgOutOfRange args:{ 1 . maxExtentId }.
   anExtentId < 1 ifTrue:[ ^ 1 ].
   anExtentId > maxExtentId ifTrue:[ ^ maxExtentId].
].
^ anExtentId .
%

category: 'Listing Instances'
method: Repository
listInstancesToHiddenSet: aClass withMaxThreads: maxThreads maxCpuUsage: aPercentage

"Scans the entire Repository and generates a list of instances of aClass
 into hidden set 1 .  The previous state of the hidden set 1 is cleared.
 Does not include objects in VM's temporary object memory.

 Returns the number of instances found.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData. The entire
 Repository is scanned once."

| aSet |

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

aSet := IdentitySet with: aClass .
^ self _scanPomWithMaxThreads: maxThreads waitForLock: 60 pageBufSize: 8
       percentCpuActiveLimit: aPercentage identSet: aSet limit: 0
       scanKind: 1 toDirectory: nil
%

category: 'Private'
method: Repository
_aggressiveMaxThreadCount

"Answer the number of threads to start for an aggressive, 
 multithreaded repository scan.  The calculation below is really just a
 guess.  We used twice the number of CPU cores on the host since it is expected
 that threads will often be in I/O wait when scanning large repositories."

^ 2 max: ((System maxSessionId - 7) min: ((System hostCpuCount * 2) min: 16))
%

category: 'Listing Instances'
method: Repository
listInstancesToHiddenSet: aClass 

"Scans the entire Repository and generates a list of instances of aClass
 into hidden set 1 .  The previous state of the hidden set ListInstancesResult
 is cleared.  
 Does not include objects in VM's temporary object memory.

 Returns the number of instances found.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData. The entire
 Repository is scanned once.

 The scan is done in an non-aggressive manner and uses only 2 threads.  To 
 complete the operation in less time, see the #fastListInstancesToHiddenSet:
 method."

^ self listInstancesToHiddenSet: aClass withMaxThreads: 2 maxCpuUsage: 90
%

category: 'Listing Instances'
method: Repository
fastListInstancesToHiddenSet: aClass 

"Same as the #listInstancesToHiddenSet: method except the scan is performed 
 aggressively in order to complete in as little time as possible.

 This method may consume most or all host system resources while it
 is in progress."

^ self listInstancesToHiddenSet: aClass 
       withMaxThreads: self _aggressiveMaxThreadCount
       maxCpuUsage: 95
%

category: 'Listing Instances'
method: Repository
listInstances: anArray toDirectory: aString

"Scan the repository for the instances of classes in anArray and write
 the results to binary bitmap files in the directory specified by
 aString.  Binary bitmap files have an extension of .bm and may be
 loaded into hidden sets using class methods in class System.
 aString must specify a path to a writable directory.

 Does not include objects in VM's temporary object memory.

 Bitmap files are named as follows:

 <ClassName>-<classOop>-instances.bm

 where className is the name of the class and classOop is the object
 ID of the class.

 The result is an Array of pairs.  For each element of the argument anArray,
 the result array contains  <aClass>, <numberOfInstancesFound>.
 The numberOfInstances is the total number written to the output bitmap file
 
 Note: the order of classes in the result array may differ from that of 
 the argument anArray.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.  The entire
 Repository is scanned at least once.

 If the argument anArray contains more than 2000 unique elements then
 the entire Repository will be scanned once for each group of 2000 unique
 elements, or fraction thereof.

 The scan is done in an non-aggressive manner and uses only 2 threads.  To 
 complete the operation in less time, see the #fastListInstances:toDirectory:
 method."

aString _validateClass: String.
anArray _validateClass: Array.
^ self listInstances: anArray limit: 0 toDirectory: aString withMaxThreads: 2
 maxCpuUsage: 90 memoryOnly: false
%

category: 'Listing Instances'
method: Repository
fastListInstances: anArray toDirectory: aString

"Same as the #listInstances:toDirectory: method except the scan is performed 
 aggressively in order to complete in as little time as possible,
 and does not include in-memory objects.
 

 This method may consume most or all host system resources while it
 is in progress."

aString _validateClass: String.
anArray _validateClass: Array.
^ self listInstances: anArray limit: 0 toDirectory: aString 
      withMaxThreads: self _aggressiveMaxThreadCount  maxCpuUsage: 95
	memoryOnly: false
%

category: 'Listing Instances'
method: Repository
listInstances: anArray limit: aSmallInt

"Returns a list of instances on the receiver that belong to one of the
 classes listed in anArray. .  Includes in-memory objects.

 aSmallInt is the maximum number of instances to report for each class,
 or 0 for unlimited . 

 The result is an Array of pairs, for each element of the argument anArray,
 the result array contains  <numberOfInstances>, <arrayOfInstances> .
 The numberOfInstances is the total number in the repository .
 The arrayOfInstances will be limited by the argument aSmallInt .

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.  The entire        
 Repository is scanned at least once.

 If the argument anArray contains more than 2000 unique elements then
 the entire Repository will be scanned once for each group of 2000 unique
 elements, or fraction thereof.

 The scan is done in an non-aggressive manner and uses only 2 threads.  To 
 complete the operation in less time, see the #fastListInstances:limit:
 method."

^ self listInstances: anArray limit: aSmallInt toDirectory: nil withMaxThreads: 2
 maxCpuUsage: 90 memoryOnly: false
%

category: 'Listing Instances'
method: Repository
fastListInstances: anArray limit: aSmallInt

"Same as the #listInstances:limit: method except the scan is performed 
 aggressively in order to complete in as little time as possible.

 This method may consume most or all host system resources while it is in progress."

^ self listInstances: anArray limit: aSmallInt toDirectory: nil 
      withMaxThreads: self _aggressiveMaxThreadCount maxCpuUsage: 95
	memoryOnly: false 
%


category: 'Listing Instances'
method: Repository
listInstances: anArray limit: aSmallInt toDirectory: directoryString
 withMaxThreads: maxThreads maxCpuUsage: aPercentage 

 ^ self listInstances: anArray limit: aSmallInt toDirectory: directoryString
   withMaxThreads: maxThreads maxCpuUsage: aPercentage memoryOnly: false
%

method: Repository
listInstances: anArray limit: aSmallInt toDirectory: directoryString
 withMaxThreads: maxThreads maxCpuUsage: aPercentage memoryOnly: memOnlyBool

"If directoryString == nil, result includes in-memory objects.

 If memBool == true, result contains only the in-memory objects,
  and maxThreads and aPercentage  are ignored."

| inputSet resultInSetOrder result code scanBlk |

memOnlyBool ifFalse:[ 
  System needsCommit ifTrue: [ self _error: #rtErrAbortWouldLoseData ]
]. 
inputSet := self _arrayOfClassesAsSet: anArray .
inputSet size < 1 ifTrue:[ ^ { } ] .

memOnlyBool ifFalse:[
 scanBlk := [ :scanSetThisTime |
     self _scanPomWithMaxThreads: maxThreads waitForLock: 60 pageBufSize: 8
                             percentCpuActiveLimit: aPercentage  
                             identSet: scanSetThisTime limit: aSmallInt
                             scanKind: code toDirectory: directoryString 
  ].
].
code := directoryString ifNotNil:[ 2 ] ifNil:[ 0 ].
inputSet := IdentitySet withAll: anArray .

resultInSetOrder:= self _doListInstancesFrom: inputSet with: scanBlk 
			includeMemory:  directoryString == nil .

directoryString ifNotNil:[
  result := resultInSetOrder "primitive wrote the files, so we are done"
] ifNil:[ | inputArraySize |
  inputArraySize := anArray size .
  result := Array new: inputArraySize * 2.
  1 to: inputArraySize do:[:j| | soIdx resIdx anObj |
      anObj := anArray at: j .
      soIdx := (inputSet _offsetOf: anObj) * 2 .
      resIdx := j * 2 .
      result at: resIdx - 1 put: ( resultInSetOrder at: soIdx - 1 ) ."totalCount"
      result at: resIdx put: ( resultInSetOrder at: soIdx) . "array of instances"
  ].
].
^ result
%

category: 'Private'
method: Repository
_arrayOfClassesAsSet: anArray
 | set |
  set := IdentitySet withAll: anArray .
  set do: [:each | each _validateClass: Behavior].
  ^ set
%

category: 'Private'
method: Repository
_doListInstancesFrom: inputSet with: diskBlock includeMemory: inMemoryBool
| inputSetSize result inputIdx |

inputIdx := 1 .
inputSetSize := inputSet size .
result := { }  .  "pairs  size, array of instances, in order of input set"
[ inputIdx <= inputSetSize] whileTrue:[ | scanSetThisTime idx |
   scanSetThisTime := IdentitySet new .
   [ scanSetThisTime size < 2000 and:[inputIdx <= inputSetSize ]] whileTrue:[
      scanSetThisTime add: (inputSet _at: inputIdx) .
      inputIdx := inputIdx + 1 .
    ].
   idx := result size + 1 .
   diskBlock ifNotNil:[
     result addAll: (diskBlock value: scanSetThisTime ).
   ].
   inMemoryBool ifTrue:[ | memResult |
     memResult := self _listInstancesInMemory: (Array withAll: scanSetThisTime).
     1 to: memResult size do:[:j | | elem memRes |
       memRes := memResult at: j .
       (elem := result atOrNil: idx) ifNil:[
          result at: idx put: memRes size ;
                 at: idx + 1 put: memRes .
       ] ifNotNil:[ | ne |
          elem := result at: idx + 1. 
          (elem := IdentitySet withAll: elem) addAll: memRes .
          result at: idx put: elem size ;
                 at: idx + 1 put: elem asArray .
       ].
       idx := idx + 2
     ].
   ]
 ].
^ result
%

category: 'Listing Instances'
method: Repository
listInstances: anArray

"Returns a list of instances on the receiver that belong to one of the
 classes listed in anArray.  The result of this method is an Array of Arrays,
 where the contents of each inner array consists of all instances 
 whose class is equal to the corresponding element in anArray.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 If anArray contains multiple occurrences of a class, then the result
 will contain corresponding multiple occurrences of the same Array that lists
 the instances of that class.

 If anArray contains an element that is not a kind of Behavior, an error
 is generated.

 Scans the entire Repository at least once.

 If the argument anArray contains more than 2000 unique elements then
 the entire Repository will be scanned once for each group of 2000 unique
 elements, or fraction thereof.

 The scan is done in an non-aggressive manner and uses only 2 threads.  To 
 complete the operation in less time, see the #listInstances:limit:
 method."

| reportArr res |
reportArr := self listInstances: anArray limit: 0 .
res := Array new: anArray size .
1 to: anArray size do:[:j |
  res at:j put: (reportArr at: j * 2 ) . "array of instances"
].
^ res
%

category: 'Listing Instances'
method: Repository
fastListInstances: anArray

"Same as the #listInstances: method except the scan is performed 
 aggressively in order to complete in as little time as possible.

 This method may consume most or all host system resources while it
 is in progress."

| reportArr res |
reportArr := self fastListInstances: anArray limit: 0 .
res := Array new: anArray size .
1 to: anArray size do:[:j |
  res at:j put: (reportArr at: j * 2 ) . "array of instances"
].
^ res
%

method: Repository
listInstancesInMemory: anArray

"Returns a list of instances of the classes specified in anArray
 that are found in temporary object memory.

 The result of this method is an Array of Arrays, where the 
 contents of each inner array consists of all instances
 of the corresponding element in anArray.

 The result contains in-memory objects only.  Objects which are not
 currently in memory are not analyzed.

 Does not abort the current transaction."

| pairs res |
pairs := self listInstances: anArray limit: 0 toDirectory: nil 
      withMaxThreads: 1 maxCpuUsage: 95 memoryOnly: true .
res := Array new: anArray size .
1 to: anArray size do:[:j |
  res at:j put: (pairs at: j * 2 ) . "array of instances"
].
^ res
%

category: 'Private'
method: Repository
_listInstancesInMemory: sortedArray

"sortedArray must be sorted by identity.
 Returns an Array, each element is an Array of instances for
 corresponding element of sortedArray."

<primitive: 389>
sortedArray _validateClass: Array.
(sortedArray size > 2034 "virtual machine constant")
  ifTrue:[ sortedArray _error: #errArgTooLarge args:{ 2034 } ] .
^ self _primitiveFailed: #_listInstancesInMemory: args: { sortedArray } 
%

category: 'Listing References'
method: Repository
listReferencesInMemory: anArray

"Returns a list of instances in temporary object memory that 
 have a reference to one of the objects specified in anArray.  
 The result of this method is an Array of Arrays, where the 
 contents of each inner array consists of all instances
 that have a reference to the corresponding element in anArray.
 
 The result contains in-memory objects only.  Objects which are not
 currently in memory are not analyzed.

 Does not abort the current transaction."

<primitive: 393>
anArray _validateClass: Array.
(anArray size > 2034 "virtual machine constant")
  ifTrue:[ anArray _error: #errArgTooLarge args:{ 2034 } ] .
^ self _primitiveFailed: #listReferencesInMemory: args: { anArray } 
%

category: 'Class Management'
method: Repository
_loadedClasses: includeModulesBoolean

"Returns an Array of all Classes currently loaded in memory.
 If includeModulesBoolean==true, the result also includes instances of Module.
 Result does not include meta classes, meta modules, or Ruby virtual classes."

<primitive: 869> 
includeModulesBoolean _validateClass: Boolean .
self _primitiveFailed: #_loadedClasses: args: { includeModulesBoolean } 
%

category: 'Listing References'
method: Repository
listReferences: anArray 

"Returns a list of instances in the Repository that have a reference to one of
 the objects specified in anArray.  The result of this method is an Array of
 Arrays, where the contents of each inner array consists of all instances 
 that have a reference to the corresponding element in anArray.  

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 The result contains only committed objects. 

 The scan is done in an non-aggressive manner and uses only 2 threads.  To 
 complete the operation in less time, see the #fastListReferences
 method."

^ self listReferences: anArray withLimit: 0 withMaxThreads: 2 maxCpuUsage: 90
%

category: 'Listing References'
method: Repository
listReferences: anArray withLimit: aSmallInt

"Returns a list of instances in the Repository that have a reference to one of
 the objects specified in anArray.  The result of this method is an Array of
 Arrays, where the contents of each inner array consists of all instances 
 that have a reference to the corresponding element in anArray.  
 The number in each inner array is limited to approximately aSmallInt.  

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 The result contains only committed objects. 

 The scan is done in an non-aggressive manner and uses only 2 threads.  To 
 complete the operation in less time, see the #fastListReferences
 method."

^ self listReferences: anArray withLimit: aSmallInt withMaxThreads: 2 maxCpuUsage: 90
%

category: 'Listing References'
method: Repository
fastListReferences: anArray 

"Same as the #listReferences: method except the scan is performed 
 aggressively in order to complete in as little time as possible.

 This method may consume most or all host system resources while it is in progress."

^ self listReferences: anArray withLimit: 0 
       withMaxThreads: self _aggressiveMaxThreadCount maxCpuUsage: 95
%
      
category: 'Listing References'
method: Repository
listReferences: anArray withLimit: aSmallInt withMaxThreads: maxThreads
 maxCpuUsage: aPercentage

"Returns a list of instances in the Repository that have a reference to one of
 the objects specified in anArray.  The result of this method is an Array of
 Arrays, where the contents of each inner array consists of instances 
 that have a reference to the corresponding element in anArray.  
 The number in each inner array is limited to approximately aSmallInt.  

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 The result contains both permanent and temporary objects.  The temporary
 objects found may vary from run to run.

 This method is very expensive.  It scans the entire Repository and looks
 at every instance variable of every object."

| outerArr innerArr |

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

outerArr := self _listReferencesWithMaxThreads: maxThreads waitForLock: 60
      pageBufSize: 8 percentCpuActiveLimit: aPercentage
      targets: anArray withLimit: aSmallInt.
aSmallInt > 0 ifTrue:[
  1 to: outerArr size  do:[ :j|
    innerArr := outerArr at: j .
    innerArr size > aSmallInt ifTrue:[ innerArr size: aSmallInt ]. 
  ].
].
^ outerArr.
%
      
category: 'Private'
method: Repository
_listReferencesWithMaxThreads: maxThreads waitForLock: lockWaitTime 
      pageBufSize: aBufSize percentCpuActiveLimit: aPercent
      targets: anArray withLimit: aSmallInt

"Private. Returns a list of instances in the Repository that have a 
 reference to one of the objects specified in the targets array.  The 
 result of this method is an  Array of Arrays, where the contents of 
 each inner array consists of instances that have a reference to the 
 corresponding element in anArray.  

 This method aborts the current transaction.

 The number in inner array is limited to approximately aSmallInt,
 and may exceed aSmallInt by the maximum number of objects in a data page.
 If aSmallInt is <= 0, the result size is unlimited. 

 The result contains only permanent objects.

 This primitive method performs a multithreaded scan of the Repository.

 See _scanPomWithMaxThreads primitive for a description of the multithreaded args
"

<primitive: 899>
 maxThreads _validateClass: SmallInteger.
 lockWaitTime _validateClass: SmallInteger.
 aBufSize _validateClass: SmallInteger ; _validateIsPowerOf2 .
 self _validatePercentage: aPercent .
 aSmallInt _validateClass: SmallInteger.
 anArray _validateClass: Array.
 (anArray size > 2034 "virtual machine constant")
  ifTrue:[ anArray _error: #errArgTooLarge args:{ 2034 } ] .
 ^ self _primitiveFailed: 
    #_listReferencesWithMaxThreads:waitForLock:pageBufSize:percentCpuActiveLimit:targets:withLimit:
    args: {  maxThreads . lockWaitTime . aBufSize . 
             aPercent . anArray . aSmallInt }
%

category: 'Deprecated'
method: Repository
listObjectsInSegments: anArray 
"Deprecated"

self deprecated: 'Repository>>listObjectsInSegments: deprecated v3.0. 
Use #listObjectsInObjectSecurityPolicies: instead.'.
^ self listObjectsInObjectSecurityPolicies: anArray 
%

category: 'Listing By Security Policy'
method: Repository
listObjectsInObjectSecurityPolicies: anArray 

"Returns a list of objects in the receiver that have the specified 
 objectSecurityPolicyIds.  The result of this method is an Array
 of Arrays, where the contents of each inner array consists of
 objects whose objectSecurityPolicyId is the objectSecurityPolicyId of
 of the corresponding element in anArray.

 anArray must be an Array of at most 2034 objectSecurityPolicyIds 

 The value zero can be used to indicate you wish to get the objects that
 have no objectSecurityPolicyId.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 The result contains both permanent and temporary objects.  The temporary
 objects found may vary from run to run.

 The scan is done in an non-aggressive manner and uses only 2 threads.  To 
 complete the operation in less time, see the #fastListInstances:toDirectory:
 method."

^ self listObjectsInObjectSecurityPolicies: anArray withLimit: 0
%

category: 'Listing By Security Policy'
method: Repository
fastListObjectsInObjectSecurityPolicies: anArray 

"Same as the #listObjectsInObjectSecurityPolicies: method except the scan
 is performed aggressively in order to complete in as little time as possible.

 This method may consume most or all host system resources while it is
 in progress."

^ self fastListObjectsInObjectSecurityPolicies: anArray withLimit: 0
%

category: 'Deprecated'
method: Repository
_validateSegmentId: anObjectSecurityPolicyId 
"Deprecrated, use #_validateObjectSecurityPolicyId: instead of this method."

^ self _validateObjectSecurityPolicyId: anObjectSecurityPolicyId 
%

category: 'Private'
method: Repository
_validateObjectSecurityPolicyId: anObjectSecurityPolicyId 

"Generates an error if anObjectSecurityPolicyId is not a SmallInteger
 >= 0 and <= self size .  
 Note that objectSecurityPolicyId 0 specifies world write,
 and does not have an associated GsObjectSecurityPolicy . "
| sz |
anObjectSecurityPolicyId _validateClass: SmallInteger .
(anObjectSecurityPolicyId < 0 or:[ anObjectSecurityPolicyId > (sz := self size) ]) 
  ifTrue:[anObjectSecurityPolicyId _error: #rtErrArgOutOfRange args:{ 0 . sz }
].
% 

category: 'Deprecated'
method: Repository
listObjectsInSegmentToHiddenSet: anObjectSecurityPolicyId
"Deprecated"

self deprecated: 'Repository>>listObjectsInSegmentToHiddenSet: deprecated v3.0. 
Use #listObjectsInObjectSecurityPolicyToHiddenSet: instead.'.

^ self listObjectsInObjectSecurityPolicyToHiddenSet: anObjectSecurityPolicyId
%

category: 'Listing By Security Policy'
method: Repository
listObjectsInObjectSecurityPolicyToHiddenSet: anObjectSecurityPolicyId

"Scans the entire Repository and generates a list of objects that have specified 
 objectSecurityPolicyId into hidden set 1 .  The previous state of hidden set 1
 is cleared.

 The value zero can be used to indicate you wish to get the objects that
 have no objectSecurityPolicyId.

 Returns the number of instances found.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData. The entire
 repository is scanned once.

 The scan is done in an non-aggressive manner and uses only 2 threads.  To 
 complete the operation in less time, see the 
 #fastListObjectsInObjectSecurityPolicyToHiddenSet: method."

self _validateObjectSecurityPolicyId: anObjectSecurityPolicyId .
^ self _listObjectsInObjectSecurityPolicies: { anObjectSecurityPolicyId }  
       limit: 0 scanKind: 5 toDirectory: nil withMaxThreads: 2 maxCpuUsage: 90
%

category: 'Listing By Security Policy'
method: Repository
fastListObjectsInObjectSecurityPolicyToHiddenSet: anObjectSecurityPolicyId

"Same as the #listObjectsInObjectSecurityPolicyToHiddenSet: method except the scan is
 performed aggressively in order to complete in as little time as possible.

 This method may consume most or all host system resources while it is in progress."

self _validateObjectSecurityPolicyId: anObjectSecurityPolicyId .
^ self _listObjectsInObjectSecurityPolicies: { anObjectSecurityPolicyId }  
       limit: 0 scanKind: 5 toDirectory: nil 
       withMaxThreads: self _aggressiveMaxThreadCount maxCpuUsage: 95
%

category: 'Deprecated'
method: Repository
listObjectsInSegments: anArray toDirectory: aString
"Deprecated"

self deprecated: 'Repository>>listObjectsInSegment:toDirectory: deprecated v3.0. 
Use #listObjectsInObjectSecurityPolicies:toDirectory: instead.'.

^ self listObjectsInObjectSecurityPolicies: anArray toDirectory: aString.
%

category: 'Listing By Security Policy'
method: Repository
listObjectsInObjectSecurityPolicies: anArray toDirectory: aString

"Scan the repository for objects that have the specified objectSecurityPolicyIds
 and writes the results to binary bitmap files in the directory specified by
 aString.  Binary bitmap files have an extension of .bm and may be
 loaded into hidden sets using class methods in class System.
 aString must specify a path to a writable directory.

 anArray must be an Array of at most 2034 objectSecurityPolicyIds. 

 The value zero can be used to indicate you wish to get the objects that
 have no objectSecurityPolicyId.

 Bitmap files are named as follows:  
   segment<objectSecurityPolicyId>-objects.bm
 where objectSecurityPolicyId is an element of anArray .

 The result is an Array of pairs.  For each element of the argument anArray,
 the result array contains  <objectSecurityPolicyId>, <numberOfInstancesFound>.
 The numberOfInstances is the total number written to the output bitmap file.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData."

aString _validateClass: String .
^ self _listObjectsInObjectSecurityPolicies: anArray limit: 0 scanKind: 6
       toDirectory: aString withMaxThreads: 2 maxCpuUsage: 90
%

category: 'Listing By Security Policy'
method: Repository
fastListObjectsInObjectSecurityPolicies: anArray toDirectory: aString

"Same as the #listObjectsInObjectSecurityPolicies:toDirectory: method
 except the scan is performed aggressively in order to complete in as
 little time as possible.

 This method may consume most or all host system resources while it is in progress."

aString _validateClass: String .
^ self _listObjectsInObjectSecurityPolicies: anArray limit: 0 scanKind: 6
       toDirectory: aString withMaxThreads: self _aggressiveMaxThreadCount
       maxCpuUsage: 95
%

category: 'Deprecated'
method: Repository
listObjectsInSegments: anArray withLimit:  aSmallInt
"Deprecrated"

  self deprecated: 'Repository>>listObjectsInSegments:withLimit: deprecated v3.0. 
Use #listObjectsInObjectSecurityPolicies:withLimit: instead.'.
^ self listObjectsInObjectSecurityPolicies: anArray withLimit:  aSmallInt
%

category: 'Listing By Security Policy'
method: Repository
listObjectsInObjectSecurityPolicies: anArray withLimit:  aSmallInt

"Returns a list of objects in the receiver that have the specified
 objectSecurityPolicyIds.  The result of this method is an Array
 of Arrays, where the contents of each inner array consists of
 objects whose objectSecurityPolicyId is the objectSecurityPolicyId of
 of the corresponding element in anArray.  The number in each inner
 array is limited to approximately aSmallInt.  

 anArray must be an Array of at most 2034 objectSecurityPolicyIds

 The value zero can be used to indicate you wish to get the objects that
 have no objectSecurityPolicyId.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 The result contains both permanent and temporary objects.  The temporary
 objects found may vary from run to run.

 This method is very expensive.  It scans the entire Repository and looks
 at the object header of every object."

^ self _listObjectsInObjectSecurityPolicies: anArray limit: aSmallInt 
      scanKind: 4 toDirectory: nil withMaxThreads: 2 maxCpuUsage: 90
%


category: 'Listing By Security Policy'
method: Repository
fastListObjectsInObjectSecurityPolicies: anArray withLimit:  aSmallInt

"Same as the #listObjectsInObjectSecurityPolicies:withLimit: method except
 the scan is performed aggressively in order to complete in as little time
 as possible.

 This method may consume most or all host system resources while it is in progress."

^ self _listObjectsInObjectSecurityPolicies: anArray limit: aSmallInt 
      scanKind: 4 toDirectory: nil withMaxThreads: self _aggressiveMaxThreadCount
      maxCpuUsage: 95
%

category: 'Private'
method: Repository
_listObjectsInObjectSecurityPolicies: anArray limit: aLimit scanKind: kind
 toDirectory: aStringOrNil withMaxThreads: maxThreads maxCpuUsage: aPercentage

 "aLimit is a SmallInteger, max number of objects to report, or 0 for unlimited.

 anArray must be an Array of at most 2034 objectSecurityPolicyIds. "
| sortedSegIds inputSegIds primResult primResultOfs result |

System needsCommit ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

sortedSegIds := SortedCollection new .
inputSegIds := { }  .
anArray _validateClass: Array .
anArray size > 2034 ifTrue:[ anArray _error: #errArgTooLarge args:{ 2034 } ].
1 to: anArray size do:[ :k | |segId |
  segId := anArray at: k .
  self _validateObjectSecurityPolicyId:  segId .
  inputSegIds add: segId .
  sortedSegIds add: segId
].
primResult := self _scanPomWithMaxThreads: maxThreads waitForLock: 90 pageBufSize: 8 
                percentCpuActiveLimit: aPercentage
                identSet: (IdentitySet withAll: sortedSegIds)
                limit: aLimit scanKind: kind toDirectory: aStringOrNil .

kind == 4 ifFalse:[ ^ primResult ].
result := Array new: anArray size * 2 .

primResultOfs := 1 .
1 to: sortedSegIds size do:[:segIdOfs | | inOfs |
  inOfs := inputSegIds indexOfIdentical: (sortedSegIds at: segIdOfs) .
  result at: (inOfs * 2) -1  put: (primResult at: primResultOfs ) .
  result at: (inOfs * 2)   put: (primResult at: primResultOfs + 1) .
  primResultOfs := primResultOfs + 2 .
].
^ result
%


! _replaceAllObsSym deleted

category: 'Storing and Loading'
method: Repository
writeTo: aPassiveObject

"Instances of Repository cannot be converted to passive form.  This method
 writes nil to aPassiveObject and stops GemStone Smalltalk execution with a
 notifier."

aPassiveObject writeObject: nil.
self _error: #rtErrAttemptToPassivateInvalidObject
%

category: 'Formatting'
method: Repository
printOn: aStream

"Puts a displayable representation of the receiver on the given stream."

"Copy the implementation from Object so we don't inherit it from Collection."

aStream nextPutAll: self asString
%

category: 'Private'  
method: Repository 
_contentsOfPage: aPage
 
"If aPage is an invalid page ID, then the primitive fails.  Returns an Array of
 size at least 5 for all kinds of pages.  For data pages, the Array size is at
 least 15 and varies depending upon the number of objects in the page.

 The aPage argument should be a positive Integer specifying a valid page ID.
  
 The contents of the returned Array are as follows:

    index       description
     1.         pageKind
     2.         sessionId for user that wrote the page.
     3.         low beginId  (transaction begin ID)
     4.         high beginId
     5.         bytesUsed
                     (elements 6 and above are only present for data pages )
     6.         nil  (was approx page creation time (a DateTime) in gs64v1.0)
     7.         clusterId
     8.         freeBytes
     9.         numFreeChunks  
    10.         sizeLargestFree
    11.         numUsedBytes   
    12.         numberOfObjects
    13.         maxObjSize
    14.         minObjSize
    15.         numberOfValidObjects  
    16.         numberOfInvalidObjects
    (17.. 17 + numberOfValidObjects - 1) list of valid objects  
    (17 + numberOfValidObjects ...)  list of OOPs of invalid objects.  "

<primitive: 399>
self _primitiveFailed: #_contentsOfPage: args: { aPage }
%

! deleted transformAllObsSym: , fix 41974

! category: 'Repository Conversion'
! method: Repository
! createObjListFrom: fileName ofKind: className
! 
! "Private. Returns an Array of (valid) objects based on the list of
! OOPs contained in the text file fileName. The objects must all be of
! kind className."
! 
! <primitive: 478>
! self _primitiveFailed: #createObjListFrom:ofKind: 
!      args: { fileName . className } .
! self _uncontinuableError
! %

! category: 'Repository Conversion'
! method: Repository
! createObjListFrom: fileName ofClass: className
! 
! "Private. Returns an Array of (valid) objects based on the list of
! OOPs contained in the text file fileName. The objects must all be of 
! class className."
! 
! <primitive: 481>
! self _primitiveFailed: #createObjListFrom:ofClass: 
!      args: { fileName . className } .
! self _uncontinuableError
! %

! deleted _getContainerList: obsSymSet result: anArray

! deleted _replaceAllSymIn: containerList obsSymList: oldSymList
! mappingDict: mappingDict

category: 'Private'
method: Repository
_shadowPagesByExtent

"This method aborts the current transaction."

<primitive: 526>
self _primitiveFailed: #_shadowPagesByExtent
%

category: 'Private'
method: Repository
_shadowPagesCount

"This method aborts the current transaction."

<primitive: 527>
self _primitiveFailed: #_shadowPagesCount
%

category: 'Private'
method: Repository
_deadNotReclaimedCount

"Return the number of dead objects pending reclamation"

^System _deadNotReclaimedCount
%

category: 'Private'
method: Repository
findObjsConnectedTo: anArray
"Return an array of all objects reachable from the objects in the given 
 array.  Objects implicitly referenced from the header like the class are 
 not included in the sweep.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData."         

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

System _hiddenSetReinit: 26.
System _addAll: anArray to: 26.
^ self _findConnectedObjs.
%

category: 'Private'
method: Repository
_findConnectedObjs
"This method aborts the current transaction."
<primitive: 560>
^ self _primitiveFailed: #_findConnectedObjs
%

category: 'Private'
method: Repository
_findOopsOnPages: anArrayOfPageIds
"Scan the object table and collect all OOPs that reference any page
 in the input array of page IDs."
<primitive: 561>
self _primitiveFailed: #_findOopsOnPages: args: { anArrayOfPageIds }
%

category: 'Private'
method: Repository
_pageForOop: anInteger
"Returns the page ID for the given object ID by looking in the
 object table.  A result of -1 indicates the object ID is a 
 free oop, or the object ID is not in the object table."
<primitive: 562>
self _primitiveFailed: #_pageForOop: args: { anInteger }
%
category: 'Private'
method: Repository
_findPagesContainingOops: anArrayOfInts
"Returns an array of page ID's that contain the oops in the input
 array.  The input array must contain the oop as a SmallInteger or
 LargeInteger, not the object itself.  

 This method does not look at the object table, rather is scans
 the data pages directly from the extents.  Note that the object
 may appear on more than 1 page if it is shadowed."
<primitive: 563>
self _primitiveFailed: #_findPagesContainingOops: args: { anArrayOfInts }
%

category: 'Private'
method: Repository
_oopIsDead: anInteger
"Returns a boolean indicating if the given oop is present in the
 stone's dead object set.  Primitive fails if the integer is not
 a valid disk oop value."

<primitive: 564>
self _primitiveFailed: #_oopIsDead: args: { anInteger } 
%

category: 'Deprecated'
method: Repository
quickObjectAuditWithLevel: anInt

" Deprecated, use #fastObjectAudit instead of this method."

^ self fastObjectAudit
%


! changes to fix 32188
category: 'Repository Analysis'
method: Repository
buildLimitSetForRefPathScan

"Build a list of objects to serve as the limit set for a reference path scan.
 The reference path scan will end when a reference from any objects
 returned by this method is found during the scan."

|set symUser |

set := IdentitySet new.
set addAll: SystemRepository. "Add all objectSecurityPolicies"
"Get all symbol list dictionaries and all versions of all classes from all user profiles.
 Using a set prevents duplicates."
AllUsers do:[:aUserPro| |symbolList|
  set add: aUserPro .
  symbolList := aUserPro symbolList.
  1 to: symbolList size do:[:n| |eachSymbolListDictionary|
    eachSymbolListDictionary := symbolList at: n.
    (set includes: eachSymbolListDictionary) ifFalse:[ 
      set add: eachSymbolListDictionary.
      eachSymbolListDictionary keysAndValuesDo:[:k :v|
        v isBehavior ifTrue:[
          set add: v .
          set add: v class .
	  1 to: v classHistory size do:[:p| |class|
	    class := v classHistory at: p.
            class ~~ v ifTrue:[
	      set add: class.
	      set add: class class
            ]
          ]
        ].
      ].
    ].
  ].
].
set add: (Globals at: #SharedDependencyLists).

	"gemstone64, AllSymbols is now in SymbolUser's UserGlobals . "
symUser := AllUsers userWithId:'SymbolUser' .
set add: (symUser resolveSymbol: #AllSymbols) value .

^set asArray
%

category: 'Repository Analysis'
method: Repository
findReferencePathToObject: anObject

"Finds a reference path to a single object.  See the comments in method
 Repository>>findReferencePathToObjects:findAllRefs:printToLog: for more
 information.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData."

|result limitArray|
limitArray := self buildLimitSetForRefPathScan.
result := self findReferencePathToObjs: { anObject } limitObjArray: limitArray
                findAllRefs: true printToLog: true.
result == nil
	ifTrue:[^nil].
^ result first.
%
! fix 41770: add GsSecurityPolicy to the list of objects automatically considered as part of the limitArray
category: 'Repository Analysis'
method: Repository
findReferencePathToObjects: anArray findAllRefs: findAllRefs printToLog: printToLog

"Finds a reference path to an array of objects.  Returns an array of arrays showing how each
 object is connected to the repository by reference.  The first element of a reference path will
 usually (see below) be one of the objects in the limit set, which is computed by the method
 Repository>>buildLimitSetForRefPathScan.   The last object in a reference path
 will be anObject.

 NOTE: this method is not intended to work for classes or metaclasses.

 Only a single reference path to each object is returned. It's possible for any object to be
 connected to the repository through multiple reference paths.

 Objects in the stone's list of dead objects are excluded from the scan.  It is an error if
 any input object is in the stone's dead objects set, in which case this method will return nil.

 It is an error if any input object is not committed.  It is also an error if any object is in
 both the search array and the limit set.  nil will be returned in both of these cases.

 It is possible for an object to be disonnected from the repository and therefore have no reference path.
 In this case, the second element of the result array will be false, indicating the reference path was not
 found.

 If findAllRefs is false, the scan will return when the first complete reference path is found or all objects
 are determined to be dead.  If it is true, the scan will continue until all reference paths are found.

 If printToLog is true, messages showing progress of the scan will be printed to the stdout of the gem
 calling this method.  

 The progressCount statistic indicates how many objects have been processed for each scan and is reset
 to zero at the end of each scan.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.         

 This method operates in a transaction and will begin one if necessary.  This method may run
 for long periods of time and should not be called in production repositories.

 A reference path scan for an object will be terminated and considered completed if any class object 
 or GsObjectSecurityPolicy is found in the reference path, even if that object is not otherwise present in
 the limit set.  Normally all classes are connected to the repository and will be included in
 the limit set.  However, for garbage collection purposes, disconnected classes are
 considered live objects if any other object references them.  In other words, a class is a live object
 if any live instances of that class exist in the repository.

 Returns an array of arrays, or nil on error.  Each element of the returned array corresponds to 
 an object in the input array and has the following format:
 1 - the search object
 2 - aBoolean - true if the reference path was found, false if not.
 (reference path follows only if element 2 is true)
 3 - first element of the reference path (this object will be normally be in the limit set)
 4 - second element of the reference path.
 ...
 n - last element of the reference path, which will be the search object."

| limitArray |
limitArray := self buildLimitSetForRefPathScan.
^ self findReferencePathToObjs: anArray limitObjArray: limitArray findAllRefs: findAllRefs printToLog: printToLog.
%
category: 'Repository Analysis'
method: Repository
findReferencePathToObjs: anArray limitObjArray: limitArray findAllRefs: findAllRefs printToLog: printToLog

 "Finds an array of reference paths to objects.  See the comments in method
 Repository>>findReferencePathToObjects:findAllRefs:printToLog: for more
 information.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData."
System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^ self _findReferencePathToObjs: anArray limitObjArray: limitArray findAllRefs: findAllRefs printToLog: printToLog
%

category: 'Repository Analysis'
method: Repository
_findReferencePathToObjs: anArray limitObjArray: limitArray findAllRefs: aBoolean printToLog: anotherBoolean

 "Finds an array of reference paths to objects.  See the comments in method
 Repository>>findReferencePathToObjects:findAllRefs:printToLog: for more
 information.

 This method aborts the current transaction."

<primitive: 568>
anArray _validateClass: Array.
limitArray _validateClass: Array.
aBoolean _validateClass: Boolean.
anotherBoolean _validateClass: Boolean.
(anArray size ==  0)
  ifTrue:[ ^anArray  _error: #errArgTooSmall args:{ 1 }  ] .
(limitArray size ==  0)
  ifTrue:[ ^limitArray _error: #errArgTooSmall args:{ 1 }  ] .
^ self _primitiveFailed: 
       #_findReferencePathToObjs:limitObjArray:findAllRefs:printToLog: 
       args: { anArray . limitArray . aBoolean . anotherBoolean }
%

category: 'Private'
method: Repository
_makePageWithIdScavengable: anInt

"Private, this method should only be used when instructed to do so by
 VMware, Inc. personnel.

 Add the page ID represented by anInt to the stone's list of scavengable
 pages.  anInt must be either a SmallInteger or LargeInteger
 representing a valid data page ID.

 Returns true if the page was successfully added.  Returns false if the
 page is already on the scavengable pages list.  Returns nil if the page
 does not exist or is not a valid data page ID.  This method causes the
 system to do 2 or more commits and a checkpoint.  

 Only SystemUser is allowed to execute this method.
 After executing this method ,   SystemRepository reclaimAll
 should be used to wait for the page to be reclaimed. 

 This method will fail if another session holds the commitToken,
 or if checkpoints are suspended. "

<primitive: 569>
self _primitiveFailed: #_makePageWithIdScavengable: args: { anInt }
%

category: 'Class Membership'
method: Repository
species

"Returns a class similar to, or the same as, the receiver's class which
 can be used for containing derived copies of the receiver."

^ Array 
%

category: 'Class Membership'
method: Repository
speciesForSelect

"Returns a class similar to, or the same as, the receiver's class which
 can be used for containing derived copies of the receiver for select:
 and reject: queries."

^ Array
%
category: 'Class Membership'
method: Repository
speciesForCollect

"Returns a Class, an instance of which should be used as the result of
 collect: or other projections applied to the receiver."

^ Array
%

category: 'Repository Conversion'
method: Repository
_migrateGroups

"Migrates group collections in all the ObjectSecurityPolicies of the receiver."

self do:[ :anObjectSecurityPolicy | anObjectSecurityPolicy ~~ nil 
  ifTrue:[ anObjectSecurityPolicy _migrateGroups] ]
%

category: 'Searching'
method: Repository
indexOf: anObject

"Returns the index of the first element in the receiver that is equivalent to
 the argument. If the receiver does not have any elements equivalent to the
 argument, returns, zero.

 Since ObjectSecurityPolicies are canonicalized by the GsObjectSecurityPolicy creation methods, and
 only referenced from SystemRepository, this is implemented with
 indexOfIdentical:
"

^ self indexOfIdentical: anObject
%

category: 'Searching'
method: Repository
indexOfIdentical: anObject

"Returns the index of the first element in the receiver that is identical to
 the argument. If the receiver does not have any elements identical to the
 argument, returns, zero."

<primitive: 494>
self _primitiveFailed: #indexOfIdentical: args: { anObject } 
%

category: 'Searching'
method: Repository
indexOf: anObject ifAbsent: anExceptionBlock

"Returns the index of the first element in the receiver that is equivalent to
 the argument. If the receiver does not have any elements equivalent to the
 argument, returns the value of evaluating anExceptionBlock."

| ofs |
ofs := self indexOfIdentical: anObject .
ofs > 0 ifTrue:[ ^ ofs ].

^ anExceptionBlock value.
%

category: 'conversion'
method: GsObjectSecurityPolicy
_setObjectSecurityPolicyId: anInt

self _unsafeAt: 7 put: anInt
%

! ------------ conversion of pre-v2.2  SystemRepository
expectvalue %SmallInteger
run
 "Check number of old objectSecurityPolicies needing conversion"
 | oldRep |
 oldRep := Globals at:#SystemRepository .
 oldRep class superClass == Array ifTrue:[ | oldSize maxSegId |
   oldSize := oldRep size .
   maxSegId := 16rFFFE .
   oldSize <= maxSegId ifTrue:[
      ^ oldSize 
   ] ifFalse:[
     ^ 'old SystemRepository too big, only the first ' , maxSegId asString ,
	' out of ' , oldSize asString , ' will be converted.' 
   ]
 ] ifFalse:[
   ^ 0 "no conversion needed"
 ]
%
expectvalue %String
run
  "Create aliases for changed names in Globals"
| report |
report := String new .
#(
  #(#'Segment'				#'GsObjectSecurityPolicy')
  #(#'DataCuratorSegment'		#'DataCuratorObjectSecurityPolicy')
  #(#'SystemSegment'			#'SystemObjectSecurityPolicy')
  #(#'SecurityDataSegment'	#'SecurityDataObjectSecurityPolicy')
  #(#'GsIndexingSegment'		#'GsIndexingObjectSecurityPolicy')
) do: [:pair |
  | oldName newName oldValue newValue |
  oldName := pair at: 1.
  newName := pair at: 2.
  oldValue := Globals at: oldName otherwise: nil.
  newValue := Globals at: newName otherwise: nil.
  (oldValue == nil and:[ newValue ~~ nil]) ifTrue: [
    Globals at: oldName put: newValue.
    report add: 'added '; add: oldName ; lf .
  ].
  (newValue == nil and:[ oldValue ~~ nil]) ifTrue: [
     Globals at: newName put: oldValue.
     report add: 'added '; add: newName ; lf .
  ].
].
^ report
%

! fix 41803 merged from 2.4.5
expectvalue %String 
run
 "Now convert or initialize the objectSecurityPolicies in the repository."
|  oldRep inConversion 
   rep count report seg sysUser installSegBlock segsCreated |
report := String new .
oldRep := Globals at:#SystemRepository .
oldRep class superClass == Array ifTrue:[ | newSize newRep |
   inConversion := true .
   newSize := oldRep size min: 16rFFFE .  "note message above"
   newSize < 2 ifTrue:[ | dcSeg |
     dcSeg := Globals at:#DataCuratorObjectSecurityPolicy otherwise: nil  .
     dcSeg ~~ nil ifTrue:[
       "looks like a 2.0 or 2.1 repository, need to fixup DataCuratorObjectSecurityPolicy"
       newSize := 2.
       oldRep at:2 put: DataCuratorObjectSecurityPolicy .
       { #GsIndexingObjectSecurityPolicy . #PublishedObjectSecurityPolicy . 
	 #GsTimeZoneObjectSecurityPolicy . #SecurityDataObjectSecurityPolicy } 
		do:[ :aKey | | aSeg |
          aSeg := Globals at: aKey otherwise: nil .
          aSeg == (oldRep at: 1) ifTrue:[
            "delete this ref to SystemObjectSecurityPolicy"  
            Globals removeKey: aKey . 
          ]. 
        ]
     ].
   ].
   newRep := Repository new: newSize .   "new: not yet disallowed"
   newRep instVarAt:1 put: (oldRep instVarAt:1) "name"  .
   newRep instVarAt:2 put: (oldRep instVarAt:2) "dataDictionary" .
   newRep _unsafeSetOop: 240641 .

   GsObjectSecurityPolicy _constraints at: 1 put: newRep class  .
   GsObjectSecurityPolicy _constraints at: 7 put: SmallInteger .

   GsObjectSecurityPolicy _constraints _unsafeAt: 1 put: newRep class  .
   GsObjectSecurityPolicy _constraints _unsafeAt: 7 put: SmallInteger .
   1 to: newSize do:[:segId| | anObjectSecurityPolicy |
     anObjectSecurityPolicy := oldRep at: segId .
     anObjectSecurityPolicy _unsafeAt:1 put: newRep .
     anObjectSecurityPolicy _unsafeAt:7 put: segId  "objectSecurityPolicyId" .
   ].
   1 to: newSize do:[:segId| | anObjectSecurityPolicy |
     anObjectSecurityPolicy := oldRep at: segId .
     newRep _basicAt:segId put: anObjectSecurityPolicy .
     anObjectSecurityPolicy _objectSecurityPolicy: DataCuratorObjectSecurityPolicy .
   ].
   newRep _objectSecurityPolicy:  DataCuratorObjectSecurityPolicy .
   Globals at:#SystemRepository put: newRep .
   Globals at:#OldSystemRepository put: oldRep .
   report addAll: 'converted ' ; addAll: newSize asString ;
     addAll: ' objectSecurityPolicies. '
] ifFalse:[
   inConversion := false .
   report addAll: 'conversion of SystemRepository not needed. '
].

"Ensure all ObjectSecurityPolicies from a virgin 2.2 exist . 
 Gs64 v2.2 bom.c only creates the first 2 . "

rep := Globals at:#SystemRepository .
count := 0 .
segsCreated := { false . false . false . false .  false . false . false . false } .
rep size < 8 ifTrue:[
  [ rep size < 8 ] whileTrue:[
    count := count + 1 .
    GsObjectSecurityPolicy newInRepository: rep . 
    segsCreated at: rep size put: true .
  ].
  report addAll:'created '; addAll: count asString ; addAll: ' objectSecurityPolicies. '
].
sysUser := AllUsers userWithId:'SystemUser' .
rep := Globals at:#SystemRepository .

installSegBlock := [ :aKey :aSeg | | oldSeg |
  oldSeg := Globals at: aKey otherwise: nil .
  oldSeg == nil ifTrue:[
    Globals at: aKey put: aSeg
  ] ifFalse:[
    oldSeg == aSeg ifFalse:[
      nil error:'invalid GsObjectSecurityPolicy reference in Globals'
    ]
  ]
].

seg := rep at: 1 .		"repeats bom work, needed for conversion"
installSegBlock value:#SystemObjectSecurityPolicy value: seg .
((segsCreated at: 1) or:[ inConversion]) ifTrue:[
  seg  owner: sysUser ; ownerAuthorization: #write  ;
    worldAuthorization: #read  .
].

seg := rep at: 2 .		"repeats bom work, needed for conversion"
installSegBlock value:#DataCuratorObjectSecurityPolicy value: seg .
((segsCreated at: 2) or:[ inConversion]) ifTrue:[
  seg owner: (AllUsers userWithId:'DataCurator') ;
    ownerAuthorization: #write  ;
    worldAuthorization: #read  .
].

				"remaining not done in bom"
seg := rep at: 3 .	"GsTimeZoneObjectSecurityPolicy no longer used"	

seg := rep at: 4 .
installSegBlock value:#GsIndexingObjectSecurityPolicy value: seg .
((segsCreated at: 4) or:[ inConversion]) ifTrue:[
  seg isInvariant ifFalse:[
    seg owner: sysUser ; ownerAuthorization: #write  ;
      worldAuthorization: #write  ; immediateInvariant .
].
].

seg := rep at: 5 .
installSegBlock value:#SecurityDataObjectSecurityPolicy value: seg .
seg isInvariant ifFalse:[
  seg owner: sysUser ; ownerAuthorization: #write  ;
    worldAuthorization: #none  ;  immediateInvariant; _unsafeSetOop: 235777 .
].

seg := rep at: 6 .
installSegBlock value:#PublishedObjectSecurityPolicy value: seg .
((segsCreated at: 6) or:[ inConversion]) ifTrue:[
  seg isInvariant ifFalse:[
    seg owner: sysUser ;   "group authorizations done later in bomlast"
      ownerAuthorization: #write ;
    worldAuthorization: #none .
  ].
].

"GsObjectSecurityPolicy 7 , owned by GcUser , fixed up later"
"GsObjectSecurityPolicy 8, owned by Nameless, fixed up later"

inConversion ifTrue:[
  "converting from a pre-Gs64 v2.2 repository.
   Fix 35966 , repositories from Gs64 v2.0.x and v2.1.x  may 
   have objectSecurityPolicyId == OOP_NIL (10r20) in disk object headers ,
   so preallocate a GsObjectSecurityPolicy for objectSecurityPolicyId 20 that is world read-write "
   rep size < 20 ifTrue:[ | newSeg |
     "Possibly a Gs64 v2.0.x or v2.1.x  repository "
     newSeg := GsObjectSecurityPolicy newInRepository: rep .  
     newSeg owner: sysUser; ownerAuthorization: #write;
            worldAuthorization:#write ; immediateInvariant  .
     newSeg objectSecurityPolicyId >= 20 ifTrue:[
        nil error:'invalid objectSecurityPolicyId' 
     ].
     newSeg _setObjectSecurityPolicyId: 20. 		"fix 38084"
     rep size: 20 .
     rep at: 20 put: newSeg .
     report addAll: ' created GsObjectSecurityPolicy 20 .'.
   ]
].
^ report 
%

expectvalue %String
run
  "Create aliases for changed names in Globals"
| report list |
report := String new .
list := #(
  #(#'Segment'				#'GsObjectSecurityPolicy')
  #(#'DataCuratorSegment'		#'DataCuratorObjectSecurityPolicy')
  #(#'SystemSegment'			#'SystemObjectSecurityPolicy')
  #(#'SecurityDataSegment'	#'SecurityDataObjectSecurityPolicy')
  #(#'GsIndexingSegment'		#'GsIndexingObjectSecurityPolicy')).
list do: [:pair |
  | oldName newName oldValue newValue |
  oldName := pair at: 1.
  newName := pair at: 2.
  oldValue := Globals at: oldName otherwise: nil.
  newValue := Globals at: newName otherwise: nil.
  (oldValue == nil and:[ newValue ~~ nil]) ifTrue: [
    Globals at: oldName put: newValue.
    report add: 'added '; add: oldName ; lf .
  ].
  (newValue == nil and:[ oldValue ~~ nil]) ifTrue: [
     Globals at: newName put: oldValue.
     report add: 'added '; add: newName ; lf .
  ].
].
list do:[:pair | pair do:[:name | (Globals associationAt: name) immediateInvariant ]].
^ report
%


expectvalue true
run
GsObjectSecurityPolicy removeSelector: #_setObjectSecurityPolicyId: .
^ true
%

! ------- instance creation methods disallowed
category: 'Instance Creation'
classmethod: Repository
new

"Disallowed."
self shouldNotImplement: #new
%

category: 'Instance Creation'
classmethod: Repository
new: anInteger

"Disallowed."

self shouldNotImplement: #new:
%


! with:...  disallowed because new: disallowed

!--------------- methods inherited from Object that are disallowed
set class Repository

category: 'Updating'
method
at: offset put: aValue
"Disallowed"
self shouldNotImplement: #at:put:
%

method
basicAt: offset put: aValue
"Disallowed"
self shouldNotImplement: #basicAt:put:
%

method
instVarAt: offset put: aValue 
"Disallowed"
self shouldNotImplement: #instVarAt:put:
%

method
nilFields
"Disallowed"
self shouldNotImplement: #nilFields
%

method
size: aSize
"Disallowed"
self shouldNotImplement: #size:
%

method
_at: offset put: aValue
"Disallowed"
self shouldNotImplement: #_at:put:
%

method
_basicAt: offset put: aValue
"Disallowed"
self shouldNotImplement: #_basicAt:put:
%

method
_basicSize: aSize
"Disallowed"
self shouldNotImplement: #_basicSize:
%

method
_primitiveAt: offset put: aValue
"Disallowed"
self shouldNotImplement: #_primitiveAt:put:
%

method
_unsafeAt: offset put: aValue
"Disallowed"
self shouldNotImplement: #_unsafeAt:put:
%

method
objectSecurityPolicy: aSeg
"Disallowed"
self shouldNotImplement: #objectSecurityPolicy:
%

method
_objectSecurityPolicy: aSeg
"Disallowed"
self shouldNotImplement: #_objectSecurityPolicy:
%

category: 'Storing and Loading'

method
basicLoadFrom: passiveObj
"Disallowed."
self shouldNotImplement: #basicLoadFrom:
%

method
basicLoadFrom: passiveObj size: aSize
"Disallowed"
self shouldNotImplement: #basicLoadFrom:size:
%

method
basicLoadFromNoRead: passiveObj
"Disallowed"
self shouldNotImplement: #basicLoadFromNoRead:
%

method
basicLoadFromNoRead: passiveObj size: aSize
"Disallowed"
self shouldNotImplement: #basicLoadFromNoRead:size:
%

method
basicLoadFromOld: passiveObj
"Disallowed"
self shouldNotImplement: #basicLoadFromOld:
%

method
loadFrom: passiveObj
"Disallowed"
self shouldNotImplement: #loadFrom:
%

method
loadNamedIVsFrom: passiveObj
"Disallowed"
self shouldNotImplement: #loadNamedIVsFrom:
%

method
loadVaryingFrom: passiveObj
"Disallowed"
self shouldNotImplement: #loadVaryingFrom:
%

method
loadVaryingFrom: passiveObj size: aSize
"Disallowed"
self shouldNotImplement: #loadVaryingFrom:size:
%

category: 'Instance migration'

method
become: anObj
"Disallowed"
self shouldNotImplement: #become:
%

method
_primitiveBecome: anObject forDict: aBoolean
"Disallowed"
self shouldNotImplement: #_primitiveBecome:forDict:
%

method
migrate
"Disallowed"
self shouldNotImplement: #migrate
%

method
migrateFrom: anObj
"Disallowed"
self shouldNotImplement: #migrateFrom:
%

method
migrateFrom: anObj instVarMap: aDict
"Disallowed"
self shouldNotImplement: #migrateFrom:instVarMap:
%

method
migrateIndexable: anObj myClass: aCls otherClass: secondCls
"Disallowed"
self shouldNotImplement: #migrateIndexable:myClass:otherClass:
%

method
changeClassTo: aCls
"Disallowed"
self shouldNotImplement: #changeClassTo:
%

category: 'Copying
method
copy
"Disallowed"
self shouldNotImplement: #copy
%

method
deepCopy
"Disallowed"
self shouldNotImplement: #deepCopy
%

method
shallowCopy
"Disallowed"
self shouldNotImplement: #shallowCopy
%


! --------------  methods inherited from Collection  that are disallowed 
method
add: anObj
"Disallowed"
self shouldNotImplement: #add:
%

method
addAll: aCollection
"Disallowed"
self shouldNotImplement: #addAll:
%

method
remove: anObj
"Disallowed"
self shouldNotImplement: #remove:
%

method
remove: anObj ifAbsent: aBlock
"Disallowed"
self shouldNotImplement: #remove:ifAbsent:
%

method
removeAll: aCollection
"Disallowed"
self shouldNotImplement: #removeAll:
%

method
removeAllIdentical: aCollection
"Disallowed"
self shouldNotImplement: #removeAllIdentical:
%

method
removeIdentical: anObj
"Disallowed"
self shouldNotImplement: #removeIdentical:
%

method
removeIdentical: anObj ifAbsent: aBlock
"Disallowed"
self shouldNotImplement: #removeIdentical:ifAbsent:
%

category: 'Garbage Collection'
method: Repository
loadGcCandidatesFromFile: aString intoHiddenSet: hiddenSetId

"Given a file which is binary list of dead objects (produced
 by one of the #findDisconnectedObjects: methods),
 load the objects into the hidden set with the given ID.

 Only objects which exist and are not in the stone's dead object
 set are added to the hidden set.

 Returns an Array containing 3 SmallIntegers:
  1 - Total number of objects in the file.
  2 - Number of valid objects loaded into the hidden set.
  3 - Number of invalid objects not loaded.

 Requires the GarbageCollection privilege.

 If this session contains uncommitted changes to the repository, the method 
 signals a error: #rtErrAbortWouldLoseData, to indicate that data could be lost.
 Otherwise it puts the session into auto-begin transaction mode and aborts."

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

^ self primLoadGcCandidatesFromFile: aString intoHiddenSet: hiddenSetId
%

category: 'Private'
method: Repository
primLoadGcCandidatesFromFile: aString intoHiddenSet: hiddenSetId

<primitive: 597>
aString _validateClass: String .
hiddenSetId _validateClass: SmallInteger .
self _primitiveFailed: #loadGcCandidatesFromFile:intoHiddenSet: 
     args: { aString . hiddenSetId }
%

category: 'Instance Counting'
method: Repository
countInstances: anArray withMaxThreads: maxThreads maxCpuUsage: aPercentage

"Scans the entire Repository and generates an array of instance counts of the 
 the classes contained in anArray. Does not include in-memory objects.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData. The entire
 Repository is scanned once."

| inputSet resultInSetOrder result inputArraySize anObj scanBlk |

System needsCommit ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

inputSet := self _arrayOfClassesAsSet: anArray .
inputSet size < 1 ifTrue:[ ^ { } ] .

scanBlk := [ :scanSetThisTime |
      self _scanPomWithMaxThreads: maxThreads waitForLock: 90 pageBufSize: 8
       percentCpuActiveLimit: aPercentage  identSet: scanSetThisTime 
       limit: 0 scanKind: 2 toDirectory: nil ].

resultInSetOrder:= self _doListInstancesFrom: inputSet with: scanBlk 
			includeMemory: false . 

inputArraySize := anArray size .
result := Array new: inputArraySize .
1 to: inputArraySize do:[:j|
  anObj := anArray at: j .
  result at: j put: ( resultInSetOrder at: (inputSet _offsetOf: anObj) * 2)
  ].
^ result
%

category: 'Instance Counting'
method: Repository
countInstances: anArray

"Scans the entire Repository and generates an array of instance counts of the 
 the classes contained in anArray.

 The scan is done in an non-aggressive manner and uses only 2 threads.  To 
 complete the operation in less time, see the #fastCountInstances:
 method."

^ self countInstances: anArray withMaxThreads: 2 maxCpuUsage: 90
%

category: 'Instance Counting'
method: Repository
fastCountInstances: anArray

"Same as the #countInstances: method except the scan is performed 
 aggressively in order to complete in as little time as possible.

 This method may consume most or all host system resources while it is in progress."

^ self countInstances: anArray withMaxThreads: self _aggressiveMaxThreadCount
       maxCpuUsage: 95
%

category: 'Page-Order Operations'
method: Repository
listInstancesInPageOrder: anArrayOfClasses toFile: aString 
 withMaxThreads: maxThreads maxCpuUsage: aPercentage

"Scans the repository for all committed instances of the classes
 contained in anArray and stores the results into a new file
 specified in aString.
 Does not include objects in VM's temporary object memory.

 anArray must be an instance of Array which contains at least 1 and
 not more than 2034 classes.

 aString must be the full path to a new file on the gem's host system 
 which will be created by this method.  It is an error if the file
 already exists.

 All results of the scan (all instances of the classes in anArray) are
 stored in the file in page order (i.e., the same order as the  objects
 are stored on disk) as of the time of the scan.   Note that unlike other
 #listInstances methods,  the file contains the complete list of all
 instances of all classes contained in anArray.

 The ultimate file size will be approximately 5 bytes for each object
 found by the scan, plus 24 bytes.  For example, a scan which locates 10 
 million objects will generate a file approximately 50 megabytes in size.

 Objects created and committed after starting this method are not included
 in the result file.

 This method aborts the current transaction; if an abort would cause unsaved
 changes to be lost, it signals an error, #rtErrAbortWouldLoseData.

 This method runs in a transaction and is therefore not recommended for 
 production systems.

 Operations which process a large volume of objects (e.g.: instance migration,
 etc) often run significantly faster if the objects are processed in page order
 rather than in object ID order.  Other #listInstances: methods always return
 results in object ID order.

 Objects are routinely moved from old pages to new pages whenever modified.
 The Reclaim GC gems also move objects to new pages.  Therefore the actual
 order the objects appear on disk will diverge over time from the order of
 the objects in the result file.

 Returns a SmallInteger representing the total number of objects written to the
 file."
 
System needsCommit
  ifTrue: [ ^ self _error: #rtErrAbortWouldLoseData ] .

^ self _scanPomWithMaxThreads: maxThreads waitForLock: 90 pageBufSize: 8 
                  percentCpuActiveLimit: aPercentage
                  identSet: (IdentitySet withAll: anArrayOfClasses)
                  limit: 0 scanKind: 3 toDirectory: aString
%


category: 'Page-Order Operations'
method: Repository
listInstancesInPageOrder: anArrayOfClasses toFile: aString
"See the method #listInstancesInPageOrder:toFile:withMaxThreads:maxCpuUsage: 
 for more information on this method.

 The scan is done in an non-aggressive manner and uses only 2 threads.  To 
 complete the operation in less time, see the #fastListInstancesToHiddenSet:
 method."

^ self listInstancesInPageOrder: anArrayOfClasses toFile: aString 
       withMaxThreads: 2 maxCpuUsage: 90
%

category: 'Page-Order Operations'
method: Repository
fastListInstancesInPageOrder: anArrayOfClasses toFile: aString
"Same as the #listInstancesInPageOrder:toFile: method except the scan is performed 
 aggressively in order to complete in as little time as possible.

 This method may consume most or all host system resources while it is in progress."

^ self listInstancesInPageOrder: anArrayOfClasses toFile: aString 
       withMaxThreads: self _aggressiveMaxThreadCount maxCpuUsage: 95

%

category: 'Page-Order Operations'
method: Repository
openPageOrderOopFile: aString

"Opens a binary page-ordered file of object ID's created by the
 Repository>>listInstancesInPageOrder:toFile: method.

 Raises an error if the file could not be opened for reading, the file
 is corrupt, or aString is not an instance of String.

 Returns a SmallInteger which is the ID of the opened file.  This ID must
 be used to read the contents of the file (see the method
 readObjectsFromFileWithId:startingAt:upTo:into: in this class).

 This method acquires and retains the garbage collection lock.   No garbage
 collection operations can be started until the
 Repository>>closePageOrderOopFileWithId: method is executed or the session
 logs out."

<primitive: 830>
aString _validateClass: String .
aString size == 0 ifTrue:[ aString _error: #errArgTooSmall args:{ 1 } ].
^ self _primitiveFailed: #openPageOrderOopFile: args: { aString }
%

category: 'Page-Order Operations'
method: Repository
readObjectsFromFileWithId: aSmallInt startingAt: startIndex upTo: endIndex into: anArray

"Reads and validates objects from a page-ordered file which was previously
 opened with the Repository>>openPageOrderOopFile: method.

 aSmallInt must be the result of the Repository>>openPageOrderOopFile: method.

 startIndex is the index of the first object in the file to read.  The first 
 object in the file has an index of 1.  It is an error if startIndex is less than 1
 or greater than endIndex.

 endIndex is the index of the last object to read from the file.  endIndex 
 must be greater than or equal to startIndex.  endIndex may exceed the index of
 the last object in the file.  In that case, all objects from startIndex to the
 end of the file will be returned.

 anArray must be an instance of Array when the resulting objects will be 
 stored.  It is an error if anArray is not empty.

 Returns a SmallInteger which is the size of anArray.  This will be the lesser
 of the number (endingIndex - startIndex + 1) or, all remaining objects in the
 file.

 It is possible that one or more object identifiers contained in the file
 are no longer valid due to garbage collection.  Objects which are no longer
 valid have nil stored in their place in anArray.  Objects that have been 
 garbage collected and are in the free oop list or the dead not reclaimed set
 are considered to be invalid.

 The session must be in a transaction when this method is invoked.  If the
 session is not in a transaction, an #rtErrPrimOutsideTrans error is raised.

 For best performance, it is recommended that no more than 2034 objects
 be returned by any single invocation of this method.  Returning a high
 number of objects will cause anArray to grow very large and may result in
 a fatal out-of-memory error."

^ self _primReadObjectsFromFileWithId: aSmallInt startingAt: startIndex upTo: endIndex into: anArray 
%


category: 'Page-Order Operations'
method: Repository
numberOfObjectsInPageOrderOopFileWithId: aSmallInt

"Answer the number of objects in a page order oop file previously opened 
 with the Repository>>openPageOrderOopFile: method.
 aSmallInt must be the result of the Repository>>openPageOrderOopFile: method.
 Does not close or otherwise change the state of the file."

^ self _primReadObjectsFromFileWithId: aSmallInt startingAt: nil upTo: nil into: nil
%

category: 'Private'
method: Repository
_primReadObjectsFromFileWithId: aSmallInt startingAt: startIndex upTo: endIndex into: anArray

<primitive: 831>
aSmallInt _validateClass: SmallInteger .
startIndex ifNil:[ "all args nil means return the number of oops in the file"
 endIndex _validateClass: UndefinedObject .
 anArray _validateClass: UndefinedObject .
] ifNotNil:[ "normal case: request to read the file failed."
  startIndex _validateClass: SmallInteger .
  endIndex _validateClass: SmallInteger .
  anArray _validateClass: Array .
  anArray size == 0 ifFalse:[ anArray _error: #errArgTooSmall args:{ 1 } ] .
  (startIndex < 1 or:[ startIndex > endIndex]) ifTrue:[ 
     startIndex _error: #rtErrArgOutOfRange args:{ 1 . endIndex } ] .
].
^ self _primitiveFailed: 
       #_primReadObjectsFromFileWithId:startingAt:upTo:into: 
       args: { aSmallInt . startIndex . endIndex . anArray }
%
category: 'Page-Order Operations'
method: Repository
closePageOrderOopFileWithId: aSmallInt

"Closes a file previously opened with the Repository>>openPageOrderOopFile: 
 method and releases the garbage collection lock.

 Returns true if successful; false if the file was not open or was already 
 closed."

<primitive: 832>
aSmallInt _validateClass: SmallInteger .
^ self _primitiveFailed: #closePageOrderOopFileWithId: args: { aSmallInt }
%

category: 'Page-Order Operations'
method: Repository
auditPageOrderOopFileWithId: aSmallInt

"Reads the page order oop file with the ID returned by the 
 Repository>>openPageOrderOopFile: method and counts the number of
 objects in the file which are no longer valid.  Objects are 
 considered invalid if they no longer exist (i.e., are not 
 present in the shared object table) or present in the dead objects
 set maintained by stone.

 The session must be in a transaction when this method is invoked.  If the
 session is not in a transaction, an #rtErrPrimOutsideTrans error is raised.

 Returns an array containing 2 SmallIntegers:
    array[1] - total number of oops in the file.
    array[2] - number of invalid oops in the file.

 Returns an array containing 2 nils if aSmallInt is not a valid file
 ID returned by the Repository>>openPageOrderOopFile: method."

<primitive: 833>
aSmallInt _validateClass: SmallInteger .
^ self _primitiveFailed: #auditPageOrderOopFileWithId: args: { aSmallInt }
%

category: 'Private'
method: Repository
_pageForOopIsTagged: anInt
"Lookup anInt in the shared object table and return a Boolean indicating
 if the page ID for the object is tagged, indicating it is in the shared
 dependency map.  Returns nil if anInt is not in the object table or it is
 a free oop."

<primitive: 570>
self _primitiveFailed: #_pageForOopIsTagged: args: { anInt }
%

category: 'Private'
method: Repository
_sharedDepMapLookupForOop: anInt
"Lookup anInt in the shared dependency map and return an integer which indicates
 the value in the map for the given OOP (which is the DependencyList for the
 given OOP).

 Returns nil if anInt is not present in the shared dependency map."

<primitive: 571>
self _primitiveFailed: #_sharedDepMapLookupForOop: args: { anInt }
%

category: 'Garbage Collection'
method: Repository
basicFindDisconnectedObjectsAndWriteToFile: aFileNameString
 pageBufferSize: pageBufSize saveToRepository: saveToRep
 withMaxThreads: maxThreads maxCpuUsage: aPercentage

"Perform a markForCollection on the system and write the list of
 dead objects to given file.  If saveToRep is true, also store the results
 in the array under the key #FdcResults in Globals.

 To save the dead objects to an array but not write them to a file, set
 aFileNameString to nil and saveToRep to true.  It is illegal for 
 aFileNameString to be nil and for saveToRep to be false.

 This method aborts the current transaction.

 pageBufSize must be a power of two and specifies the number of pages to buffer per thread.

 This method will return an error if the repository is in restore mode.

 See also the comments in the method
  Repository>> _mfcWithMaxThreads: maxThreads waitForLock: lockWaitTime 

 for a desciption of the result array and a more detailed description of
 this method."

| fdcResults arrayOfDead |

System needsCommit
  ifTrue: [ self _error: #rtErrAbortWouldLoseData ] .

(System transactionMode == #autoBegin)
  ifFalse:[System transactionMode: #autoBegin].

(SystemRepository restoreActive)
  ifTrue:[^ self _error: #rtErrCommitDbInRestore].

System abortTransaction.
self cleanupMgcResultsForMgc: false.
System commitTransaction
  ifFalse:[^ self _error: #rtErrGcCommitFailure].
self cleanupFdcResultsForMgc: false.
fdcResults := Globals at: #FdcResults.

arrayOfDead := fdcResults at: 3.
saveToRep
  ifTrue:[
    (arrayOfDead == nil)
        ifTrue:[fdcResults at: 3 put: Array new].]
  ifFalse:[fdcResults at: 3 put: nil].

System commitTransaction
  ifFalse:[^ self _error: #rtErrGcCommitFailure].

^ self _fdcWithMaxThreads: maxThreads
      waitForLock: 90 pageBufSize: pageBufSize percentCpuActiveLimit: aPercentage
      toFile: aFileNameString resultArray: fdcResults 
%

category 'Garbage Collection'
method: Repository
markForCollectionWithMaxThreads: maxThreads waitForLock: waitTimeSeconds
"See the markForCollection method for comments."

^ self markForCollectionWithMaxThreads: maxThreads 
                           waitForLock: waitTimeSeconds 
                           pageBufSize: 128
                 percentCpuActiveLimit: 90
%

! fixed 41773
category 'Garbage Collection'
method: Repository
markForCollectionWithMaxThreads: maxThreads waitForLock: waitTimeSeconds 
 percentCpuActiveLimit: aPercent

"See the markForCollection method for comments."

| arr ex msg |
arr := self _mfcWithMaxThreads: maxThreads waitForLock: waitTimeSeconds 
    pageBufSize: 128 percentCpuActiveLimit: aPercent .
(ex := Warning new) _number: 2515 ; args: arr .
msg := 'markForCollection found ' ,
        (arr atOrNil: 1) asString , ' live objects, ',
        (arr atOrNil: 2) asString , ' dead objects(occupying approx ' ,
        (arr atOrNil: 3) asString , ' bytes)' .
ex signal: msg .
^ ex
%

category 'Garbage Collection'
method: Repository
markForCollectionWithMaxThreads: maxThreads waitForLock: waitTimeSeconds 
 pageBufSize: pageBufSize percentCpuActiveLimit: aPercent

"See the markForCollection method for comments."

| arr ex msg |
arr := self _mfcWithMaxThreads: maxThreads waitForLock: waitTimeSeconds 
    pageBufSize: pageBufSize percentCpuActiveLimit: aPercent .
(ex := Warning new) _number: 2515 ; args: arr .
msg := 'markForCollection found ' , 
        (arr atOrNil: 1) asString , ' live objects, ',
        (arr atOrNil: 2) asString , ' dead objects(occupying approx ' ,
        (arr atOrNil: 3) asString , ' bytes)' .
ex signal: msg .
^ ex
%

category 'Garbage Collection'
method: Repository
fastMarkForCollection

"Same as the #markForCollection method except the scan is performed 
 aggressively in order to complete in as little time as possible.

 This method may consume most or all host system resources while it
 is in progress."

^ self markForCollectionWithMaxThreads: self _aggressiveMaxThreadCount 
  waitForLock: 90 pageBufSize: 128 percentCpuActiveLimit: 95
%

category 'Garbage Collection'
method: Repository
fastMarkForCollectionWithPageBufSize: pageBufSize

"Same as the #fastMarkForCollection method except you can specify 
 a larger pageBufSize than the default 128.

 This method may consume most or all host system resources while it
 is in progress."

^ self markForCollectionWithMaxThreads: self _aggressiveMaxThreadCount 
  waitForLock: 90 pageBufSize: pageBufSize percentCpuActiveLimit: 95
%

! fixed 41773
category: 'Private'
method: Repository
_mfcWithMaxThreads: maxThreads waitForLock: lockWaitTime 
      pageBufSize: aBufSize percentCpuActiveLimit: aPercent
 "Perform a multithreaded markForCollection operation using at most maxThreads.

  Returns an Array { numLiveObjects . numDeadObjects . approxSizeOfDeadObjects } .

  The maxThreads argument specifies the maximum number of slave sessions (threads) 
  that can be used during the operation.  The actual number of active threads
  can be adjusted to a lower value at runtime automatically by the main 
  thread based on the percentCpuActiveLimit and dynamically by the setting 
  the mtThreadLimit (see details below).

  The lockWaitTime argument is used to specify how many seconds method should
  wait while attempting to acquire the gcLock.  No other garbage collection 
  operations may be started or in progress while this method is running.  
  There also must be no outstanding possible dead objects in the system for 
  the GC lock to be granted.
 
  The pageBufSize, which must be a power of two, specifies the number 
  of pages to buffer per thread. This parameter in conjunction with the 
  maxThreads largely determines the memory footprint needed to perform 
  this operation.

  The cpuActiveLimit specifies a level of total cpu activity at which the algorithm
  automatically inactivates threads to prevent overload of system resources.

  This algorithm makes use of additional sessions (threads) to achieve 
  significant performance improvements.  It also makes space/time trade offs
  so that heap memory in addition to the TemporaryObjectCache (TOC)
  resources are used.  In fact, this algorithm doesn't require any TOC, so
  configuring this process for the smallest TOC space is advantageous. 
  The memory space that is needed is variable and depends upon the current 
  oopHighWater value, the number of sessions and the pageBufSize specified.

  The overhead associated with the oopHighWater value can be computed:
    oopHighWater in bytes = (stnOopHighWater + 10M) / 2

  The memory cost per thread is 50K + (180K * pageBufSize)

  To give some idea of how this scales, a system that had an oopHighWaterMark
  of 500M running 8 sessions with a pageBufSize of 512 would require about 1GB 
  of free memory to start up.

  Other tuning considerations:

  During the operation the repository is scanned repeatedly from low pageId 
  to high pageId within each extent.  Performance can be optimized by arranging 
  the extents so that the scan causes the fewest head movements on each drive 
  and the threads can simultaneously initiate I/O operations on multiple extents.

  For example:   With a single disk controller and one extent per drive the 
    following would provide an optimal configuration: 
       Disk1: extent0
      Disk2: extent1
       ... 
  With multiple controllers:
     Controller1:  Disk 1 extent0
                   Disk 2 extent2
                   Disk 3 extent4
     Controller2:  Disk 4 extent1
                   Disk 5 extent3
                   Disk 6 extent5
  
  For use with a RAID or SAN system, the device should be configured for 16K read
  operations from sequentially increasing blocks within an extent.

  This operation requires almost continuous access to the entire object table.
  Running on a system that can hold all of the object table in the cache is will
  allow it to perform optimally.  The object table size is approximately 
  stnOopHighWater * 12 bytes.

  If the machine can be dedicated to performing this operation, a good rule of thumb
  is to set the maxThreads to numCpus + numControllers and the percentCpuActiveLimit
  should be increased to 100.

  If the operation is being performed on a system that is running a live application
  the maxThreads can still be set to a high value and the cpuActiveLimit can be
  used to control the cpu resources used by this method.  If this operation pushes the
  cpu utilization above the specified limit the main thread automatically deactivates 
  slave threads until the load is again below the specified limit.  While threads are 
  not actively working on the operation, they still respond to the commit record 
  backlog and abort as needed to prevent the backlog from exceeding its specified limit.

  Finally, the pageBufSize specified must be a power of 2. Larger values can improve 
  performance.  For example, in a test case where the machine was dedicated to 
  performing this operation, increasing the pageBufSize from 64 to 512 reduced 
  the elapsed time by about 20 percent.  On the other hand, if the operation is being 
  run on a live system then smaller values may be better depending on how often the algorithm 
  must abort (and clear the buffers) to prevent the commit record backlog from growing.

  Caution should be taken when setting the numSessions and pageBufSize to prevent 
  the memory footprint for this process from being so large that the system starts
  swapping.

  The values of the arguments percentCpuActiveLimit and maxThreads are used as initial
  values for two statistics associated with this process, the mtPercentCpuActiveLimit 
  and mtThreadLimit.  The main thread monitors these statistics and uses them to control
  the number of threads that are actively working, which is displayed in the mtActiveThreads
  statistic.  Thus the operation can be tuned dynamically by setting either the 
  mtPercentCpuActiveLimit or mtThreadLimit for the process.  When the system is not performing
  adequately because the operation is consuming too much of the cpu resources, the 
  mtPercentCpuActiveLimit: method can be used to decrease the value of the corresponding
  statistic and thus remove some of the load on the system caused by this operation.
  In situations where I/O is the limiting factor on the system and the operation is 
  using too much I/O bandwidth, the number of active threads can be changed by using 
  the mtThreadsLimit: method.  The mtPercentCpuActiveLimit can be set to any value 
  between 0 and 100, mtThreadLimit can be set to values between 0 and maxThreads."
  
<primitive: 876>

maxThreads _validateClass: SmallInteger.
lockWaitTime _validateClass: SmallInteger.
aBufSize _validateClass: SmallInteger ; _validateIsPowerOf2 .
self _validatePercentage: aPercent .
^ self _primitiveFailed: #_mfcWithMaxThreads:
       args: { maxThreads . lockWaitTime . aBufSize . aPercent }
%

category: 'Private'
method: Repository
_fdcWithMaxThreads: maxThreads waitForLock: lockWaitTime 
    pageBufSize: aBufSize percentCpuActiveLimit: aPercent
    toFile: aFileNameString resultArray: anArray 

 "Run a multithreaded algorithm to determine the disconnected (dead) objects
  in the repository.  The dead objects identifiers (oops) are written
  to a file, stored in anArray, or both.  

  Returns anArray .

  This is method never results in the garbage collection of any objects.

  See _mfcWithMaxThreads:waitForLock:pageBufSize:percentCpuActiveLimit: 
  for documentation of the common arguments.

  If the toFile argument is not nil, a file is created and a list of the
  dead object identifiers found is written to the file in a binary format.
  The file must not exist when the method is called and the argument
  must be an instance or subclass of String.  The resulting file may 
  be used by the _markGcCandidatesFromFile method in this class.  
  The size of the file in bytes is approximately 8 times the number 
  of dead objects found. 

  Be careful to ensure there is sufficient disk space to write the 
  entire list of dead objects when using this method with large repositories.

  The resultArray is expected to be a committed array object of size 4.
  If successful, this method stores the following values in the resultArray:

    1 - anInteger representing the number of live objects found.
    2 - anInteger representing the number of dead objects found.
    3 - anArray containing the dead objects, false if a commit failed,
        or nil if the list of dead objects is not stored in the repository.
    4 - nil if aFileNameString was nil, otherwise a boolean indicating 
        whether the dead object ids were successfully written to the file.

  This method requires the GarbageCollection privilege and aborts the
  current transaction.

  Pre-existing dead objects, i.e., objects which reside in the stone's
  deadNotReclaimed set, are ignored and are not returned by this method.

  The output file is deleted if this method encounters an error.

  See the description of 
    #_mfcWithMaxThreads:waitForLock:pageBufSize:percentCpuActiveLimit:
  for a description of the memory costs and performance optimizations.

  See the following methods in class System for details of accessing 
  and updating the mt statistics:
     mtThreadsLimit: sessionId
     mtThreadsLimit: sessionId setValue: newVal
     mtPercentCpuActiveLimit: sessionId
     mtPercentCpuActiveLimit: sessionId setValue: newVal
     mtMaxThreads: sessionId"

<primitive: 877>

maxThreads _validateClass: SmallInteger.
lockWaitTime _validateClass: SmallInteger.
aBufSize _validateClass: SmallInteger ; _validateIsPowerOf2 .
self _validatePercentage: aPercent .
(aFileNameString == nil) 
  ifTrue:[(anArray at: 3) _validateClass: Array].
anArray _validateClass: Array.
(anArray size == 4)
  ifFalse:[anArray _error: #rtErrArgOutOfRange ].
((anArray at: 3) == nil)
  ifTrue:[aFileNameString _validateClass: String].
self _primitiveFailed: 
     #_fdcWithMaxThreads:waitForLock:pageBufSize:percentCpuActiveLimit:toFile:resultArray:
     args: {  maxThreads . lockWaitTime . aBufSize . 
              aPercent . aFileNameString . anArray } .
^ self _uncontinuableError
%

category: 'Private'
method: Repository
_scanPomWithMaxThreads: maxThreads waitForLock: lockWaitTime 
      pageBufSize: aBufSize percentCpuActiveLimit: aPercent
      identSet: anIdentitySet limit: aLimit
      scanKind: kind toDirectory: aStringOrNil

" This primitive method performs various scans of the Repository.
  The scans only take into account committed objects; i.e., this method 
  does not operate on objects in the Temporary Object Cache (TOC).

  This primitive uses a multithreaded algorithm to sweep the active
  data pages in the repository to gather the information requested.

  The maxThreads argument specifies the maximum number of slave sessions
  (threads) that will be used during the operation.  The actual number of 
  active threads can be adjusted to a lower value automaticly at runtime 
  by the main thread based on the percentCpuActiveLimit and by the user 
  by interactively setting the mtThreadLimit (see details below).

  The lockWaitTime argument is used to specify how many seconds method should
  wait while attempting to acquire the garbage collection (GC) lock.
  No other garbage collection operations may be started or in progress while
  this method is running.  Objects in the possibleDead or deadNotReclaimed 
  sets at the start of the scan are ignored by the scan.
 
  The pageBufSize, which must be a power of two, specifies the number 
  of pages to buffer per thread. This parameter in conjunction with the 
  maxThreads largely determines the memory footprint needed to perform 
  this operation.  The pageBufSize doesn't have much impact on the 
  performance of the scan, so a default size of 8 is probably sufficient.

  The cpuActiveLimit specifies a level of total cpu activity at which the 
  algorithm automatically inactivates threads to prevent overload 
  of system resources.

  This algorithm makes use of additional sessions (threads) to achieve 
  significant performance improvements.  It also makes space/time trade offs
  so that both heap memory and TemporaryObjectCache (TOC)
  resources are used.  In fact, this algorithm doesn't require much TOC, 
  except for the operations that return array results, so configuring this 
  process for a smaller TOC space is advantageous. 

  The memory space that is needed is variable and depends upon number of
  instances being searched for, the number found, the number of sessions 
  requested and the pageBufSize specified for each.

  The anIdentitySet arguement should be initialized to contain the 
  classes whose instances are to be located.

  The aLimit argument specifies the maximum number of instances to report 
  for any class or for the objsLargerThan scan.  A value of 0 specifies
  that all instances are returned (warning - unlimited searches on a large
  database can run out of TOC space).

  The behavior of this method is dependent on the scanKind as follows:

   scanKind  Function       Result
   ===========================================================================
     0     listInstances  Returns an Array.  For each element of 
                          anIdentitySet, the result array contains 2 
                          elements: the total number of instances and 
                          an array of the instances (possibly limited in size).
   
     1     listInstances  The input set must contain a single class.
                          The instances found are stored into the hidden
                          set: ListInstancesResult, and the number of 
                          elements found is the primitive return value.

     2     listInstances  Returns an Array of pairs where the odd numbered 
                          elements are the classes specified in the IdentitySet
                          (in oop order) and the even numbered elements are 
                          the number of instances of the preceding class 
                          that were found.

                          If the toDirectory argument is nil, then only 
                          array described above is returned.  If not nil, 
                          then  Bitmaps containing the object ID's of 
                          the instances are written to binary bitmap 
                          files in the specified directory.  The resulting
                          bitmap files are named:
                             <ClassName>-<classOop>-instances.bm
                          where className is the name of the class and 
                          classOop is the object ID of the class.

     3    listInPageOrder This operation requires a result file specification
                          (passed in the toDirOrFile argument).  It is an 
                          error if the file already exists.  The oops of the 
                          objects in the repository that are instances
                          of the classes specified are written to the file
                          in page order and take 5 bytes each.   Returns a 
                          SmallInteger representing the total number of 
                          objects found.

     4    objsInSecPolicy Returns an Array.  For each element of 
                          anIdentitySet, the result array contains 2 
                          elements: the total number of instances and 
                          an array of the instances (possibly limited in size)
                          that have the corresponding objectSecurityPolicyId.

     5    objsInSecPolicy The input set must contain a single securityPolicy.
                          The instances found are stored into the hidden
                          set: ListInstancesResult, and the number of 
                          elements found is the primitive return value.

     6    objsInSecPolicy Returns an Array of pairs where the odd numbered 
                          elements are the security policy ids specified 
                          in the IdentitySet (in order) and the even numbered 
                          elements are the number of instances in the preceding 
                          security policy that were found.

                          If the toDirectory argument is nil, then only 
                          array described above is returned.  If not nil, 
                          then  Bitmaps containing the object ID's of 
                          the instances are written to binary bitmap 
                          files in the specified directory.  The resulting
                          bitmap files are named: 
                             <ClassName>-<classOop>-instances.bm
                          where className is the name of the class and 
                          classOop is the object ID of the class.


     7    objsLargerThan  The input set must contain a single smallInteger
                          which is the target size for the scan. An array
                          of the objects which are larger than the target
                          size is returned. Result is limited by the
                          aLimit argument.

   ===========================================================================

   This method aborts the current transaction.  Because the view can change,
   the results of the scan can vary depending on which view was used when 
   a given object was scanned.
   
   The argument anIdentitySet must not be larger than 2030 elements.

   Scans the entire Repository once."

<primitive: 896>

maxThreads _validateClass: SmallInteger .
lockWaitTime _validateClass: SmallInteger .
aBufSize _validateClass: SmallInteger ; _validateIsPowerOf2 .
self _validatePercentage: aPercent .
anIdentitySet _validateClass: IdentitySet .
aLimit _validateClass: SmallInteger .
kind _validateClass: SmallInteger .
^ self _primitiveFailed: 
       #_scanPomWithMaxThreads:waitForLock:pageBufSize:percentCpuActiveLimit:identSet:limit:scanKind:toDirectory:
       args: { maxThreads . lockWaitTime . aBufSize . aPercent .
               anIdentitySet . aLimit . kind . aStringOrNil }
%

category: 'Multithreaded Scan Tuning'
method: Repository
setMultiThreadedScanThreads: numThreads maxCpu: aPercent

"Change the number of threads active for a multithreaded scan. 
 Also changes the maximum CPU usage parameter if aPercent is not nil.

 Returns true if successful, false if the session running the multithreaded
 scan could not be determined.

 Raises an error if the session holding the GC lock is not running on this
 host or if that session is not performing a multithreaded repository
 scan.

 Raises an error if numThreads is greater than the maximum number of threads for
 which the scan was configured (i.e., the value of the maxThreads argument
 passed to the multithreaded scan method).

 Requires the SessionAccess and SystemControl privileges."

| sesId |
sesId := System sessionIdHoldingGcLock .
sesId == 0
  ifTrue:[ RepositoryError signal: 'No scan in progress' ]. 

self getMultiThreadedScanMaxThreads == 0
  ifTrue:[ RepositoryError signal: 'No scan in progress' ]. 

  self mtThreadsLimit: sesId setValue: numThreads .
aPercent == nil
  ifFalse:[ self mtPercentCpuActiveLimit: sesId setValue: aPercent ] .
^ true "success if we get here"
%

category: 'Multithreaded Scan Tuning'
method: Repository
getMultiThreadedScanMaxThreads

"Answer the maximum number of threads that the current multithreaded scan
 is configured to use.  Answers 0 if no multithreaded scan is in progress.

 Raises an error if the session holding the GC lock is not running on this
 host or if the GC lock session is not performing a multithreaded repository
 scan.

 Requires the SessionAccess privilege."

| sesId |
sesId := System sessionIdHoldingGcLock .
sesId == 0
  ifTrue:[ RepositoryError signal: 'No scan in progress' ]. 

^ self mtMaxThreads: sesId
%

category: 'Multithreaded Scan Tuning'
method: Repository
pauseMultiThreadedScan

"Suspends any multithreaded scan in progress by setting the number of threads 
 available to the scan to 0.  The session running the scan is not stopped or 
 killed, it is in a suspended state waiting to be resumed.

 Returns true if successful, false if the session running the multithreaded
 scan could not be determined.

 Raises an error if the session holding the GC lock is not running on this
 host or if the GC lock session is not performing a multithreaded repository
 scan.

 Examples of multithreaded scan operations include:
   markForCollection
   findDisconnectedObjects
   listInstances
   listReferences
   objectAudit

 Multithreaded scan operations always hold the GC lock.  However not all 
 operations which hold the GC lock are multithreaded scans.

 The scan may be resumed by executing one of the setMultiThreadedScan* methods
 in this class.

 Requires the SessionAccess and SystemControl privileges."

^ self setMultiThreadedScanThreads: 0  maxCpu: nil "nil means no change"
%

category: 'Multithreaded Scan Tuning'
method: Repository
setMultiThreadedScanMaxAggressive

"Makes any multithreaded scan in progress as aggressive as possible.  This
 will cause the scan to complete as quickly as possible at the expense of
 consuming more system resources.

 Returns true on success, false if the multithreaded scan session could not
 be found on this host.

 Raises an error if the session holding the GC lock is not running on this
 host or if that session is not performing a multithreaded repository
 scan."

^ self setMultiThreadedScanThreads: 
       self getMultiThreadedScanMaxThreads maxCpu: 95 .
%

category: 'Multithreaded Scan Tuning'
method: Repository
setMultiThreadedScanMinAggressive

"Makes any multithreaded scan in progress run as slowly as possible without 
 suspending the scan.  This will cause the scan to take much more time to 
 complete and will greatly reduce the system resources used by the scan.

 Requires the SessionAccess and SystemControl privileges.

 Returns true on success, false if the multithreaded scan session could not
 be found on this host.

 Raises an error if the session holding the GC lock is not running on this
 host or if that session is not performing a multithreaded repository
 scan."

self getMultiThreadedScanMaxThreads == 0
  ifTrue:[ RepositoryError signal: 'No scan in progress' ]. 
^ self setMultiThreadedScanThreads: 1 maxCpu: 60 .
%

category: 'Multithreaded Scan Tuning'
method: Repository
setMultiThreadedScanMediumAggressive

"Makes any multithreaded scan in progress moderately aggressive.
 This is a compromise between completing as quickly as possible and the
 overall impact on the system.

 Returns true on success, false if the multithreaded scan session could not
 be found on this host.

 Raises an error if the session holding the GC lock is not running on this
 host or if that session is not performing a multithreaded repository
 scan."

| maxThds |
maxThds := self getMultiThreadedScanMaxThreads .

"Use 50% of the maximum number of threads configured."
maxThds := 1 min: (maxThds // 2) .

^ self setMultiThreadedScanThreads: maxThds maxCpu: 80 .
%

category 'Private'
method: Repository
_mtStat: sesId at: statCode

"Returns the value of the specified stat.
 statCode values are:
    0 - mtMaxThreads
    1 - mtThreadsLimit
    2 - mtPercentCpuActiveLimit
 
 Requires the SessionAccess and SystemControl privileges.

 Raises an error if no session with the specified ID was found on this
 host or if the session is not performing a multithreaded repository
 scan."

<primitive: 878>

sesId _validateClass: SmallInteger .
statCode _validateClass: SmallInteger .
self _primitiveFailed: #_mtStat:at: args: { sesId . statCode }
%

category 'Private'
method: Repository
_mtStat: sesId at: statCode put: newValue

"Sets the value of the specified stat.
 statCode values are:
    1 - mtThreadsLimit
    2 - mtPercentCpuActiveLimit

 Requires the SessionAccess and SystemControl privileges.

 Raises an error if no session with the specified ID was found on this
 host or if the session is not performing a multithreaded repository
 scan."

<primitive: 879>

sesId _validateClass: SmallInteger .
statCode _validateClass: SmallInteger .
self _primitiveFailed: #_mtStat:at:put: args: { sesId . statCode . newValue }
%

category: 'Multithreaded Scan Tuning'
method: Repository
mtMaxThreads: sessionId

"Returns the current value of the mtMaxThreads for the specified sessionId.
 The process executing this method must be on the same machine as the specified session.

 Requires the SessionAccess and SystemControl privileges."

^ self _mtStat: sessionId at: 0
%

category: 'Multithreaded Scan Tuning'
method: Repository
mtThreadsLimit: sessionId

"Returns the current value of the mtThreadsLimit for the specified sessionId.
 The process executing this method must be on the same machine as the specified session.

 Requires the SessionAccess and SystemControl privileges."

^ self _mtStat: sessionId at: 1
%

category: 'Multithreaded Scan Tuning'
method: Repository
mtThreadsLimit: sessionId setValue: newVal

"Sets a new value for the mtThreadsLimit for the specified sessionId.
 The process executing this method must be on the same machine as the specified session.
 The newVal may be 0, which can be used to temporarily pause the operation in progress.
 The maximum newVal allowed is determined by the current value of mtMaxSessions.
 Requires the SessionAccess and SystemControl privileges."

^ self _mtStat: sessionId at: 1 put: newVal
%

category: 'Multithreaded Scan Tuning'
method: Repository
mtPercentCpuActiveLimit: sessionId

"Returns the current value of the mtPercentCpuActiveLimit for the specified sessionId.
 The process executing this method must be on the same machine as the specified session.

 Requires the SessionAccess and SystemControl privileges."

^ self _mtStat: sessionId at: 2
%

category: 'Multithreaded Scan Tuning'
method: Repository
mtPercentCpuActiveLimit: sessionId setValue: newVal

"Sets a new value for the mtThreadsLimit for the specified sessionId.
 The process executing this method must be on the same machine as the specified session.
 The newVal may be in the range from 0 to 100.

 Requires the SessionAccess and SystemControl privileges."

^ self _mtStat: sessionId at: 2 put: newVal
%

category: 'Audit and Repair'
method: Repository
_objectAuditWithMaxThreads: maxThreads waitForLock: waitTimeSeconds 
                       pageBufSize: aBufSize 
                       percentCpuActiveLimit: aPercent
                       csvFile: aFileName
                       repair: doRepair 

"Performs a multi-threaded object audit on all objects in the repository.

 Requires SystemControl privileges.

 See _scanPomWithMaxThreads for definitions of maxThreads, waitTimeSeconds, 
 aBufSize, aPercent.

 The following errors can be generated:
   BadClassId     - object references classId which does not exist.
   BadClass       - object class reference is not a class.
   BadFormat      - the object format disagrees with the class format.
   BadSecPolicy   - the object security policy is invalid.
   BadPhysSize    - the object physical size is invalid.
   BadLogicalSize - the object logical size is greater than the physical size.
   BadLength      - the object length > bodySize.
   BadBody        - the object body contains a reference to oop illegal.
   BadReference   - the object references a non existent object at offset.
   BadOt          - is in page but has an invalid OT entry.
   InFreeOops     - the object is an a page and in the FreeOop list.
   FreeOop        - does not exist but is neither in dead nor in the FreeOop list.
   BadObjId       - invalid objectId found in object.
   BadDependency  - the OT indicates the object has a dependency but it is not in depMap.
   BadDepMapKey   - the objectId is in the depMapKeys but is not a valid object.
   BadDepMapValue - the objectId is in the depMapValues but is not a valid object.
 
 These are always printed to the GciErrorLog.  If a valid file name is provided for the 
 csvFile (comma separated values) argument then a line is written to the file for each 
 error encountered.  The lines contain the following information: 
   1. ErrorKind - the name associated with each of the errors defined above.
   2. ObjectId  - the objectId for the object with an error.
   3. ClassName  - the name of the object's class.
   4. Offset    - the offset (or other integer value, e.g. size)
   5. Reference - the reference that does not exist.
 If no errors are found, the file is deleted.

 If doRepair is true, the repairs are performed after the entire scan of the repository
 is complete.  It is possible that if a very large number of errors are found the 
 system may run out of memory to record the information needed to fix the errors.
 In this case the following warning message will be generated:
    WARNING: Ran out of memory to record errors, remaining errors will not be fixed.
 If this is found in the output, then the errors before this message will be fixed, but
 any that are printed after this message will not be fixed and the audit will need to be
 rerun to correct any additional errors.  In addion an additional warning is generated at 
 the end of the repair output.

 During the repair phase, each error will be printed out along with a description of the
 action taken to repair the problem.  The system will commit after every 2000 repair 
 operations to avoid overflowing temporary memory space.

 The only errors that cannot be corrected by the audit code are the InFreeOops and FreeOop
 error kinds.  To fix these you will need to rebuild the FreeOop list - see admin guide for 
 instructions on how to rebuild the FreeOop list."

<primitive: 392>
maxThreads _validateClass: SmallInteger.
waitTimeSeconds _validateClass: SmallInteger.
aBufSize _validateClass: SmallInteger ; _validateIsPowerOf2 .
self _validatePercentage: aPercent .
aFileName _validateClass: String.
doRepair _validateClass: Boolean.
^ self _primitiveFailed:
       #_objectAuditWithMaxThreads:waitForLock:pageBufSize:percentCpuActiveLimit:csvFile:repair:
       args: { maxThreads . waitTimeSeconds . aBufSize .
               aPercent . aFileName . doRepair }
%

category: 'Backup and Restore'
method: Repository
fullBackupTo: fileName

"Writes a full backup file containing the most recently committed 
 version of  the receiver as of the time the method is executed.

 This method always writes the file without compression. To compress
 the backup files use fullBackupCompressedTo:

 The fileName argument may use GemStone Network Resource String syntax.
 For example, this may be used to access a file on another machine, provided
 a GemStone NetLDI process is running on the remote machine.  

 If the device containing the file runs out of space, then the backup 
 terminates with a system I/O error and the partially written backup 
 file is deleted.

 If the session contains uncommitted changes to the repository, the method 
 signals a error: #rtErrAbortWouldLoseData, to indicate that data could 
 be lost.  Otherwise, it puts the session in autoBegin mode and performs 
 the backup operation.  The session may be aborted a number of times during
 the backup to prevent a commit record backlog.  

 When the backup completes, the session is set to manualBegin mode
 mode so that it does not reference a commit record  that would cause 
 the repository to grow.

 Returns true if the backup was completed.

 This method requires the FileControl privilege.

 This method performs the backup using multiple slave sessions.  
 The number of slave sessions is automatically determined from the
 number of extents in the repository, with a minimum of 2 sessions and
 a maximum of 16.  The performance can be modified during the run by 
 updating the Multithreaded Scan Tuning methods.

 A GciHardBreak during this method will terminate the session."

  ^ self fullBackupTo: { fileName } MBytes: 0
%

category 'Private'
method Repository
_validateFileNames: fileNames limits: mByteLimit
  | limitArray |

  System needsCommit
    ifTrue: [ self _error: #rtErrAbortWouldLoseData ].

  fileNames _validateClass: Array.
  fileNames size == 0 ifTrue: [ fileNames _error: #rtErrArgOutOfRange args:{ 1 . 256 }].
  1 to: fileNames size do: [ :i |
    (fileNames at: i) _validateClass: String].
  (mByteLimit isKindOfClass: SmallInteger)  ifTrue:[
     ( (mByteLimit < 0) or: [ mByteLimit > (4000 * 1024) ])
        ifTrue: [ mByteLimit _error: #rtErrArgOutOfRange args: {0 . 4000 * 1024}] .
      limitArray := Array new: fileNames size.
      1 to: limitArray size do: [ :i |
        limitArray at: i put: mByteLimit ] ]
    ifFalse: [ 
      (mByteLimit isKindOfClass: Array)
         ifTrue:  [ limitArray := mByteLimit ]
         ifFalse: [ ( ArgumentTypeError new name: 'mByteLimit' expectedClass: 
                    { Array . SmallInteger } actualArg: mByteLimit ) signal].
    ].
  limitArray at: limitArray size put: 0.
  ^limitArray
%


category: 'Backup and Restore'
method: Repository
fullBackupTo: fileNames MBytes: mByteLimit 

"The same as fullBackupTo: except that it accepts an array of filenames 
 as the first argument so that the backup may be spread across multiple
 files on different devices.  In addition, the mByteLimit argument 
 specifies the maximum size of the files.  The mByteLimit may be a single 
 integer which is applied to each file in the fileNamesArray or
 an array of integer values, where the array must be the same size 
 as the fileNamesArray.

 A zero value for the megabyte limit means that there is no limit on the 
 size of the resulting file(s) and is an easy way to specify that the backup
 should partition the data equally across the files.  The maximum value
 of a backup file even with no limit is 4TB. A mByteLimit value less 
 than zero generates an error.

 This method overides the mByteLimit for the last file in the array
 and sets it to 0 to prevent the backup from failing due to lack of 
 space unless the device used for the last backup file is full."

  | limitArray fnArray |
  (fileNames isKindOfClass: Array) 
    ifTrue: [ fnArray := fileNames ]
    ifFalse: [ fnArray := { fileNames } ].
  limitArray := self _validateFileNames: fnArray limits: mByteLimit.
  ^ self _fullBackupTo: fnArray MBytes: limitArray compress: false.
%

category: 'Backup and Restore'
method: Repository
fullBackupCompressedTo: fileName

"The same as fullBackupTo: except that the backup file is compressed
 as it is written.

 If the fileName is does not end in '.gz' then a '.gz' suffix is added."

  ^ self fullBackupCompressedTo: { fileName } MBytes: 0
%


category: 'Backup and Restore'
method: Repository
fullBackupCompressedTo: fileNames MBytes: mByteLimit

"The same as fullBackupTo:MBytes except that the backup file 
 is compressed as it is written.

 If the fileName is does not end in '.gz' then a '.gz' suffix is added."

  | limitArray fnArray |
  (fileNames isKindOfClass: Array) 
    ifTrue: [ fnArray := fileNames ]
    ifFalse: [ fnArray := { fileNames } ].
  limitArray := self _validateFileNames: fnArray limits: mByteLimit.
  ^ self _fullBackupTo: fnArray MBytes: limitArray compress: true.
%

category: 'Backup and Restore'
method: Repository
_fullBackupTo: fileNames MBytes: limitArray compress: aBool

"Private.  Provides implementation of backup methods.

 Returns true if the backup was completed."

  <primitive: 394>
  fileNames _validateClass: Array.
  limitArray _validateClass: Array.
  aBool _validateClass: Boolean.
  self _primitiveFailed: #_fullBackupTo:MBytes:compress: .
%

category: 'Backup and Restore'
method: Repository
restoreStatusInfo

"Returns an Array describing the current restore status of the Repository,  
 including the next transaction log file or backup file required to continue
 the restore.  The Array contains the following elements:
   
   1: a String describing the restore status.
   2: a SmallInteger, 0 restore not active, 2 = active from log
   3: a SmallInteger, next fileId to restore 
   4: a SmallInteger, -1 (unused element)
   5: an Integer,  firstStartupTime of the repository   
   6: an Integer,  commitSequence.high of the repository
   7: an Integer,  commitSequence.low of the repository
   8: a SmallInteger, fileId of oldest tranlog that must be present for restore.
   9: a SmallInteger, fileId of current restore position
   10: a SmallInteger, blockId of current restore position
   11: a String, the time of the last checkpoint restored to.
   12: a SmallInteger, 0 no log receiver, 1 log receiver connecte to stone.

 If restore is not active, the elements 3 to 11 are nil.

 Restore status is an attribute of the Repository, not of the session,
 and persists across logout/login and stopstone/startstone."

^ System _zeroArgPrim: 143
%

category: 'Backup and Restore'
method: Repository
restoreStatus
  
"Returns a String describing the current restore status of the Repository,
 including the next transaction log file or backup file required to continue
 the restore, and the oldest transaction log file that must be present
 for a restore from log."

^ self restoreStatusInfo at: 1
%
     
category: 'Backup and Restore'
method: Repository
restoreActive

"Returns true if the restore is currently active, false if not."

^(self restoreStatusInfo at: 2) ~= 0
%

category: 'Backup and Restore'
method: Repository
restoreStatusNextFileId
    
"Returns a SmallInteger, the fileId of the next transaction log or backup that
 should be restored, or nil if restore not active."
  
^ self restoreStatusInfo at: 3
%

category: 'Backup and Restore'
method: Repository
restoreStatusOldestFileId
 
"Returns a SmallInteger, the fileId of the oldest transaction log
 that needs to be present for the next restore from log operation,
 or nil if restore not active."
 
^ self restoreStatusInfo at: 8
%   

category: 'Backup and Restore'
method: Repository
restoreStatusTimeRestoredTo
 
"Returns a String which represents time of the last checkpoint 
 that the repository was restored to or nil if restore not active."
 
^ self restoreStatusInfo at: 11 
%   

category: 'Backup and Restore'
method: Repository
commitRestore

"Terminates a restore operation and puts the system back into a normal 
 operating mode in which commits are allowed and reenables logins.

 The repository must be in the restoreFromLogs state, otherwise this 
 method generates an error.

 If the logs have not been restored to the end of the known logs, i.e., 
 the last restore operation was a restoreToEndOfLog: or restoreToPointInTime:
 then a warning is issued ane the restore is stopped at the current 
 location.  This use of commitRestore can result in failure to restore 
 all previously committed transactions.  However, this allows a Repository 
 to be restored as far as practical when some log files are lost or corrupted.

 Upon successful completion, the session is automatically logged out and
 the RestoreLogSuccess error (4048) is generated.

 This method requires the FileControl privilege.  It is recommended
 that it be run by either DataCurator or SystemUser.

 Because any session that logs in or performs a restore operation 
 is prevented from doing normal commits, this session is automatically
 logged out when the restore operation completes."

self restoreActive ifFalse: [
  ImproperOperation signal: 
     'You cannot commitRestore without first executing restoreFromBackup'].
^ System _zeroArgPrim: 141
%


category: 'Backup and Restore'
method: Repository
restoreFromBackup: fileName

"Disables logins and starts a full restore of the repository based 
 on the contents of the specified backup file.

 Restored objects are clustered in a manner that is similar, but not
 necessarily identical to the clustering at the time the  backup file 
 was created.  If the Repository being restored into has the same 
 number of extents as the system had when the backup file was created, 
 then distribution of objects within extents is preserved unless one 
 of the extents becomes full.  If the number of extents is different 
 than the number when the backup was created, then the current 
 DBF_ALLOCATION_MODE configuration controls the distribution of 
 objects across the extents.  

 If the backup file was made when in partial-logging mode, then logins
 are reenabled and the system is fully operational when the restore 
 completes.  If the backup was made when in full-logging mode, then 
 the system is not fully operational until tranlogs have been 
 restored and the commitRestore method is executed.

 If the backup file was created with the compress: argument true or
 if the file was compressed using 'gunzip', then the input file should
 be specified with the '.gz' suffix.

 To minimize the size of the resulting repository the stone should be
 restarted on a copy of the the initial repository (copied from 
 $GEMSTONE/bin/extent0.dbf).

 To optimize the time to restore the backup, the exents in the new
 repository should be pregrown to the minimum expected size of 
 the restored repository.

 Upon successful completion, the session is automatically logged out and
 the RestoreBackupSuccess error (4046) is generated.

 A GciHardBreak during this method will terminate the session.

 This method requires the FileControl privilege.  It is recommended
 that it be run by either DataCurator or SystemUser.

 This method performs the restore using multiple slave sessions.  
 The number of slave sessions is automatically determined from the
 number of extents in the repository, with a minimum of 2 sessions and
 a maximum of 16.  The performance can be modified during the run by 
 updating the Multithreaded Scan Tuning methods."

^ self restoreFromBackups: { fileName }
%

category: 'Backup and Restore'
method: Repository
restoreFromBackups: arrayOfFileNames

"Similar to restoreFromBackup: except that it accepts an array of fileNames."

^ self _restoreBackups: arrayOfFileNames scavPerentFree: 100
%

category: 'Backup and Restore'
method: Repository
restoreFromBackups: arrayOfFileNames scavengePagesWithPercentFree: aPercent

"Same as restoreFromBackups, except that restored pages that have greater
 than aPercent free space are added to the scavengable pages at the end
 of the restore.  WARNING: a small percentage, less than 5 percent can cause
 the reclaim gems to be very busy after the restore."

^ self _restoreBackups: arrayOfFileNames scavPerentFree: aPercent
%

category: 'Backup and Restore'
method: Repository
_restoreBackups: arrayOfFileNames scavPerentFree: aPercent

"Private.  Provides implementation of restoreBackups."

<primitive: 822>
arrayOfFileNames _validateClass: Array.
aPercent _validateClass: SmallInteger.
self _primitiveFailed: #restoreBackups:scavPercentFree: .
self _uncontinuableError
%

category 'Private'
method: Repository
_checkRestoreActive

self restoreActive ifFalse: [
  ImproperOperation signal: 
     'You cannot restoreLogs without first executing restoreFromBackup'].
%

category: 'Backup and Restore'
method: Repository
restoreFromCurrentLogs

"After a restoreFromBackup: successfully completes this method may be 
 executed to replay transactions which occurred since the backup was made.
 If the restorestatus of the repository is not activeFromLog then an error 
 is generated.
 
 This method attempts to sequentially replay all transaction logs from 
 the first transaction log required to the end of the current logs. 

 If previous restore operation was a variant of restoreFromArchiveLogs,
 the last archive log will be reopened and any data written since
 the last EOF postion restored, before restoring the first current log.

 The restoreStatus or restoreStatusNextFileId methods can be used to determine 
 the next file required for a restoreLogs operation.

 If some of the tranlogs have been archived and are not in the locations
 configured with the STN_TRAN_LOG_DIRECTORIES, then the restoreFromArchiveLogs:
 methods should be used to retore up until the next files are in the 
 current tranlog directories.

 When opening a log file on a file system, if the filename ends in '.gz',
 the file is expected to be in gzip compressed format.  Restoring a compressed 
 tranlog from a raw partition is not supported.

 Upon successful completion, the session is automatically logged out and
 the RestoreLogSuccess error (4048) is generated.

 If an error occurs, such as a corrupted tranlog or a missing tranlog in 
 the sequence then an error is generated and the session is not loggged out.

 This method requires the FileControl privilege.  It is recommended
 that it be run by either DataCurator or SystemUser."

self _checkRestoreActive.
^ self _restoreLogs: -1 time: 0 archiveLogs: 0
%


category: 'Backup and Restore'
method: Repository
restoreToEndOfLog: fileId

"Similar to restoreFromCurrentLogs except that the restore stops when it has 
 completed replaying the transactions in the file specified by fileId.

 The fileId argument must be a positive SmallInteger.

 A subsequent restore operation may be used to continue restoring logs
 past the log location at which this operation stopped."

self _checkRestoreActive.
fileId _validateClass: SmallInteger.
fileId > 0 ifFalse:[ fileId _error: #rtErrArgNotPositive ].
^ self _restoreLogs: fileId time: 0 archiveLogs: 0
%

category: 'Private'
method: Repository
_checkPointInTime: aDateTime

| checkpointTime aTimeT |
checkpointTime := DateTime fromString: 
  ((self restoreStatusTimeRestoredTo) copyFrom: 1 to: 17) usingFormat: #( 1 2 3 $/ 1 2 $: true true false ).
aDateTime < checkpointTime  ifTrue: [ 
  ImproperOperation signal: ( 'aDateTime ' , aDateTime asString ,  
    ' is less than restoreStatusTimeRestoredTo' ,
     self restoreStatusTimeRestoredTo)].
aTimeT := aDateTime asSecondsGmt - 2177452800.
(aTimeT < 0 or:[ aTimeT > 16r7fffffff]) ifTrue:[
  aDateTime _error: #rtErrArgOutOfRange args:{ 0 . 16r7fffffff }
  ].
^ aTimeT
%

category: 'Backup and Restore'
method: Repository
restoreToPointInTime: aDateTime

"Similar to restoreToEndOfLog except that the restore stops the restore at 
 the first checkpoint which originally occurred at or after aDateTime.

 An error is generated if aDateTime precedes the time of the last restored
 checkpoint."

| aTimeT |
self _checkRestoreActive.
aTimeT := self _checkPointInTime: aDateTime.
self _restoreLogs: -1 time: aTimeT archiveLogs: 0
%

category: 'Backup and Restore'
method: Repository
_restoreLogs: fileId time: aTime archiveLogs: useArchiveInt

"Private.  Provides implementation of restoreFromLogs.
 If _primSetArchiveLogDirectories was called and
   useArchiveInt == 1 , archive logs will be used ;
   useArchiveInt == 2 , continuous restore entered using archive logs.
   useArchiveInt == 3 , stop continuous restore.
 "

<primitive: 395>
fileId _validateClass: SmallInteger.
aTime _validateClass: SmallInteger.
useArchiveInt _validateClass: SmallInteger. 
useArchiveInt >= 2 ifTrue:[
  fileId == -1 ifFalse:[  
    ArgumentError signal: 'fileId must be -1 for continuous restore'.
  ].
  aTime == 0 ifFalse:[
    ArgumentError signal: 'aTime must be 0 for continuous restore'.
  ].
].
self _primitiveFailed: #_restoreLogs:time:archiveLogs:.
self _uncontinuableError
%

category: 'Backup and Restore'
method: Repository
restoreFromArchiveLogs: archiveLogDir

"Similar to restoreFromLogs except that the restore operation looks
 for the logs to restore in the directory or array of directories 
 specified instead of the currently configured STN_TRAN_LOG_DIRECTORIES.

 archiveLogDir may be a single directory specification, 
 or an Array  of directory specifications.
 
 Each directory specification must be a String that names a file system 
 directory.  It is an error if any of the directories specified does
 not exist.  The directory specified may be empty.

 Environment variables , if any, in the directory specifications 
 are expanded using the environment of the gem  or topaz -l) process
 running this session, before the directory specifications are passed to 
 stone for execution of the restore operation.  
 Environment variables are the suggested way to alleviate interactive 
 usability issues caused by removal of the setArchiveLogDirectories: method.

 Terminates when all of the logs in the specified directories have
 been restored or if an error is discovered while processing the logs."

self restoreFromArchiveLogs: archiveLogDir withPrefix: nil
%

category: 'Backup and Restore'
method: Repository
restoreFromArchiveLogs: archiveLogDir withPrefix: aPrefix

"Similar to restoreFromLogs except that the restore operation looks
 for the logs to restore in the directory or array of directories 
 specified instead of the currently configured STN_TRAN_LOG_DIRECTORIES.
 Each directory specification must be a String that names a file system 
 directory.  It is an error if any of the directories specified does
 not exist.  The directory specified may be empty.
 
 aPrefix specifies file name pattern in sequence instead of standard tranlog.
 aPrefix must be astring if nil default tranlog<n> is used
 
 Terminates when all of the logs in the specified directories have
 been restored or if an error is discovered while processing the logs."

self _checkRestoreActive.
self _setArchiveLogs: archiveLogDir withPrefix: aPrefix.
self _restoreLogs: -1 time: 0 archiveLogs: 1 .
%

method: Repository
continuousRestoreFromArchiveLogs: archiveLogDir

"Similar to restoreFromArchiveLogs: except that stone enters
 continuous restore mode, and control is returned to this session
 executing this method. Intended for use on a hot standby stone 
 where the archive logs are being written by a log receiver process
 which is receiving data from the master stone's log feeder process.
 Waits up to 15 seconds for a log receiver to connect to stone, after
 which if no log recever is detected, signals an error.

 The restore can be terminated by  Repository >> stopContinuousRestore.

 Repository >> restoreStatus is usable while the continuous restore
 is running.
"

| ready printed count |
ready := false .
printed := false .
count := 0 .
[ ready ] whileFalse:[ | info |
  info := self restoreStatusInfo .
  (info at: 2) == 0 ifTrue:[ self _checkRestoreActive ].
  (info at: 12) == 1 ifTrue:[ 
    ready := true 
  ] ifFalse:[
    printed ifFalse:[ printed := true .
       GsFile gciLogClient:'  [Info:] Waiting for a log receiver to connect to stone.' ].
    count := count + 1 .
    count > 15 ifTrue:[ ImproperOperation signal:
	'continuousRestore cannot start, no log receiver connected to stone.' ].
    System sleep: 1 .
  ]
].
self _setArchiveLogs: archiveLogDir withPrefix: nil.
self _restoreLogs: -1 time: 0 archiveLogs: 2 .
%

method: Repository
stopContinuousRestore

"Terminate a previous continuousRestoreFromArchiveLogs: .
 Waits for tranlogs found in the archive logs directory
 to be restored, but does not wait for more logs to be
 received by the logreceiver process .
 Has no effect if no such restore is active."

  self _checkRestoreActive.
  self _restoreLogs: -1 time: 0 archiveLogs: 3
%


category: 'Backup and Restore'
method: Repository
restoreFromArchiveLogs: archiveLogDir toEndOfLog: fileId

"Similar to restoreFromArchiveLogs: except that the restore stops 
 when the end of the log with fileId is reached."

self restoreFromArchiveLogs: archiveLogDir toEndOfLog: fileId withPrefix: nil
%

category: 'Backup and Restore'
method: Repository
restoreFromArchiveLogs: archiveLogDir toPointInTime: aDateTime

"Similar to restoreFromArchiveLogs: except that the restore stops at 
 the first checkpoint which originally occurred at or after aDateTime.

 An error is generated if aDateTime precedes the time of the last restored
 checkpoint."

 self restoreFromArchiveLogs: archiveLogDir toPointInTime: aDateTime withPrefix: nil
%

category: 'Backup and Restore'
method: Repository
restoreFromArchiveLogs: archiveLogDir toEndOfLog: fileId withPrefix: prefixString

"Similar to restoreFromArchiveLogs:toEndOfLog: except that the specified
 prefixString is used instead of the value of the configuration parameter
 specified by STN_TRAN_LOG_PREFIX."

self _checkRestoreActive.
self _setArchiveLogs: archiveLogDir withPrefix: prefixString.
self _restoreLogs: fileId time: 0 archiveLogs: 1
%

category: 'Backup and Restore'
method: Repository
restoreFromArchiveLogs: archiveLogDir toPointInTime: aDateTime withPrefix: prefixString

"Similar to restoreFromArchiveLogs:toPointInTime: except that the specified
 prefixString is used instead of the value of the configuration parameter
 specified by STN_TRAN_LOG_PREFIX."

| aTime |
self _checkRestoreActive.
aTime := self _checkPointInTime: aDateTime.
self _setArchiveLogs: archiveLogDir withPrefix: prefixString.
^ self _restoreLogs: -1 time: aTime archiveLogs: 1
%

category: 'Private'
method: Repository
_validateString: aString

aString == nil ifFalse: [
  (aString class isSubclassOf: String) ifFalse:[ 
    aString _errorExpectedClass: String ].
  aString size > 1023 ifTrue:[ aString _error: #errArgTooLarge args:{ 1023 } ].
  ].
%

category: 'Private'
method: Repository
_setArchiveLogs: dirArg withPrefix: prefixString
  "dirArg may be a String or an Array of Strings"
| logDirs |
dirArg _isOneByteString
   ifTrue: [ logDirs := { dirArg } ]
   ifFalse: [ logDirs := dirArg ].
logDirs _validateClass: Array.
logDirs size < 1 ifTrue:[
  logDirs _error: #errArgTooSmall args:{ 1 }.
].
logDirs do:[ :aString | self _validateString: aString ].
self _validateString: prefixString.
^ self _primSetArchiveLogDirectories: logDirs withPrefix: prefixString
%

category: 'Private'
method: Repository
_primSetArchiveLogDirectories: logDirs withPrefix: prefixString

"Send archive log directories specification to Stone.  Sender has
 done class kind checks of all the arguments.  Stone checks for
 existence of each directory.  arrayOfDirectorySpecs should be non-empty.
 If the prefixString is nil, the STN_TRAN_LOG_PREFIX configuration 
 parameter is used."

<primitive: 469>
self _primitiveFailed: #_primSetArchiveLogDirectories:withPrefix:
     args: { logDirs . prefixString }
%
category: 'Deprecated'
method: Repository
setArchiveLogDirectory: aDirectorySpec

  "Disallowed. You must specify the archive log directory on
   each restoreFromArchiveLogs:  invocation.  This avoids confusion
   caused by archive log locations being remembered by stone and
   thus not being visible to the topaz script."
  
  ^ self shouldNotImplement: #setArchiveLogDirectory:
%
method: Repository
setArchiveLogDirectories: arrayOfDirectorySpecs

  "Disallowed. You must specify the archive log directories on
   each restoreFromArchiveLogs:  invocation.  This avoids confusion
   caused by archive log locations being remembered by stone and
   thus not being visible to the topaz script."

  ^ self shouldNotImplement: #setArchiveLogDirectories:
%
