!=========================================================================
! Copyright (C) VMware, Inc. 1986-2012.  All Rights Reserved.
!
! $Id: writest.gs 28927 2012-08-22 22:26:20Z monty $
!
! Superclass Hierarchy:
!   WriteStream, PositionableStream, Stream, Object.
!
!=========================================================================

removeallmethods WriteStream
removeallclassmethods WriteStream

category: 'Documentation'
classmethod:
comment
^ 'WriteStream / WriteStreamLegacy is a PositionableStream that allows its 
 objects to be written, but not read.'.
%

category: 'Instance Creation'
classmethod: WriteStream
new

"Disallowed.  To create a new WriteStream, use the class method on: instead."

self shouldNotImplement: #new
%

category: 'Instance Creation'
classmethod: WriteStream
with: aCollection

"Returns an instance of the receiver open for writing. that is positioned at the end of 
 aCollection."
   | res | 
   (res := self on: aCollection) setToEnd .
   ^ res
%

category: 'Accessing'
method: WriteStream
next

"Disallowed.  You cannot read a WriteStream."

self shouldNotImplement: #next
%

! fix 37918/36521: add _nextPut:

category: 'Adding'
method: WriteStream
_nextPut: anObject
"see PrintStream>>_nextPut:"

itsCollection at: position put: anObject.
position := position + 1.
^ anObject
%

category: 'Adding'
method: WriteStream
nextPut: anObject

"Inserts anObject as the next element that the receiver can access for writing.
 Returns anObject."

^ self _nextPut: anObject
%

category: 'Adding'
method: WriteStream
nextPutAll: aCollection

"Inserts the elements of aCollection as the next elements that the receiver can
 access.  Returns aCollection."

position == (itsCollection size + 1)
ifFalse:
  [ ^ super nextPutAll: aCollection ].
itsCollection addAll: aCollection.
position := position + (aCollection size).
^ aCollection
%

category: 'Adding'
method: WriteStream
nextPutAllBytes: aCharacterCollection

"Inserts the byte contents of aCharacterCollection as the next elements 
 that the receiver can access.  Returns aCollection.  The receiver's collection
 must be a type of String."

"Used in the implementation of PassiveObject."

position == (itsCollection size + 1)
ifFalse:
  [ ^ super nextPutAllBytes: aCharacterCollection ].
itsCollection addAllBytes: aCharacterCollection.
position := position + (aCharacterCollection _basicSize).
^ aCharacterCollection
%

category: 'Accessing'
method: WriteStream
contents

"WriteStreams return the portion of their collection that has been written:
 the collection up to the next write-position."

position < 2 ifTrue: [ ^itsCollection species new ].
^itsCollection copyFrom: 1 to: position-1
%

category: 'ANSI'
method: WriteStream
flush
	"Update a stream's backing store.
	Upon return, if the receiver is a write-back stream, the state of the
	stream backing store must be consistent with the current state of the 
	receiver.
	If the receiver is not a write-back stream, the effect of this 
	message is unspecified."

	"We do nothing; this method is provided for ANSI compatibility"
%
